/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */

import jasco.Jasco;
import jasco.util.CommandLine;
import jasco.util.CommandRunner;
import jasco.util.InvalidCommandLineFlag;
import jasco.util.logging.Logger;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.Iterator;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.Vector;

public class Main 
{	
	private static final String FLAG_VERSION = "version";
	private static final String FLAG_CLASSPATH = "classpath";
	private static final String FLAG_OUTPUTDIR = "outputdir";
	private static final String FLAG_TARGET = "target";
	private static final String FLAG_X = "X";
	private static final String FLAG_K = "k";
	private static final String FLAG_D = "D";
	private static final String FLAG_A = "A";
	private static final String FLAG_L = "l";
	private static final String FLAG_DEBUG = "debug";
	private static final String FLAG_INIT = "init";
	private static final String FLAG_NOEVENTS = "noevents";
	private static final String FLAG_DISTRIBUTION = "distributed";

	public static void main(String [] args) 
	{	
	
		try {
			parseCommandLine(args);
		} catch (FileNotFoundException e) {
			Logger.getInstance().showError("File not found: "+e.getMessage());
		} catch (IOException e) {
			Logger.getInstance().showError("IO Exception: "+e.getMessage());
		} catch (InvalidCommandLineFlag e) {
			Logger.getInstance().showError("Invalid flag: "+e.getMessage());
		}
		
		
		
		
	}
	
	public static void parseCommandLine(String[] args) throws IOException, InvalidCommandLineFlag, FileNotFoundException {
//		jasco.util.Logger.setInstance(new jasco.util.ExceptionLogger());
		jasco.options.Options.loadOptions(false);
		
		CommandLine commandLine = new CommandLine(args,2);
		boolean correct = commandLine.parse();
		
		TreeMap flags = commandLine.getFlags();
		
		String version = (String) flags.get(FLAG_VERSION);
		if (version!=null)  {
			System.out.println(jasco.options.Options.getVersion());
			return;
		}
		
		String init = (String) flags.get(FLAG_INIT);
		if (init!=null) {
			jasco.options.Options.generateScripts();
			return;
		}
		
		if (commandLine.getArgs().size() == 0) {
			Logger.getInstance().showError( "Command required" );
			return;
		}
		String command = (String) commandLine.getArgs().get(0);
		
		String classPath = (String)flags.get(FLAG_CLASSPATH);
		String outputDir = (String)flags.get(FLAG_OUTPUTDIR);
		
		//default stuff: Generated
		if(classPath==null&&outputDir==null&&!jasco.options.Options.isContainedInClasspath("Generated"))
			classPath="Generated";
		
		if (classPath!=null)  {
			
			jasco.options.Options.addToClassPath(classPath);
			jasco.util.CommandRunner runner = new jasco.util.CommandRunner(false);
			String props=" -D\"jasco.installpath="+jasco.options.Options.getInstallPath()+"\"";
			props+=" -D\"jasco.debug="+jasco.options.Options.showDebugOutput()+"\"";
			String cmd = "java "+props+" -classpath \""+jasco.options.Options.getClassPath()+"\" Main";
			for(int i=0;i<args.length;++i)  
				if(!args[i].startsWith("-classpath"))
					cmd+=" \""+args[i]+"\" ";
			
			CommandRunner.runCommand(cmd,false);
			return;
		}
		//System.out.println(jasco.options.Options.getClassPath());
		
		String target = (String)flags.get(FLAG_TARGET);
		if(target!=null)
			jasco.options.Options.setTarget(target);
	
		
		jasco.Jasco aspect = new jasco.Jasco();
		
		if (command.equals("Introspect")) 
		{
			Jasco.createIntrospectorWindow();
			return;
		}
		
		Set legalFlags = new TreeSet();
		legalFlags.add(FLAG_OUTPUTDIR);
		legalFlags.add(FLAG_TARGET);
		legalFlags.add(FLAG_CLASSPATH);
		legalFlags.add(FLAG_K);
		legalFlags.add(FLAG_DEBUG);
		legalFlags.add(FLAG_DISTRIBUTION);
		
		boolean X = false;
		//System.out.println(flags.get("X"));
		if(flags.get(FLAG_X)!=null)
			X=true;	
		
		
		boolean A = false;
		//System.out.println(flags.get("X"));
		if(flags.get(FLAG_A)!=null)
			A=true;	
		boolean D = false;
		//System.out.println(flags.get("X"));
		if(flags.get(FLAG_D)!=null)
			D=true;
		boolean noevents = false;
		//System.out.println(flags.get("X"));
		if(flags.get(FLAG_NOEVENTS)!=null)
			noevents=true;
		boolean distribution = false;
		if(flags.get(FLAG_DISTRIBUTION)!=null)
			distribution=true;
		boolean k = false;
		//System.out.println(flags.get("X"));
		if(flags.get(FLAG_K)!=null)  
			k=true; 	
		if(flags.get(FLAG_DEBUG)!=null)
			jasco.options.Options.showDebugOutput(true);	
		
		if(k)
			jasco.options.Options.setDeleteTempFiles(false);
		if(noevents)
			jasco.options.Options.doTransformEvents(false);
		else jasco.options.Options.doTransformEvents(true);
		
		if(distribution)
			jasco.options.Options.setDistributed(true);
		 
		
		String file = "";
		if (!correct)  {
			printUsage(command);
			return;
		}
		else file=(String) commandLine.getArgs().elementAt(1);
		
		if(outputDir!=null)
			jasco.options.Options.setOutputDir(new File(outputDir).getCanonicalPath());	
		
			
		if(command.equals("CompileAspect"))  {
			legalFlags.add(FLAG_X);
			commandLine.checkFlags(legalFlags);
			jasco.options.Options.setTransformHooks(X);
			if (!checkFile(file))  {
				Logger.getInstance().showError( "Non-existing file: "+file);
				return;
			}
			
			Jasco.compileAspect(file);
		}
		else if(command.equals("CompileRefinement"))  {
			commandLine.checkFlags(legalFlags);
			if (!checkFile(file))  {
				Logger.getInstance().showError( "Non-existing file: "+file);
				return;
			}
			
			Jasco.compileRefinement(file);
		}
		else if(command.equals("CompileConnector"))  {
			//jasco.options.Options.loadClass("jasco.connectorparser.Cutpoint");
			legalFlags.add(FLAG_X);
			commandLine.checkFlags(legalFlags);
			if (!checkFile(file))  {
				Logger.getInstance().showError( "Non-existing file: "+file);
				return;
			} 
			
			if(X)
				jasco.options.Options.setTransformBean(true);	
			Jasco.compileConnector(file);
		}
		else if (command.equals("RemoveConnector")) 
		{
			Jasco.removeConnectorInOutputDir(file);
		}
		else if (command.equals("TransformBean"))  {
			legalFlags.add(FLAG_X);
			legalFlags.add(FLAG_A);
			legalFlags.add(FLAG_NOEVENTS);
			legalFlags.add(FLAG_D);
			legalFlags.add(FLAG_L);
			commandLine.checkFlags(legalFlags);
			
			jasco.options.Options.addSinglePathToClassPathFirst(jasco.options.Options.getOutputDir());
			Vector excludedmethods = new Vector();
			if (X) 
			{
				String value = (String)flags.get(FLAG_X);
				String excludedmethodstring = value;
				excludedmethods = parseExcludedMethods(excludedmethodstring);
			}
			if(A)
				jasco.options.Options.setTransformAllMethods(true);
			
			if (flags.get(FLAG_L)!=null)  {
				try {
				
				File filelist = new File(file);
				Logger.getInstance().showOutput("START TRANSFORIMG, please wait....");
				BufferedReader reader = new BufferedReader(new FileReader(file));
				String line = reader.readLine();
				while (line!=null)  {
					
					Jasco.transformBeanInOutputDir(line,excludedmethods);
					line=reader.readLine();
				}
				reader.close();
				Logger.getInstance().showOutput("TRANSFORIMG DONE");
				}catch(Exception ex)  {Logger.getInstance().showError("Exception occured while accessing file list:" +file);
				}
			}
			else if (D)   {
				Logger.getInstance().showOutput("Searching all java beans in "+jasco.options.Options.getOutputDir()+"....");
				
				Vector v = getListOfBeansInDir(new File(file));
				Logger.getInstance().showOutput("START TRANSFORIMG, please wait....");
				Iterator i = v.iterator();
				while (i.hasNext())  {
					String next = (String) i.next();
					Jasco.transformBeanInOutputDir(next,excludedmethods);
				}
			}
			else {
			Jasco.transformBeanInOutputDir(file,excludedmethods);
			}
		}
		else if (command.equalsIgnoreCase("CompileTraversal"))  {
			commandLine.checkFlags(legalFlags);
			if (!checkFile(file))  {
				Logger.getInstance().showError( "Non-existing file: "+file);
				return;
			}
			Jasco.compileTraversal(file);
		}
		else System.err.println("Unknown command! "+command);
		
		System.exit(0);
	}
	
	protected static Vector getListOfBeansInDir(File dir)  {
		
		if (!dir.exists())  {
			Logger.getInstance().showError("File "+dir.getAbsolutePath()+" does not exist!");
			return new Vector();
		}
		if (!dir.isDirectory())  {
			Logger.getInstance().showError("File "+dir.getAbsolutePath()+" is not a directory!");
			return new Vector();
		}
			
 		Vector res= new Vector();
		String[] files = dir.list();
		for(int i=0; i<files.length; ++i)  {
			File f = new File(dir,files[i]);
			
			if(!f.isDirectory())
				continue;
			
			res.addAll(getListOfBeansInDirAndPackage(f,files[i]));
		}
		return res;					
	}
	protected static Vector getListOfBeansInDirAndPackage(File dir, String p)  {
	
		Vector res= new Vector();
		String[] files = dir.list();
		for(int i=0; i<files.length; ++i)  {
			File f = new File(dir,files[i]);
			if (f.isDirectory())  
				res.addAll(getListOfBeansInDirAndPackage(f,p+"."+files[i]));
			else  {
				String classname = files[i];
				if(!classname.endsWith(".class"))
					continue;
				if(classname.indexOf("$")!=-1)
					continue;	
				int j = classname.lastIndexOf('.');
				if(j!=-1)
					classname=classname.substring(0,j);
				String cClass = p+"."+classname;
				res.add(cClass);	
			}
		}
		return res;
	}
	
	protected static Vector parseExcludedMethods(String theMethods) 
	{	
		
		Vector output = new Vector();
		StringTokenizer t = new StringTokenizer(theMethods,jasco.options.Options.PATH_SEPARATOR);
		while (t.hasMoreTokens())
		{
			String res=t.nextToken();
			
			output.add(res);
		}
		return output;
	}
	
	protected static boolean checkFile(String f)  
	{		
		return new File(f).exists();
	}
			
	public static void printUsage(String command)  {
	
		System.err.println();
		String filename = "filename";
		if(command.equalsIgnoreCase("transformBean"))
			filename="fullclassname";
		System.err.println("Usage: "+command+" [flags] "+filename+" ");
		System.err.println();
		System.err.println("Flags:");
		System.err.println("-classpath:");
		System.err.println("-outputdir:");
		System.err.println("-X:");
		System.err.println("-version:");
		System.err.println("-target:");
		if (command.equalsIgnoreCase("transformBean"))  {
			System.err.println("-A");
			System.err.println("-D");
			System.err.println("-noevents");
			
		}
		System.err.println();
		System.err.println("More info: http://ssel.vub.ac.be/jasco/tools.php");
	}
		
	
	
	

	
}
