/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.runtime;

import gnu.regexp.RE;
import jasco.options.Options;
import jasco.runtime.connector.ISignatureMatcher;
import jasco.runtime.connector.MethodSignature;
import jasco.util.RegexpMatcher;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringReader;
import java.io.StringWriter;
import java.util.Iterator;
import java.util.Vector;

public class RuntimeContext  {

	private String[] cflow;
	private String calledMethod;
	private boolean isExecution;
	private MethodJoinpoint jascoMethod;
	
	public RuntimeContext(String calledMethod, MethodJoinpoint jascoMethod)  {
	
		this(calledMethod,jascoMethod,true);
	}
	
	public RuntimeContext(String calledMethod,  MethodJoinpoint jascoMethod, boolean isExecution)  {
		
			this.calledMethod=makeGNUReg(calledMethod);
			this.isExecution=isExecution;
			this.jascoMethod=jascoMethod;
	}
	
	public RuntimeContext(MethodJoinpoint jascoMethod)  {
		
			this(jascoMethod.getFullName(),jascoMethod,jascoMethod.isExecution()); 
	}
	
	public Object getTargetObject() {
		
		return jascoMethod.getCalledObject();
	} 
	
	public java.lang.annotation.Annotation[] getAnnotations() {
		
		return jascoMethod.getAnnotations();
	}
	
	public Class getStaticClass() {
		
		return jascoMethod.loadClass();
	}
	
	public MethodJoinpoint getJP() {
		
		return jascoMethod;
	}
	
	public boolean isCall() {
		
		return !isExecution;
	}
	
	public boolean isExecution() {
		
		return isExecution;
	}
	
	protected void initCflow()  {
		
		cflow=StackTraceConstructor.getInstance().getStackTrace();
	}
	
	protected String[] getCflow() {
		
		if(cflow==null) {
			initCflow();
		}
		return cflow;
	}
	
	public boolean containsMethodName(MethodSignature sig)  {
		
		String s = sig.getType()+"."+sig.getMethodname();
		for(int i=0; i<getCflow().length; ++i)
			if(RegexpMatcher.match(s,getCflow()[i]))
				return true;
				
		
		return false;	
	}
	
	
	public boolean containsMethodName(RE re)  {
		
		for(int i=0; i<getCflow().length; ++i)
			if(re.isMatch(getCflow()[i]))
				return true;
				
		
		return false;	
	}
	
	public boolean isDirectlyCalled(RE re)  {
		
		if(getCflow().length>1)
			return re.isMatch(getCflow()[1]);
		else return false;	
	}
	
	public boolean isDirectlyCalled(MethodSignature sig)  {
		
		if(getCflow().length>1) {
			
			String s = sig.getType()+"."+sig.getMethodname();
			return RegexpMatcher.match(s,getCflow()[1]);
		}
		else return false;	
	}
	
	public String getCalledMethod()  {
	
		return calledMethod;
	}
	
	public boolean isCalledMethod(RE methodName)  {
	
		return methodName.isMatch(calledMethod);
	}
	
	public void setCalledMethod(String methodName)  {
	
		calledMethod=makeGNUReg(methodName);
	}
	
	protected static boolean isRegularExp(String s)  {
	
		return s.indexOf('*') != -1;
	}
	
	public static String makeGNUReg(String s)  {
		
		return jasco.util.RegexpMatcher.makeGNUReg(s);
	}

	/**
	 * @param matcher
	 * @return
	 */
	public boolean containsMethodName(ISignatureMatcher matcher) {
		Iterator sigs = matcher.getSignatures();
		while(sigs.hasNext()) {
			
			MethodSignature sig = (MethodSignature) sigs.next();
			if(containsMethodName(sig))
				return true;
		}
		return false;
	}

	/**
	 * 
	 * @return
	 */
	public boolean isDirectlyCalled(ISignatureMatcher matcher) {
		
		Iterator sigs = matcher.getSignatures();
		while(sigs.hasNext()) {
			
			MethodSignature sig = (MethodSignature) sigs.next();
			if(isDirectlyCalled(sig))
				return true;
		}
		return false;
	}
	
	/**
	 * LDBN 19-7-2005. This method generates a array of strings with the stack contents
	 * */
	public static String[] getStackContents(){
		return StackTraceConstructor.getInstance().getStackTrace();
	}
		
}

abstract class StackTraceConstructor  {

	private static  StackTraceConstructor instance;
	
	static  {
	
		if (Options.isJavaAtLeastVersion(1,4))  
			instance=new DefaultStackTraceConstructor();
		else instance = new OldStackTraceConstructor();
	}
	
	public static void setInstance(StackTraceConstructor c)  {
	
		instance=c;
	}
	
	public static StackTraceConstructor getInstance()  {
	
		return instance;
	}
	
	abstract public String[] getStackTrace();
}

class DefaultStackTraceConstructor  extends StackTraceConstructor {
	
	public String[] getStackTrace()  {
		
		StackTraceElement[] elements = new Exception("boe").getStackTrace();
		//cflow = new String[elements.length];
		
		Vector temp = new Vector();
		for(int i=0; i<elements.length; ++i)  {
			//System.out.println(elements[i].getClassName()+"."+elements[i].getMethodName());
			if(elements[i].getClassName().startsWith("jasco"))
				continue;
			if(elements[i].getClassName().startsWith("sun.reflect"))
				continue;
			if(elements[i].getClassName().startsWith("java.lang.reflect"))
				continue;	
			if(elements[i].getMethodName().startsWith("_Jasco_"))
				continue;	
			//System.out.println(elements[i].getClassName()+"."+elements[i].getMethodName());
			//if(Options.isAspect(elements[i].getClassName()))
			//		continue;
			temp.add(RuntimeContext.makeGNUReg(elements[i].getClassName()+"."+elements[i].getMethodName()));
		}
		//System.out.println();
		String[] tmp = new String[temp.size()];
		temp.copyInto(tmp);
		return tmp;
	}
}
	
class OldStackTraceConstructor extends StackTraceConstructor {

	public String[] getStackTrace()  {
		
		try {
		Vector v = new Vector();
		StringWriter s = new StringWriter();
		PrintWriter w = new PrintWriter(s);
		new Throwable("boe").printStackTrace(w);
		
		BufferedReader reader = new BufferedReader(new StringReader(s.toString()));
		String tmp=reader.readLine();
		while (tmp!=null)  {
			
			if (tmp.startsWith("\t"))  {
				int index = tmp.indexOf("at");
				if (index!=-1)  {
					String res = tmp.substring(index+3);
					int index2=res.indexOf("(");
					if(index2!=-1)
						res=res.substring(0,index2);
					if (!res.startsWith("jasco")&&!(res.indexOf("_Jasco_")!=-1))  {
						v.addElement(RuntimeContext.makeGNUReg(res));
					
					}
				}
			}
			tmp=reader.readLine();
		}
	
		String[] tmpa = new String[v.size()];
		v.copyInto(tmpa);
		
		return tmpa;
		}catch (IOException ex)  {
			
		}
		return new String[0];
		//return new String[0];
	}
}