/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.runtime.aspect;

import jasco.runtime.JascoMethod;
import jasco.runtime.hotswap.JAsCoClassLoader;
import jasco.util.logging.Logger;
import javassist.CtClass;
import javassist.NotFoundException;

//import jasco.connectorparser.*;

public class ConstructorMethod  {

	private Class[] args;
	private Class returnType;
	private boolean multipleElements;
	private String[] argnames;

	public ConstructorMethod(Class returnType, Class[] args, String[] argnames, boolean multipleElements)  {
	
		this.returnType=returnType;
		this.args=args;
		this.argnames=argnames;
		this.multipleElements=multipleElements;
	}
	
	public boolean hasReturnType()  {
	
		return returnType!= null;
	}
	
	public Class getReturnType()  {
		
		return returnType;
	}
	
	public Class[] getArgTypes() {
		
		return args;
	}
	
	public String[] getArgNames() {
		
		return argnames;
	}
	
	public boolean hasMultipleElements()  {
	
		return multipleElements;
	}
	
	public boolean isCompatibleReturnJavassist(JascoMethod jp)  {

		if(!hasReturnType())
			return true;
		
		try {
		CtClass thisReturnType = JAsCoClassLoader.getSingleton().loadClass(returnType.getName());
		CtClass checkReturnType = jp.getJavassistMethod().getReturnType();
		if(thisReturnType.equals(checkReturnType))
			return true;
		
		return checkReturnType.subtypeOf(thisReturnType);
		}catch(NotFoundException ex) {Logger.getInstance().showError(ex);return false;}
	}
	
	public boolean isCompatibleReturn(Class checkReturnType)  {

		if(!hasReturnType())
			return true;
		
		if(checkReturnType.equals(returnType))
			return true;
			//System.err.println("checking"+checkReturnType.getName()+":"+returnType.getName()+";"+returnType.isAssignableFrom(checkReturnType));
		return returnType.isAssignableFrom(checkReturnType);
	}
	
	protected boolean isAlwaysCompatibleArg(int i) {
		
		if(i>=args.length&&hasMultipleElements())
			return true;
		else return false;
	} 
	
	protected boolean isPossibleCompatibleArg(int i) {
		
		if(i>=args.length&&!hasMultipleElements())	
			return false;
		else return true;
	}
	
	public boolean isCompatibleArg(Class cl,int i)  {
	 
		if(isAlwaysCompatibleArg(i))
			return true;
		else if(!isPossibleCompatibleArg(i))
			return false;
		else return args[i].isAssignableFrom(cl);
	}
	
	public boolean isCompatibleArgJavassist(CtClass cl,int i)  {
		 
			if(isAlwaysCompatibleArg(i))
				return true;
			else if(!isPossibleCompatibleArg(i))
				return false;
			else {
				try {
				CtClass clz = JAsCoClassLoader.getSingleton().loadClass(args[i].getName());
				return cl.subtypeOf(clz);
				}catch(NotFoundException ex) {Logger.getInstance().showError(ex);return false;}
			}
		}

	/**
	 * @param args2
	 * @return
	 */
	public boolean isCompatibleArgsJavassist(CtClass[] args2) {
		
		if(args2.length<args.length) //if less arguments than expected whoehoe, problems!!!!
			return false;
		
		for(int i=0;i<args2.length;++i)
			if(!isCompatibleArgJavassist(args2[i],i))
				return false;
		
		return true;
	}
	
	public boolean isCompatibleArgs(Class[] args2) {
		
		if(args2.length<args.length) //if less arguments than expected whoehoe, problems!!!!
			return false;
		
		for(int i=0;i<args2.length;++i)
			if(!isCompatibleArg(args2[i],i))
				return false;
		
		return true;
	}
}