/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.runtime.aspect;

public class HookContextProperty  {
	
	private boolean reflArgs = false;
	private boolean directArgsArray = false;
	private boolean directCalledObject = false;
	private boolean directCalledMethod = false;
	private boolean reflCalledObject = false;
	private boolean proceedWithDefinedArgs = false;
	
	public static final String GLOBAL = "global";
	
		/**
	 * @param reflArgs
	 * @param directArgsArray
	 * @param directCalledObject
	 * @param directCalledMethod
	 * @param reflCalledObject
	 * 
	 */
	public HookContextProperty(boolean reflArgs, boolean directArgsArray,
			boolean directCalledObject, boolean directCalledMethod,
			boolean reflCalledObject, boolean proceedWithDefinedArgs) {
		super();
		this.reflArgs = reflArgs;
		this.directArgsArray = directArgsArray;
		this.directCalledObject = directCalledObject;
		this.directCalledMethod = directCalledMethod;
		this.reflCalledObject = reflCalledObject;
		this.proceedWithDefinedArgs = proceedWithDefinedArgs;
	}
	
	/**
	 * @return Returns the directArgsArray.
	 */
	public boolean isDirectArgsArray() {
		return directArgsArray;
	}
	/**
	 * @return Returns the directCalledMethod.
	 */
	public boolean isDirectCalledMethod() {
		return directCalledMethod;
	}
	/**
	 * @return Returns the directCalledObject.
	 */
	public boolean isDirectCalledObject() {
		return directCalledObject;
	}
	/**
	 * @return Returns the onlyDefaultProceed.
	 */
	public boolean isProceedWithDefinedArgs() {
		return proceedWithDefinedArgs;
	}
	/**
	 * @return Returns the reflArgs.
	 */
	public boolean isReflArgs() {
		return reflArgs;
	}
	/**
	 * @return Returns the reflCalledObject.
	 */
	public boolean isReflCalledObject() {
		return reflCalledObject;
	}
	
	public static HookContextProperty intersection(HookContextProperty h1, HookContextProperty h2) {
		
		return new HookContextProperty(h1.reflArgs||h2.reflArgs,h1.directArgsArray||h2.directArgsArray,
				h1.directCalledObject||h2.directCalledObject,h1.directCalledMethod||h2.directCalledMethod,
				h1.reflCalledObject||h2.reflCalledObject, h1.proceedWithDefinedArgs||h2.proceedWithDefinedArgs);
	}
	
	public static HookContextProperty intersection(HookContextProperty[] props) {
		
		HookContextProperty result = new HookContextProperty(false,false,false,false,false,false);
		for(int i=0;i<props.length;++i) {
			HookContextProperty next = props[i];
			if(next==null) continue;
			if(next.reflArgs)
				result.reflArgs=true;
			if(next.directArgsArray)
				result.directArgsArray=true;
			if(next.directCalledObject)
				result.directCalledObject=true;
			if(next.directCalledMethod)
				result.directCalledMethod=true;
			if(next.reflCalledObject)
				result.reflCalledObject=true;
			if(next.proceedWithDefinedArgs)
				result.proceedWithDefinedArgs=true;
		}
		return result;
	}
	
	public String toSourceCode() {
		
		return "new jasco.runtime.aspect.HookContextProperty("+reflArgs+", "+directArgsArray+", "+directCalledObject+", "+
		directCalledMethod+", "+reflCalledObject+", "+proceedWithDefinedArgs+")";
	}
}
		 