package jasco.runtime.aspect.factory;

import jasco.util.logging.Logger;
import jasco.runtime.MethodJoinpoint;
import jasco.runtime.aspect.IHook;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;

public class PerBindingAspectFactory extends DefaultAspectFactory {

	private static final String TRIGGERING_METHOD_NAME = "_Jasco_getTriggeringBinding";

	public PerBindingAspectFactory() {
		super();
	}

	public IHook getAspectInstance(MethodJoinpoint jp,
			IAspectInstanceCreator creator) {
		if(jp==null)
			return null;
		Object key = getKey(jp,creator.getHookClass());
		IHook currentHook =  this.getAspectInstance(key,creator);
		return currentHook;
	}

	public Object getKey(MethodJoinpoint jp) {
		throw new InternalError("No longer used");
	}
	
	public Object getKey(MethodJoinpoint jp, Class<? extends IHook> hookclass) {
		try {
			Method getbinding = hookclass.getMethod(TRIGGERING_METHOD_NAME,MethodJoinpoint.class);
			Object binding = getbinding.invoke(null,jp);
			return binding;
		} catch(NoSuchMethodException nsme) {
			Logger.getInstance().showError(
				"Expected method "+TRIGGERING_METHOD_NAME+
				" not found in "+hookclass.getName()
			);
			return null;
		} catch(IllegalAccessException iae) {
			Logger.getInstance().showError(
				"No access to method "+TRIGGERING_METHOD_NAME+
				" in "+hookclass.getName()
			);
			return null;
		} catch(InvocationTargetException ite) {
			Logger.getInstance().showError(
				"Exception from "+TRIGGERING_METHOD_NAME+
				" in "+hookclass.getName()
			);
			Logger.getInstance().showFullError(ite);
			return null;
		}
	}
}