/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.runtime.connector;

import jasco.runtime.aspect.IHook;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

public class HookList 
{
	private Hashtable originalHooks = new Hashtable();
	private Vector hooks = new Vector();
	private boolean isChanged = false;
	
	public HookList() 
	{
	}
	
	public HookList(Vector aVector) 
	{	
			Enumeration i = aVector.elements();
			while (i.hasMoreElements())   {
				HookElement cp = (HookElement) i.nextElement();
				IHook c = cp.getCutpoint();
				originalHooks.put(c,cp);
				hooks.addElement(c);
			}
	}
	
	public void resetChanged() {
		
		if(!isChanged)
			return;
		hooks.clear();
		hooks.addAll(originalHooks.keySet());
		isChanged=false;
		
	}
	
	protected void setChanged() {
		
		isChanged=true;
	}
	
	public boolean contains(Object aHook) 
	{	

		return hooks.contains(aHook);	
	
	}
	
	public void remove(Object aHook)
	{	
		hooks.removeElement(aHook);
		setChanged();
	}
	
	public void add(Object aHook)
	{
		if(originalHooks.get(aHook)!=null)
			hooks.addElement(aHook);
		setChanged();
	}
	
	/**
	 * used by traversal connectors!
	 * @param aHook
	 */
	public void addImpl(Object aHook)
	{
		hooks.addElement(aHook);
		setChanged();
	}

	
	public Vector getHooks() 
	{
		return hooks;
	}
	
	public Vector getCutpointElements()  {
		
		Vector res = new Vector();
		Enumeration i = hooks.elements();
		while (i.hasMoreElements())   {
			IHook cp = (IHook) i.nextElement();
			HookElement element = (HookElement) originalHooks.get(cp);
			res.addElement(element);
		}
		return res;
			
			
	}

	/**
	 * @return
	 */
	public boolean isChanged() {
		
		return isChanged;
	}
	
	
}
	
	