/*
 * Created on 24-feb-2005
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package jasco.runtime.connector;

import jasco.runtime.JascoMethod;
import jasco.runtime.MethodJoinpoint;
import jasco.runtime.hotswap.JAsCoClassLoader;
import jasco.util.RegexpMatcher;
import jasco.util.logging.Logger;

import java.lang.annotation.Annotation;
import java.util.Iterator;
import java.util.Vector;

import javassist.CtClass;
import javassist.CtMethod;
import javassist.NotFoundException;

/* 
 Copyright (C) 2002-2005 JAsCo Development Team
 http://ssel.vub.ac.be/jasco
 
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as
 published by the Free Software Foundation; either version 2 of the
 License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 USA 
 */

/**
 * class that represents the method signature defined in a connector. Is bound to exactly one abstract 
 * method parameter of a hook's constructor.
 * 
 * See CombinedMethodSignature for multiple bindings.
 * 
 * 
 * @author Wim
 */

public class MethodSignature implements ISignatureMatcher {
 
	public final static int NO_SUBTYPES = 0;
	public final static int IMPL_SUBTYPES = 1;
	public final static int ALL_SUBTYPES = 2;
	
	private String type;
	private String returnType;
	private String methodname;
	private int matchSubTypes=NO_SUBTYPES;
	private Vector annotations = new Vector();
	private Vector arguments = new Vector();
	private String argumentsString = "";
	
	private Vector tempv = new Vector();
	
	public MethodSignature(String returnType, String type, String methodname) {
		super();
		
		this.type=type;
		this.returnType=returnType;
		this.methodname=methodname;
		
		tempv.add(this);
	} 
	
	public boolean matchesJoinpoint(MethodJoinpoint jp) {
		
		
		if(!RegexpMatcher.match(this.getReturnType(), jp.getReturnTypeName()))
			return false;
		
		
		if(!matchesMethod(jp,jp.getClassLoader()))
			return false;
		
		if(!matchesClass(jp.getClassName(),jp.getClassLoader()))
			return false;
		
		
		if(!RegexpMatcher.match(this.getVariableString(), this.getVariableString(jp)))
			return false;
		
		if(jp instanceof JascoMethod && ((JascoMethod) jp).dontLoadClass()) //solves javassist annotations bug
			return true;
		
		Iterator i = annotations.iterator();
		while(i.hasNext()) {
			String annotation  = i.next().toString();
			Annotation[] realAnnotations = jp.getAnnotations();
			boolean match = false;
			for(int j=0;j<realAnnotations.length;++j) {
				if(RegexpMatcher.match(annotation, realAnnotations[j].annotationType().getName())) {
					match=true;
					break;
				}
			}
			if(!match)
				return false;
		}
		
		return true;
	}
	
	
	/**
	 * @param jp
	 * @param classLoader
	 * @return
	 */
	private boolean matchesMethod(MethodJoinpoint jp, ClassLoader classLoader) {
		
		if(!RegexpMatcher.match(this.getMethodname(), jp.getName()))
			return false;
		
		switch(getMatchSubTypes()) {
			case IMPL_SUBTYPES:
				try {
				JAsCoClassLoader loader = JAsCoClassLoader.getSingleton();
				loader.appendPathsToClassPath(classLoader);
				
				CtClass superClass = loader.loadClass(this.getType());
				CtMethod[] methods = superClass.getMethods();
				for(int i=0;i<methods.length;++i)
					if(equals(jp,methods[i])) //implemented method
						return true;
					
				return false;
				}catch(Exception ex) {Logger.getInstance().showError(ex);return false;}
				
			case ALL_SUBTYPES:	
			case NO_SUBTYPES:
		    	return true;
			default:
				throw new IllegalArgumentException("illegal subtype matching id: "+getMatchSubTypes());
		}
	}
	
	protected boolean equals(MethodJoinpoint jp, CtMethod method) {
		
		if(!jp.getName().equals(method.getName()))
			return false;
		
		String[] jpargs = jp.getFormalArgumentTypeNames();
		CtClass[] margs;
		try {
			margs = method.getParameterTypes();
		} catch (NotFoundException e) {
			Logger.getInstance().showError(e);
			return false;
		}
		if(jpargs.length!=margs.length)
			return false;
		
		for(int i=0;i<jpargs.length;++i)
			if(!jpargs[i].equals(margs[i].getName()))
				return false;
			
		return true;
	}

	public boolean matchesClass(String clzName, ClassLoader classLoader) {
		
		if(!RegexpMatcher.match(this.getType(), clzName)) {
			switch(getMatchSubTypes()) {
			
			    case ALL_SUBTYPES:
			    case IMPL_SUBTYPES:
			    	try {
						JAsCoClassLoader loader = JAsCoClassLoader.getSingleton();
						loader.appendPathsToClassPath(classLoader);
						
						CtClass superClass = loader.loadClass(this.getType());
						CtClass subClass = loader.loadClass(clzName);
						
						if(subClass.subtypeOf(superClass))
							return true;
						else return false;
						
					}catch(Exception ex) {Logger.getInstance().showError(ex);return false;}
					
			    case NO_SUBTYPES:
			    	return false;
				default:
					throw new IllegalArgumentException("illegal subtype matching id: "+getMatchSubTypes());
			}
		}
		
		
		return true;
	}
	
	public boolean matchesClass(String clzName) {
		
		return matchesClass(clzName,ClassLoader.getSystemClassLoader());
	}
	
	public boolean matchesClass(String clzName, javassist.ClassPool pool) {
		
		if(!RegexpMatcher.match(this.getType(), clzName)) {
			switch(getMatchSubTypes()) {
				
			    case ALL_SUBTYPES:
			    case IMPL_SUBTYPES:
			    	try{
					CtClass superClass = pool.get(this.getType());
					CtClass subClass = pool.get(clzName);
					
					if(subClass.subtypeOf(superClass))
						return true;
					else return false;
			    	}catch(Exception ex) {Logger.getInstance().showError(ex);return false;}
				
		 		case NO_SUBTYPES:
		 			return false;
		 		default:
		 			throw new IllegalArgumentException("illegal subtype matching id: "+getMatchSubTypes());
			}
		}
		
		
		return true;
	}
	
	public String toString() {
		
		String result = "";
		
		Iterator i = annotations.iterator();
		while(i.hasNext()) {
			String annotation  = i.next().toString();
			result+="@"+annotation;
			result+=" ";
		}
		
		result+=this.getReturnType().toString()+" ";
		result+=this.getType()+".";
		result+=this.getMethodname();
		result+="(";
		result+=getVariableString();
		result+=")";
		
		return result;
	}
	
	public String getVariableString() {
		
		String result ="";
		
		Iterator i = arguments.iterator();
		while(i.hasNext()) {
			String s = i.next().toString();
			result+=s;
			if(i.hasNext())
				result+=", ";
		}
		
		return result;
	}
	
	public String getVariableString(MethodJoinpoint jp) {
		
		String result ="";
		
		String[] clz = jp.getFormalArgumentTypeNames();
		for(int i=0;i<clz.length;++i) {
			result+=clz[i];
			if(i<(clz.length-1))
				result+=", ";
		}

		return result;
	}
	
	public Iterator getAnnotations() {
		
		return annotations.iterator();
	}
	
	public Iterator getArguments() {
		
		return arguments.iterator();
	}
	
	protected void addAnnotation(String annotation) {
		
		annotations.add(annotation);
	}
	
	protected void addArgument(String argument) {
		
		arguments.add(argument);
	}
	

	/**
	 * @return Returns the matchSubTypes.
	 */
	public int getMatchSubTypes() {
		return matchSubTypes;
	}
	/**
	 * @param matchSubTypes The matchSubTypes to set.
	 */
	protected void setMatchSubTypes(int matchSubTypes) {
		this.matchSubTypes = matchSubTypes;
	}
	/**
	 * @return Returns the methodname.
	 */
	public String getMethodname() {
		return methodname;
	}
	/**
	 * @param methodname The methodname to set.
	 */
	protected void setMethodname(String methodname) {
		this.methodname = methodname;
	}
	/**
	 * @return Returns the returnType.
	 */
	public String getReturnType() {
		return returnType;
	}
	/**
	 * @param returnType The returnType to set.
	 */
	protected void setReturnType(String returnType) {
		this.returnType = returnType;
	}
	/**
	 * @return Returns the type.
	 */
	public String getType() {
		return type;
	}
	/**
	 * @param type The type to set.
	 */
	protected void setType(String type) {
		this.type = type;
	}
	
	
	public static MethodSignature build(String returnType, String type, String methodName, String[] arguments, String[] annotations, int matchSubtypes) {
		
		MethodSignature result = new MethodSignature(returnType,type,methodName);
		
		for(int i=0;i<arguments.length;++i)
			result.addArgument(arguments[i]);
		
		for(int i=0;i<annotations.length;++i)
			result.addAnnotation(annotations[i]);
		
		result.setMatchSubTypes(matchSubtypes);
		
		return result;
		
	}

	/* (non-Javadoc)
	 * @see jasco.runtime.connector.ICombinedMethodSignature#getSignatures()
	 */
	public Iterator getSignatures() {
		
		return tempv.iterator();
	}

}
