/**
 * 
 */
package jasco.runtime.distribution;

import jasco.runtime.CRChangeEvent;
import jasco.runtime.ConnectorRegistryListener;
import jasco.runtime.connector.Connector;
import jasco.runtime.distribution.annotation.Distribution;

import java.io.IOException;

import javassist.CannotCompileException;
import javassist.ClassPool;
import javassist.CtClass;
import javassist.NotFoundException;

/**
 * @author lbenavid
 * 
 */
public class DistributedConnectorListener implements ConnectorRegistryListener {

	/**
	 * 
	 */
	public DistributedConnectorListener() {
		super();
		// TODO Auto-generated constructor stub
	}

	/**
	 * This method will distribute a coonector, making the distributed load of
	 * Aspects
	 * 
	 * @see jasco.runtime.ConnectorRegistryListener#registryChanged(jasco.runtime.CRChangeEvent)
	 */
	public void registryChanged(CRChangeEvent event) {
		// The connector events are only distributed widely if the action of
		// adding
		// or deleting happens with a connector that was not received in a
		// distributed way.
		// Meaning that the deletion is executed locally if the connector was
		// received by distribution.
		if (DistributedConnectorHandler.getInstance().isLocal(
				(Connector) event.getParam())) {
			try {
				if (event.getType() == CRChangeEvent.CONNECTOR_ADDED) {
					Connector c = (Connector) event.getParam();
					if(!shouldNotify(c))
						return;
					ClassPool pool = ClassPool.getDefault();
					CtClass connClass = jasco.runtime.hotswap.JAsCoClassLoader
							.getSingleton().loadClass(
									event.getParam().getClass().getName());
					byte[] byteCode = connClass.toBytecode();
					connClass.defrost();

					DistributedConnectorHandler.getInstance().sendNotification(
							new DistributedConnectorMessage(byteCode,
									DistributedConnectorMessage.ADD_CONNECTOR));
				} else if (event.getType() == CRChangeEvent.CONNECTOR_REMOVED) {
					Connector c = (Connector) event.getParam();
					if(!shouldNotify(c))
						return;
					DistributedConnectorHandler
							.getInstance()
							.sendNotification(
									new DistributedConnectorMessage(
											event.getParam().getClass()
													.getName(),
											DistributedConnectorMessage.REMOVE_CONNECTOR));
				}

			} catch (NotFoundException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (CannotCompileException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
	}

	private boolean shouldNotify(Connector c) {
		Distribution distribution = c.getClass().getAnnotation(Distribution.class);
		if(distribution==null)
			return true;
		if(distribution.value().equals(Distribution.Type.LOCAL))
			return false; 
		else return true;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see jasco.runtime.ConnectorRegistryListener#registryUpdated()
	 */
	public void registryUpdated() {
		// TODO Auto-generated method stub

	}

}
