package jasco.runtime.distribution;

import jasco.runtime.distribution.hostselection.DefaultHostSelectionStrategy;
import jasco.runtime.distribution.hostselection.IHostSelectionStrategy;

import java.util.Iterator;
import java.util.Vector;

public class HostSet implements IHostSet, INeighborhood {
	
	private Vector<Host> hosts;
	private String name = "";
	private IHostSelectionStrategy strategy = null;
	
	HostSet(String name, IHostSelectionStrategy strategy) {
		this(new Vector<Host>(),name,strategy);
	}
	HostSet(Vector<Host> hosts, String name, IHostSelectionStrategy strategy) {
		super(); 
		this.hosts=hosts;
		this.name=name;
		this.strategy=strategy;
	}

	HostSet(Vector<Host> v) {
		this(v,"dummy",new DefaultHostSelectionStrategy());
	}
	public boolean containsLocalHost() {
		
		Iterator<Host> hi = hosts.iterator();
		while(hi.hasNext()) {
			Host next = hi.next();
			if(next.isLocalHost())
				return true;
		}
		
		return false;
	}
	
	public String toString() {
		
		return IHostSetBuilder.NEIGH+name;
	}

	public boolean isLocalHost() {
		Iterator<Host> hi = hosts.iterator();
		while(hi.hasNext()) {
			Host next = hi.next();
			if(!next.isLocalHost())
				return false;
		}
		
		return true;
	}

	public Host selectOneHostByStrategy() {
		
		if(hosts.size()>0)
			return hosts.firstElement();
		return null;
	}
	
	public IHostSet selectHostsByStrategy() {
		
		Vector<Host> hostsCopy= (Vector<Host>) hosts.clone();
		return new HostSet(strategy.selectHosts(hostsCopy));
	}
	
	public boolean contains(Host host) {
		
		Iterator<Host> hi = hosts.iterator();
		while(hi.hasNext()) {
			Host next = hi.next();
			if(next.equals(host))
				return true;
		}
		return false;
	}
	
	public void addHost(Host host) {
		this.hosts.add(host);
		
	}
	public void removeHost(Host host) {
		this.hosts.remove(host);
	} 
	public Iterator<Host> getHosts() {
		return hosts.iterator();
	}
	
}
