package jasco.runtime.distribution;

import jasco.options.Options;
import jasco.runtime.distribution.hostselection.DefaultHostSelectionStrategy;
import jasco.runtime.distribution.hostselection.IHostSelectionStrategy;

import java.util.Hashtable;
import java.util.Map;
import java.util.Vector;

public class IHostSetBuilder {
	
	public static final String NEIGH = "neigh:";
	
	private static final Vector<IHostSet> hostsets = new Vector<IHostSet>();
	private static final Map<String,Vector<Host> > hostscontents = new Hashtable<String,Vector<Host> >();
	 
	
	private IHostSetBuilder() {
		super();
	}
	
	public static IHostSet buildFrom(String desc,String strategyStr) {
		
		if(strategyStr==null)
			return buildFrom(desc,(IHostSelectionStrategy)null);
		
		IHostSelectionStrategy strategy = null;
		
		try {
		Class clz = Options.loadClass(strategyStr);
		
		strategy = (IHostSelectionStrategy) clz.newInstance();
		} catch (ClassNotFoundException e) {
			jasco.util.logging.Logger.getInstance().showError("Host selection strategy not found: "+strategyStr);
		} catch (InstantiationException e) {
			jasco.util.logging.Logger.getInstance().showError(e);
		} catch (IllegalAccessException e) {
			jasco.util.logging.Logger.getInstance().showError(e);
		}
		 
		return buildFrom(desc,strategy);
	}
	
	public static IHostSet buildFrom(String desc,IHostSelectionStrategy strategy) {
		
		if(strategy==null)
			strategy=new DefaultHostSelectionStrategy();
		
		if(desc.startsWith(NEIGH)) {
			desc=desc.substring(NEIGH.length());
			IHostSet hset = new HostSet(getSharedVector(desc),desc,strategy);
			return hset;
		}
		else return HostWrapper.buildFrom(desc);
	}
	
	private static Vector<Host> getSharedVector(String desc) {
		if(!hostscontents.containsKey(desc)) {
			Vector<Host> v = new Vector();
			hostscontents.put(desc,v);
		}
		return hostscontents.get(desc);
	}
	public static IHostSet buildFrom(String desc) {
		
		return buildFrom(desc,(IHostSelectionStrategy)null);
	}


}
