/*
 * Created on 24-mrt-2005
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package jasco.runtime.hotswap;

import jasco.options.Options;
import jasco.runtime.ConnectorRegistry;
import jasco.runtime.JascoMethod;
import jasco.runtime.MethodJoinpoint;
import jasco.runtime.annotation.NoJAsCoAOP;
import jasco.runtime.connector.Connector;
import jasco.runtime.distribution.DistributedJascoMethod;
import jasco.runtime.transform.TransformerConstants;
import jasco.util.JavassistHelper;
import jasco.util.logging.Logger;

import java.util.Set;
import java.util.TreeSet;
import java.util.Vector;

import javassist.CtClass;
import javassist.CtMethod;
import javassist.NotFoundException;

/* 
 Copyright (C) 2002-2005 JAsCo Development Team
 http://ssel.vub.ac.be/jasco
 
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as
 published by the Free Software Foundation; either version 2 of the
 License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 USA 
 */

/**
 * @author Wim
 */
public class AspectApplicationComputer {

	private static AspectApplicationComputer instance = new AspectApplicationComputer();
	
	private AspectApplicationComputer() {
		super();
	}

	public static AspectApplicationComputer getInstance () {
		
		return instance;
	} 
	
	public Vector computeApplicableHooks(JascoMethod jp) {
		
		Vector hooks = ConnectorRegistry.findCutpointForMethod(jp.getFullName(),null,jp,true);
		return hooks;
	}
	
	public Set getMethodsWhereHooksAreApplicable(CtClass clz, ClassLoader loader) {
		
		
		TreeSet result = new TreeSet();
		
		JascoMethod[] jps = computeJoinPoints(clz,loader);
	
		for(int i=0;i<jps.length;++i) {
			Vector hooks = ConnectorRegistry.findCutpointForMethod(jps[i].getFullName(),null,jps[i],true);
			if(hooks.size()>0)
				result.add(TransformerConstants.getFullName(jps[i].getJavassistMethod()));
		}
	
		return result;
	}
	
	public Set getMethodsWhereHooksAreApplicable(String classname) {
		
		CtClass clz;
		try {
			clz = JAsCoClassLoader.getSingleton().loadClass(classname);
		
			return getMethodsWhereHooksAreApplicable(clz,ClassLoader.getSystemClassLoader());
		} catch (NotFoundException e) {
			Logger.getInstance().showError(e);
		}
		
		return null;
	}
	
	public Set getMethodsWhereHooksAreApplicable(String classname, Connector theConnector) {
		
		TreeSet result = new TreeSet();
		
		try {
			CtClass clz = JAsCoClassLoader.getSingleton().loadClass(classname);
			JascoMethod[] jps = computeJoinPoints(clz,ClassLoader.getSystemClassLoader());
			for(int i=0;i<jps.length;++i) {
				Vector hooks = theConnector.findCutpointForMethod(null,jps[i],0,true);
				if(!hooks.isEmpty())
					result.add(TransformerConstants.getFullName(jps[i]));
				
			}
		} catch (NotFoundException e) {
			Logger.getInstance().showError(e);
		}
		
		return result;
	}

	/**
	 * @param clz
	 * @return
	 */
	public JascoMethod[] computeJoinPoints(CtClass clz,ClassLoader loader) {
		
		if(JavassistHelper.hasAnnotation(clz,NoJAsCoAOP.class)) //no AOP for this class
			return new JascoMethod[0];
		
		CtMethod[] methods = clz.getDeclaredMethods();
		Vector result = new Vector();
		try {
		for(int i=0;i<methods.length;++i)
			if(TransformerConstants.isPossibleJoinpoint(methods[i])) {
				MethodJoinpoint jp = JoinpointContainer.getInstance().getJoinpoint(methods[i]);
				if(jp==null) {
					if(Options.isDistributed()) 
						jp=DistributedJascoMethod.buildFrom(methods[i],loader,i);
					else jp=JascoMethod.buildFrom(methods[i],loader,i);
					JoinpointContainer.getInstance().addJoinpoint(methods[i],jp);
				}
				result.add(jp); 
			}
				
		
		} catch (NotFoundException e) {
			Logger.getInstance().showError(e);
		}
		
		JascoMethod[] jps = new JascoMethod[result.size()];
		result.copyInto(jps);
		
		return jps;
	}

	/**
	 * @param jps
	 */
	public void releaseDoNotLoadFlag(JascoMethod[] jps) {
		for(int i=0;i<jps.length;++i)
			jps[i].dontLoadClass(false);
		
	}

}
