/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.runtime.hotswap1;

import jasco.util.logging.Logger;

import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.util.Iterator;
import java.util.Vector;

public class VMNetwork  {
	
	private DatagramSocket sendSocket;
	private DatagramSocket socket;
	
	private int listenport;
	private InetAddress sendaddress;
	private int sendport;
	private ListenThread listenThread;
	private Vector listeners = new Vector();
	
	public VMNetwork(int listenport, InetAddress sendaddress, int sendport)  {
		System.out.println("kkk"+listenport);
		this.sendaddress=sendaddress;
		this.listenport=listenport;
		this.sendport=sendport;
		startListening();
	}
	
	public int getBufferLength() {
	
		return 1024;
	}
	
	public DatagramSocket getSocket() {
	
		return socket;
	}
	
	public boolean hasSocket() {
	
		return socket!=null;
	}
	
	public void startListening() {
	
		try {
		if (listenThread!=null)  {
			listenThread.stop();
			getSocket().close();
			socket=null;
		}
		socket = new DatagramSocket(listenport);
		listenThread = new ListenThread();
		listenThread.start();
		}
		catch(Exception ex) {Logger.getInstance().showError(ex);
		}
	}
	
	public void stopListening() {
		
		listenThread.stop();
		socket = null;
	}
	
	public void sendPacket(String message) {
	
		if(!hasSocket()) {
			startListening();
		}
		
		try {
		byte[] buf = message.getBytes();
		DatagramPacket packet = new DatagramPacket(buf, buf.length, sendaddress, sendport);
		System.out.println("sending packet"+sendport);
		socket.send(packet);
		}
		catch(Exception ex) {
			
			Logger.getInstance().showError(ex);
		}
	}

	public void packetReceived(DatagramPacket packet) {
		
		System.out.println("received");
		String message = new String(packet.getData(),packet.getOffset(),packet.getLength());
		notifyReceive(message);
	}
		
	class ListenThread extends Thread {
		
		public 	ListenThread()  {
		
			setDaemon(true);
		}
	
		public void run() {
			
			while(true) {
				try{
				System.out.println("waiting for packet on"+listenport);
				byte[] buf = new byte[getBufferLength()];
				DatagramPacket packet = new DatagramPacket(buf, buf.length);
				getSocket().receive(packet);
				packetReceived(packet);
				System.out.println(" packet received on"+listenport);
				}
				catch(Exception ex) {System.err.println(ex.getMessage());
				ex.printStackTrace();
				}
			
			}
		}
	}
	
	public void addNetworkListener(NetworkListener l)  {
	
		listeners.add(l);
	}
	
	public void removeNetworkListener(NetworkListener l)  {
	
		listeners.remove(l);
	}
	
	protected void notifyReceive(String packet)  {
		
		Iterator i = listeners.iterator();
		while (i.hasNext())  {
			NetworkListener listener = (NetworkListener) i.next();
			listener.receive(packet);
		}
	}
}