/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.runtime.inline;

import jasco.options.Options;
import jasco.runtime.ConnectorRegistry;
import jasco.runtime.JascoMethod;
import jasco.runtime.aspect.IHook;
import jasco.runtime.transform.TransformerConstants;
import jasco.util.logging.Logger;

import java.lang.annotation.Annotation;
import java.util.Hashtable;
import java.util.Vector;

import javassist.CtClass;

public class InlineCompiler  {
	
	private static InlineCompiler inline = new InlineCompiler();
	private boolean enabled = false;
	private boolean jumpingaspects = false;
	private Hashtable adaptedMethods = new Hashtable();
	private InlineHeuristics doInlineMethod = new DefaultInlineHeuristics();
	private long logid = 0;
	
	public static InlineCompiler getDefault()  {
	
		return inline;
	} 
	
	private InlineCompiler()  {
	
		String tmp = System.getProperty("jasco.hotswap.inlinecompiler");
		if(tmp!=null&&!tmp.equals(""))
			enabled = new Boolean(tmp).booleanValue()&&!Options.useReflectionForProceed()&&!Options.isDistributed();
		
		if (isEnabled())  {
			System.out.println("JAsCo prototype run-time weaver enabled!");
		}
		else System.out.println("JAsCo prototype run-time weaver disabled!");
		
		
		tmp = System.getProperty("jasco.hotswap.inlinecompiler.heuristics");
		if (tmp!=null&&!tmp.equals(""))  {
			try  {
			Class inlineH = Class.forName(tmp);
			doInlineMethod=(InlineHeuristics)inlineH.newInstance();	
			} catch(Exception ex)  {
				Logger.getInstance().showError("Failed instantiating inline heuristics: "+tmp);
			}
		}
		
		tmp = System.getProperty("jasco.hotswap.inlinecompiler.jumpingaspects");
		if(tmp!=null&&!tmp.equals(""))
			jumpingaspects = new Boolean(tmp).booleanValue();
	}
	
	private boolean isJumpingAspectsEnabled() {
		
		return jumpingaspects;
	}
	
	public boolean isJumpingAspectsEnabled(IHook cp) {
		
		if(!isJumpingAspectsEnabled())
			return false;
		
		Annotation[] annots = cp.getClass().getAnnotations();
		for(int i=0;i<annots.length;++i)
			if(annots[i].annotationType().getName().equals(JumpingAspectsWeaver.WEAVE_ALL_ANNOT))
				return false;
			
		return true;
	}
	
	public Hashtable getInlineCompiledMethods()  {
		
		return adaptedMethods;
	}
	
	public boolean isEnabled()  {
	
		return enabled&&jasco.options.Options.hotswap1_5Enabled();
	}
	
	
	public boolean inlineCompile(CtClass theclass, Vector hooks, JascoMethod method, InlineJutta ijutta, boolean schemaChages)  {
	
		if(!isEnabled())
			return false;
		
		if(!doInlineMethod.inlineCompile(method,hooks))
			return false;
		
		if(!ConnectorRegistry.areConnectorsCachable())
			return false;
		
		//Logger.getInstance().showDebug("inline COMPILED: "+method.getFullName()+"---"+inlineCompiled(method));	
		
		if(inlineCompiled(method)) { //already inline compiled, though waiting to be inserted because code was running
			Logger.getInstance().showError("ALREADY inline compiled: "+method.getSignature());
			return false; //do not inline compile	
		}
		
		try {
		
		if(Options.isJAsCoBean(theclass)) //no statcially transformedf beans...
			return false;

		Logger.getInstance().showDebug("inline compiling: "+method.getFullName());	
		
		if(schemaChages)
			PerInstanceOptimizer.getInstance().processClass(theclass,hooks);
		ijutta.process(method,hooks,ClassLoader.getSystemClassLoader());
	

		addInlineCompiledMethod(method);
		
		method.setInlineCompiled(true);
		
		return true;
		
		}catch(InlineCompilingFailed ex)  {
			Logger.getInstance().showError(ex.getMessage());
		}
		catch(Exception ex)  {
			Logger.getInstance().showFullError(ex);
		}
		
		return false;
	}
	
	public boolean inlineCompiled(JascoMethod method)  {
		
		//test not necessary any more, never comes here twice
		return false;
		/*
		String clzname = method.getClassName();
		
		Vector v = (Vector) adaptedMethods.get(clzname);
		if(v==null)
			return false;
		else return v.contains(TransformerConstants.getFullName(method));*/
	} 
	
	public void addInlineCompiledMethod(JascoMethod method)  {
		
		String clzname = method.getClassName();
		String methodname = TransformerConstants.getFullName(method);
		
		Vector v = (Vector) adaptedMethods.get(clzname);
		if (v==null)  {
			v=new Vector();
			adaptedMethods.put(clzname,v);
		}
		v.add(methodname);
	}
	
	public void removeInlineCompiledMethods(String clzname, Vector methodnames)  {
	
		Vector v = (Vector) adaptedMethods.get(clzname);
		if (v==null)
			return;
		
		v.removeAll(methodnames);
	}
	
	public void removeInlineCompiledMethods(String clzname)  {
	
		adaptedMethods.remove(clzname);
	}
	

	
		
}
			
	
	