/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 
  USA
 */
package jasco.runtime.inline;

import jasco.options.Options;
import jasco.runtime.JascoMethod;
import jasco.runtime.MethodJoinpoint;
import jasco.runtime.aspect.Dummy;
import jasco.runtime.aspect.HookContextProperty;
import jasco.runtime.aspect.HookHelper;
import jasco.runtime.aspect.IHook;
import jasco.runtime.aspect.PCutpointConstructorRegular;
import jasco.runtime.aspect.factory.IPredefinedAspectFactory;
import jasco.runtime.cache.CachedAspectRep;
import jasco.runtime.cache.CombinedHooksFactory;
import jasco.runtime.connector.DynamicHookElement;
import jasco.runtime.connector.HookElement;
import jasco.runtime.hotswap.JAsCoClassLoader;
import jasco.runtime.hotswap.JoinpointContainer;
import jasco.runtime.transform.CallbackGenerator;
import jasco.runtime.transform.ClassProcessor;
import jasco.runtime.transform.FieldGenerator;
import jasco.runtime.transform.FieldGeneratorForInlineJutta;
import jasco.runtime.transform.MethodProcessor;
import jasco.runtime.transform.TransformerConstants;
import jasco.runtime.transform.TrapInserter;
import jasco.tools.aspectparser.AspectClassesGenerator;
import jasco.util.generators.JavaGenerator;
import jasco.util.generators.MethodGenerator;
import jasco.util.logging.Logger;

import java.io.File;
import java.io.FileWriter;
import java.lang.annotation.Annotation;
import java.util.Arrays;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Vector;

import javassist.CtClass;
import javassist.CtField;
import javassist.CtMethod;
import javassist.CtNewMethod;
import javassist.Modifier;
import javassist.NotFoundException;
import javassist.bytecode.CodeAttribute;

public class InlineJutta extends CombinedHooksFactory  {

	private static CtClass RUNTIME_EXCEPTION;
	private long id = getNextClassId();
	private CtClass targetClass;
	private CtMethod theMethod;
	private JascoMethod jascoMethod;
	private CtClass fieldClass;
	
	private boolean inlineReplaceCompilingPossible = false;
	private boolean outputIsObject=true;
	
	private boolean callbackExists = false;
	private boolean callbackGenerated = false;
	private FieldGenerator lastFieldGenerator = null;
	private Hashtable fieldGeneratorInits = new Hashtable();
	private Hashtable fieldGeneratorInitsJP = new Hashtable();
	
	static {
		
		try {
			RUNTIME_EXCEPTION=JAsCoClassLoader.getSingleton().loadClass(RuntimeException.class.getName());
		} catch (NotFoundException e) {
			Logger.getInstance().showError(e);
		}
	}
	
	public InlineJutta(CtClass clazz)  {
	
		targetClass=clazz;
		fieldClass = JAsCoClassLoader.getSingleton().makeClass(getFieldClassName());
		fieldClass.setModifiers(Modifier.setPublic(fieldClass.getModifiers()));
	}
	
	protected void callbackGenerated() {
		
		this.callbackGenerated=true;
	}
	
	protected boolean isCallbackGenerated() {
		
		return callbackGenerated;
	}
	
	protected String accessField(String field)  {
	
		return getFieldClassName()+"#"+field; 
	}
	
	protected String getFieldClassName()  {
	
		return targetClass.getName()+"$"+"_JAsCo_JuttaFieldClass"+id;
	}
	
	protected String getSimpleFieldClassName()  {
	
		return "_JAsCo_JuttaFieldClass"+id;
	}
	
	protected CtMethod getOrigMethod()  {
	
		return theMethod;
	}
	
	protected JascoMethod getJascoMethod()  {
	
		return jascoMethod;
	}
	
	public MethodJoinpoint getJoinpoint() {
		
		return getJascoMethod();
	}
	
	public void initVars(JascoMethod originalMethod) throws Exception {
		
		CtMethod origMethod = initCtMethod(originalMethod);
		this.theMethod=origMethod;
		this.jascoMethod=originalMethod;
		inlineReplaceCompilingPossible=false;
		outputIsObject=true;
		callbackExists=false; 
	}
	
	public void process(JascoMethod originalMethod, Vector orighooks, ClassLoader loader) throws InlineCompilingFailed, Exception {
		
		
		
		Vector hooks=(Vector)orighooks.clone();
		check15hooks(hooks);
		CachedAspectRep rep = new CachedAspectRep(getFieldClassName());
		
		initVars(originalMethod);
		 
		generateConstructorCode(getFieldClassName(),hooks,originalMethod,fieldClass,rep);	
		if(!generateInvokeCode(hooks,originalMethod,fieldClass,rep))
			return;
				
		MethodGenerator constr = rep.getConstructorMethod();
		constr.addModifier(java.lang.reflect.Modifier.STATIC);
		constr.addModifier(java.lang.reflect.Modifier.PUBLIC);
		String myInitName = "init"+originalMethod.getID();
		constr.setName(myInitName);
		constr.setReturnType("void");
						
		CtMethod init = CtNewMethod.make(constr.toString(),fieldClass);
		fieldClass.addMethod(init);
		
		if(Options.showDebugOutput()) {
			File out = new File(Options.getLogDir()+File.separator+"juttainline_constructor"+id+".txt");
			FileWriter w = new FileWriter(out);
			w.write("//"+originalMethod.getSignature()+JavaGenerator.NEWLINE+constr.toString());
			w.close();
		}
		
		String invokeCode = rep.getInvokeMethod().getImplementation();
		invokeCode=rep.getInvokeInitCode()+invokeCode;
		
		Vector methods = new Vector();
		methods.add(getOrigMethod());
		
		MethodProcessor processor = new TrapInserter(classPool);
		processor.setFieldClass(fieldClass);
		processor.setProcessMethods(methods);
		
		//disabled this for now, still testing
		//FieldGenerator fgenerator = new FieldGeneratorForCallbackPerMethod(processor);
		FieldGeneratorForInlineJutta fgenerator = new FieldGeneratorForInlineJutta(processor,callbackExists());
		InlineInserter inserter = new InlineInserter(processor,invokeCode);
		
		if(!isCallbackGenerated()&&callbackExists()) { //only generate this once anymore
			CallbackGenerator cgenerator = new CallbackGenerator(processor);
			cgenerator.setProcessMethods(getCallbackProcessMethods());
			cgenerator.process(targetClass,originalMethod.getClassName());
			cgenerator.generateClass(Options.getTempDir());
			callbackGenerated();
		}
		
		fgenerator.addInitCode(FieldGeneratorPropertyStore.getInstance().generateFInitCode(fieldClass.getName(),myInitName));
		
		fgenerator.process(targetClass,originalMethod.getClassName());
		try {
		inserter.process(targetClass,originalMethod.getClassName());
		saveInlineCode(orighooks,getOrigMethod());
		}catch(Exception ex)  {
			Logger.getInstance().showError(ex);
		}
		
		lastFieldGenerator=fgenerator;
		
		FieldGeneratorPropertyStore.getInstance().addFieldGeneratorInit(fgenerator,myInitName,hooks,originalMethod);
	}
	
	/**
	 * @return
	 */
	private Vector getCallbackProcessMethods() {
		Vector res  = new Vector();
		CtClass target = getOrigMethod().getDeclaringClass();
		CtMethod[] methods = target.getDeclaredMethods();
		for(int i=0;i<methods.length;++i)
			if(JoinpointContainer.getInstance().getJoinpoint(methods[i])!=null) //is possible jp
				res.add(methods[i]);
			
		return res;
	}

	public void finalizeWeaving() {
		
		if(lastFieldGenerator==null)
			return;
		
		try {
			
		lastFieldGenerator.generateClass(Options.getTempDir());
		}catch(Throwable ex) {
			Logger.getInstance().showFullError(ex);
		}
	}
	
	/**
	 * @return
	 */
	private boolean callbackExists() {
		return callbackExists;
	}
	
	private void callbackExists(boolean b) {
		callbackExists=b;
	}

	/**
	 * @param hooks
	 */
	private void check15hooks(Vector hooks) throws InlineCompilingFailed {
		
		Iterator i = hooks.iterator();
		while(i.hasNext()) {
			HookElement cp = (HookElement) i.next();
			Annotation[] annotations = cp.getCutpoint().getClass().getAnnotations();
			boolean found =false;
			for(int j=0;j<annotations.length;++j)
				if(annotations[j].annotationType().getName().equals(AspectClassesGenerator.ANNOTATION_ANNOTATION)) {
					found=true;break;}
			if(!found)
				throw new InlineCompilingFailed(cp.getCutpoint().getClass().getName()+": Inline compiler only works when aspects are compiled for Java 1.5 (use target flag)");
		}
	}

	protected void computeInit(Vector replaceVector, Vector genericReplaceVector, HookContextProperty requiredHooksContext, boolean testReplaces) {
		
		inlineReplaceCompilingPossible = inlineReplaceCompilingPossible(replaceVector,requiredHooksContext,testReplaces);
		
		callbackExists(!genericReplaceVector.isEmpty()&&!inlineReplaceCompilingPossible);
		
		//System.out.println(genericReplaceVector.size()+";"+replaceVector.size());
		//System.out.println(genericReplaceVector);
		//System.out.println(replaceVector);
		if(genericReplaceVector.size()!=replaceVector.size()) //around throwing/returning defined! output type is object!!!!
			outputIsObject = true;
		else outputIsObject=!inlineReplaceCompilingPossible; 
	}
	
	protected void saveInlineCode(Vector hooks, CtMethod method) {
		
		
		
		//TODO: to make work with several hooks, Vector of hooks and Vector of desginators (state Ids) to
		// retrieve code
		if(hooks.size()!=1) //jumping aspects only work for 1 hook per jp for the moment...
			return;
		
		CodeAttribute code = method.getMethodInfo().getCodeAttribute();
		
		HookElement cpElement = (HookElement) hooks.elementAt(0);
		IHook cp = cpElement.getCutpoint();
		if(!InlineCompiler.getDefault().isJumpingAspectsEnabled(cp))
			return;
		
		if(cp._Jasco_getConstructorRep() instanceof PCutpointConstructorRegular) {//jumping aspect
			PCutpointConstructorRegular reg = (PCutpointConstructorRegular) cp._Jasco_getConstructorRep();
				
			InlineCodeCache.getInstance().addInlineCode(method,code);
			
		}
	}
	
	protected String getFieldClassVarName() {
		
		return "INSTANCE"+this.getJascoMethod().getID();
	}
	
	protected void addFieldClassInstance(CtClass fieldClass) throws Exception {
		
		addFieldImpl("public static final "+getFieldClassName()+" "+getFieldClassVarName()+"  =new "+getFieldClassName()+"();",fieldClass);
	}
	
	private boolean inlineCompileGenerateReplaceSucceeded=false;
	
	protected String generateReplaceSequence(HookContextProperty requiredHooksContext, Vector replaceVector, boolean testReplaces, CtClass fieldClass, boolean outputDefined) throws Exception {
		
		inlineCompileGenerateReplaceSucceeded=false;
		
		if(!inlineReplaceCompilingPossible(replaceVector,requiredHooksContext,testReplaces))
			return super.generateReplaceSequence(requiredHooksContext,replaceVector,testReplaces,fieldClass,outputDefined);
		
		String code=null;
		Vector accessHooks = new Vector();
		addFieldClassInstance(fieldClass);
		
		for(int i=0;i<replaceVector.size();++i)  {
			HookElement cpElem = (HookElement) replaceVector.elementAt(i);
			String accessHook =accessFieldHook(cpElem);
			accessHooks.add(accessHook);
		}
		
		CtMethod themethod = getOrigMethod();
		boolean isStatic = Modifier.isStatic(themethod.getModifiers());
		
		int startForArgs = 1;
		if(requiredHooksContext.isDirectCalledObject())
			startForArgs=2;
		String args = generateAdviceArgs(IHook.REPLACE,replaceVector,testReplaces);
		
		InlineReplaceGenerator generator = new InlineReplaceGenerator(fieldClass, accessHooks, replaceVector, classPool, themethod, this.getJascoMethod(), id, args, startForArgs);
		generator.start();
		
		HookElement cpElem = (HookElement) replaceVector.elementAt(0);
		String accessHook =accessFieldHook(cpElem);
		
		code=getFieldClassName()+"#"+getFieldClassVarName()+"."+InlineReplaceGenerator.getNewMethodName(IHook.REPLACE,cpElem.getCutpoint(),id,this.getJascoMethod())+"("+args+");"+JavaGenerator.NEWLINE;
			
		
		if(!themethod.getReturnType().equals(CtClass.voidType)) {
			String tmp = "";
			if(!outputDefined)
				tmp=themethod.getReturnType().getName()+" ";
			tmp+="output=";
			code=tmp+code;
		}
		
		inlineCompileGenerateReplaceSucceeded=true;	
		
		return code;
	}
	
	
	
	//TODO: update this test...
	protected boolean inlineReplaceCompilingPossible(Vector hooks, HookContextProperty requiredHooksContext, boolean testReplaces) {
		
		if(hooks.size()==0)
			return true;
		

		//if(true) return true; //FIXME: implement this method properly!
		if(testReplaces)
			return false;
		else if(requiredHooksContext.isProceedWithDefinedArgs())
			return false;
		else if(!allHooksHaveSameArgs(hooks,IHook.REPLACE))
			return false;
		else if(!hooksExactlyDefineArgs(hooks,IHook.REPLACE))
			return false;
		else return true;
	}
	
	protected String getThisInstance() {
		
		String thisInstance="$0";
		if(Modifier.isStatic(getOrigMethod().getModifiers()))
			thisInstance="jasco.runtime.NULL#instance";
		return thisInstance;
	}
	
	protected String generateInvokeInitCode(Vector replaceVector, HookContextProperty requiredHooksContext, boolean noReplace, boolean testReplaces) throws Exception {
		
		String code = "";
		
		boolean inlineReplacePossible = noReplace||inlineReplaceCompilingPossible(replaceVector,requiredHooksContext,testReplaces);
		
		boolean requiresDefaultAroundArgs = requiresDefaultAround(replaceVector,testReplaces);
		
		if(requiredHooksContext.isDirectArgsArray()||requiredHooksContext.isReflArgs()||!inlineReplacePossible||requiresDefaultAroundArgs)
			code+="Object[] args = $args;"+JavaGenerator.NEWLINE; 
		if(requiredHooksContext.isReflArgs()||!inlineReplacePossible)
			code+="method.setArgumentsArray(args);"+JavaGenerator.NEWLINE;
		if(requiredHooksContext.isReflCalledObject()||requiredHooksContext.isProceedWithDefinedArgs()||!inlineReplacePossible) {
			code+="method.setCalledObject("+getThisInstance()+");"+JavaGenerator.NEWLINE;
		}
		
		/*
		if(Modifier.isStatic(getOrigMethod().getModifiers()))
			code=TransformerConstants.getClallBackPerMethodObject(getOrigMethod())+".initialise($$);"+JavaGenerator.NEWLINE;
		else code=TransformerConstants.getClallBackPerMethodObject(getOrigMethod())+".initialise("+thisInstance+",$$);"+JavaGenerator.NEWLINE;
		*/
		
		return code;
	}
	
	protected HookContextProperty computeAdviceContext(String adviceType, Vector advices) {
		
		HookContextProperty[] combinedHooksForAdvice = new HookContextProperty[advices.size()];
		Iterator i = advices.iterator();
		int j = 0;
		while(i.hasNext()) {
			HookElement hook = (HookElement) i.next();
			HookContextProperty prop = hook.getCutpoint()._Jasco_getHookContextPropertyForAdvice(adviceType);
			combinedHooksForAdvice[j]=prop;
		}
		
		return HookContextProperty.intersection(combinedHooksForAdvice);
	}
	
	public String generateAdviceArgs(String adviceType, Vector advices,boolean testReplaces) {
		
		HookContextProperty requiredContext = computeAdviceContext(adviceType,advices);
		HookElement firstHook = (HookElement) advices.elementAt(0);
		
		if(adviceType.startsWith(IHook.REPLACE))
			return generateReplaceAdviceArgs(requiredContext,adviceType,advices,testReplaces); //TODO: implement replace
		else return generateNonReplaceAdviceArgs(requiredContext,adviceType,firstHook.getCutpoint());
	}
	
	/**
	 * prereq: all hooks have the same args and return type!!!
	 * 
	 * checks whether hooks have the same return type and expected jp arguments as the jp method itself
	 * 
	 * @param hooks
	 * @param adviceType
	 * @return
	 */
	// 
	protected boolean hooksExactlyDefineArgs(Vector hooks, String adviceType) {
		
		try {
		//		ceck whether all hooks have the same arguments...:
		Iterator hooksI = hooks.iterator();
		boolean allHooksTheSame = true;
		HookElement hook = (HookElement) hooksI.next();
			
		CtMethod advicemethod = findReflectMethod(hook.getCutpoint(),adviceType);
		if(advicemethod==null) //default method for throwing/returning advices
			return false;
			
		CtClass[] paramtypes = advicemethod.getParameterTypes();
		CtClass[] paramsJP = getOrigMethod().getParameterTypes();
		
		int startForArgs = 2;
		
		
		if(paramtypes.length-startForArgs!=paramsJP.length)
			return false;
		
		
		for(int i=startForArgs,j=0;j<paramsJP.length&&i<paramtypes.length;++i,++j)
			if(!paramtypes[i].getName().equals(paramsJP[j].getName()))
				return false;
		
			
		if(advicemethod.getReturnType().getName().equals( getOrigMethod().getReturnType().getName()))
			return true;
		else return false;
		} catch(Exception ex) {Logger.getInstance().showError(ex);}
		
		return false;
	}
	
	/**
	 * Checks whether all hooks have the same expected jp args and jp return type
	 * @param hooks
	 * @param adviceType
	 * @return
	 */
	protected boolean allHooksHaveSameArgs(Vector hooks, String adviceType) {
		
		try {
		//		ceck whether all hooks have the same arguments...:
		Iterator hooksI = hooks.iterator();
		CtClass[] args = null;
		CtClass rtType = null;
		boolean allHooksTheSame = true;
		while(hooksI.hasNext()) {
			HookElement hook = (HookElement) hooksI.next();
			
			CtMethod advicemethod = findReflectMethod(hook.getCutpoint(),adviceType);
			if(advicemethod==null) //args used, no use for inline generation...
				return false;
			
			if(args!=null) {
				allHooksTheSame=allHooksTheSame&&Arrays.equals(args,advicemethod.getParameterTypes());
				allHooksTheSame=allHooksTheSame&&advicemethod.getReturnType().equals(rtType);
			}
			else {
				args=advicemethod.getParameterTypes();
				rtType=advicemethod.getReturnType();
			}
				
		} 
		
		return allHooksTheSame;
		
		}catch(Exception ex) {
			Logger.getInstance().showError(ex);
		}
		
		return false;
		
		
		
	}
	
	/**
	 * @param cutpoint
	 * @param adviceType
	 * @return
	 */
	private CtMethod findReflectMethod(IHook cutpoint, String adviceType) {
		return findReflectMethod(cutpoint,adviceType,true);
	}

	protected boolean requiresDefaultAround(Vector replaceVector, boolean testReplaces) {
		
		return testReplaces&&replaceVector.size()>1;
	}
	
	
	//watch out: around throwing and around returning also here, proceed must work there as well! (idea BART V)
	protected String generateReplaceAdviceArgs(HookContextProperty requiredContext, String adviceType, Vector hooks,boolean testReplaces) {
		
		if(requiresDefaultAround(hooks,testReplaces))
			return generateDefaultAdviceArgs(adviceType,hooks,testReplaces);
		
		HookElement firstHook = (HookElement) hooks.elementAt(0);
		
		HookContextProperty prop = new HookContextProperty(requiredContext.isReflArgs(),requiredContext.isDirectArgsArray(),
				true, true, requiredContext.isReflCalledObject(), requiredContext.isProceedWithDefinedArgs());
		
		String s= generateNonReplaceAdviceArgs(prop,adviceType,firstHook.getCutpoint());
		/* when args at the end for optimzied around
		try {
		CtClass[] fargs = theMethod.getParameterTypes();
		s+=", "+getThisInstance();
		for(int i=1;i<=fargs.length;++i)
			s+=", $"+i;
		}catch (javassist.NotFoundException ex) {jasco.util.logging.Logger.getInstance().showError(ex);}
		System.out.println(s);
		*/
		return s;
	}
	
	/**
	 * @param adviceType
	 * @param hooks
	 * @param testReplaces
	 * @return
	 */
	private String generateDefaultAdviceArgs(String adviceType, Vector hooks, boolean testReplaces) {
		return "method, "+getThisInstance()+", args";
	}

	/*
	protected String generateCastAdviceArgReturning(String casted, String uncasted) {
		
		Class expectedType = this.getJascoMethod().getReturnType();
		if(expectedType.equals(Void.TYPE))
			return super.generateCastAdviceArgReturning(casted,uncasted);
		else {
			String code = expectedType.getName()+" "+casted+"  = ";
			if(expectedType.isPrimitive())
				code+="jasco.runtime.inline.InlineHelper#castTo_int("+uncasted+")";
			else code+="("+expectedType.getName()+")";
			code+=";";
			return code;
		}
	}*/
	
	protected String castAdviceArgs(String adviceType, String resultName) {
		//NOTE: output is always object with returning/throwing
		if(adviceType.equals(IHook.AFTER_RETURNING)||adviceType.equals(IHook.AROUND_RETURNING)) {
			CtClass expectedType;
			try {
				expectedType = this.getOrigMethod().getReturnType();
				
				if(expectedType.equals(CtClass.voidType))
					return super.castAdviceArgs(adviceType,resultName);
				else if(expectedType.isPrimitive()&&adviceType.equals(IHook.AROUND_RETURNING))
					return AspectClassesGenerator.makePrimitive(expectedType.getName(),resultName);
				else return "("+expectedType.getName()+")"+resultName;
			} catch (NotFoundException e) {
				Logger.getInstance().showError(e);
			}
			
		}
		
		return super.castAdviceArgs(adviceType,resultName);
			
	}
	
	
	protected String generateNonReplaceAdviceArgs(HookContextProperty requiredContext, String adviceType, IHook hook) {
		
		CtMethod adviceMethod  = findReflectMethod(hook,adviceType);
		CtClass[] args = null;
		try {
		
		if(adviceMethod!=null)
			args = adviceMethod.getParameterTypes();
		}
		catch(Exception ex) {
			Logger.getInstance().showError(ex);
		}
		if(adviceMethod==null||args==null) { //no advice found, can happen fe with after returning: 
			String res= "method, ";
			res+=getThisInstance()+", ";
			if(requiredContext.isDirectArgsArray())
				res+="args";
			else res+="jasco.runtime.NULL#instanceArray";
			
			return res;
		}
		
		String argCode="";
		
		int start = 0;
		if(requiredContext.isDirectCalledMethod()) {
			argCode+="method";
			start++;
		}
		if(requiredContext.isDirectCalledObject()) {
			if(start>0)
				argCode+=", ";			
			argCode+=getThisInstance();
			start++;
		}
		if(adviceType.equals(IHook.AFTER_RETURNING)||adviceType.equals(IHook.AFTER_THROWING)||adviceType.equals(IHook.AROUND_THROWING)||adviceType.equals(IHook.AROUND_RETURNING)) 
			start++; //additional argument to ignore..
		for(int i=start,j=1;i<args.length-1;++i,++j) { //untill length-1, because ends with dummy!
			
			if(i>0)
				argCode+=", ";	
			//if this var is the array with arguments, this is always last before dummy
			if(i==args.length-2&&args[i].getName().equals("java.lang.Object[]")&&requiredContext.isDirectArgsArray()) {
				
				argCode+="args";
			}
			else {
				 
				argCode+="$"+j; //access argument of original method...
			}
		}
		
		if(argCode.endsWith(", ")||argCode.endsWith(","))
			;
		else if(!argCode.equals("")) argCode+=", ";
		argCode+=Dummy.class.getName()+"#INSTANCE";
		
		return argCode;
	}
	

	/**
	 * finds the most specific advice method, 
	 * 
	 * returns null if not existing method
	 * 
	 * @param hook
	 * @param adviceType
	 * @return
	 */
	protected CtMethod findReflectMethod(IHook hook, String adviceType, boolean strictSelection) {
		
		try {
		CtClass clz = JAsCoClassLoader.getSingleton().loadClass(hook.getClass().getName());
		CtMethod[] methods = clz.getMethods();
		for(int i=0;i<methods.length;++i)
			if(methods[i].getName().equals(adviceType)) { //method found, check whether not default
				
				if(HookHelper.hasAdviceAnnotation(methods[i])) {
					if(adviceType.equals(IHook.AFTER_RETURNING)||adviceType.equals(IHook.AROUND_RETURNING)) {
						
						CtClass expectedType = this.getOrigMethod().getReturnType();
						CtClass firstParam = (methods[i].getParameterTypes()[0]);
						
						if(expectedType.subtypeOf(firstParam))//if type of return is assignable to the first arg of this method
							return methods[i];
						else if(!strictSelection&&firstParam.subtypeOf(expectedType))
							return methods[i];
						/* no auto boxing/unboxing supported yet :-(
						else if(expectedType.isPrimitive()&&firstParam.getName().equals("java.lang.Object")) //1.5 valid
							return methods[i];
						else if(firstParam.isPrimitive()&&expectedType.getName().equals("java.lang.Object")) //1.5 valid
							return methods[i];
						*/
							
					}
					else if(adviceType.equals(IHook.AFTER_THROWING)||adviceType.equals(IHook.AROUND_THROWING)) {
						
						CtClass expectedException = methods[i].getParameterTypes()[0];
						if(expectedException.subtypeOf(RUNTIME_EXCEPTION)||RUNTIME_EXCEPTION.subtypeOf(expectedException)) //runtime exception, always capture
							return methods[i];
						else {
							//only if the expected typed exception is declared it is possible that it is thrown -> so capture it only then
							// BUG!!! SUPER/SUB EXCPETION
							CtClass[] definedExceptions = this.getOrigMethod().getExceptionTypes();
							for(int j=0;j<definedExceptions.length;++j)
								if(definedExceptions[j].subtypeOf(expectedException))
									return methods[i];
						}
					}
					else return methods[i];
				}
			}
		}
		catch(Exception ex) {
			Logger.getInstance().showError(ex);
		}
		return null;
	}
	
	//TODO: find commomn ancestor!!!!!s
	protected String computeThrowingExceptionType(Vector hooks, String adviceType) {
		
		CtClass exceptionType = null;
		for (int i = 0; i < hooks.size() ; ++i) {
			HookElement cutpoint = (HookElement)hooks.elementAt(i);
			int id = cutpoint.getId();
			
			if(!HookHelper.definesAdvice(cutpoint.getCutpoint(),adviceType))
				continue;
			
			CtMethod method = findReflectMethod(cutpoint.getCutpoint(),adviceType);
			if(method==null) {
				return super.computeThrowingExceptionType(hooks,adviceType); 
			}
			
			try {
				CtClass clz =  method.getParameterTypes()[0];
				if(exceptionType==null)
					exceptionType=clz;
				else return super.computeThrowingExceptionType(hooks,adviceType);
			} catch (NotFoundException e) {
				Logger.getInstance().showError(e);
			}
			
		}	
		
		if(exceptionType==null)
			return super.computeThrowingExceptionType(hooks,adviceType);
		else return exceptionType.getName();
	}
	
	
	
	public String defineOutput(JascoMethod newMethod, HookContextProperty requiredHooksContext) throws Exception {
		
		if(outputIsObject)
			return super.defineOutput(newMethod,requiredHooksContext);
		
		CtMethod themethod = getOrigMethod();
		if (!themethod.getReturnType().equals(CtClass.voidType))  {
			return themethod.getReturnType().getName()+" output;";
		} 
		else return null;
	}
	
	protected String generateOriginalMethodInvokeNoReplace(JascoMethod method, HookContextProperty requiredHooksContext, boolean outputDefined) throws Exception {
		
		CtMethod themethod = getOrigMethod();
		
		String code="";
		String newname = TransformerConstants.getNewMethodName(themethod);
		
		
		if (!themethod.getReturnType().equals(CtClass.voidType))  {
			if(!outputDefined)
				if(outputIsObject)
					code+="Object"+ " ";
				else code+=themethod.getReturnType().getName()+ " ";
			code+="output = ";
			String invokeCode = newname+"($$)";
			if(themethod.getReturnType().isPrimitive()&&outputIsObject)
				code+=AspectClassesGenerator.makePrimitiveObject(themethod.getReturnType().getName(),invokeCode);
			else code+=invokeCode;
			
		} else code+=newname+"($$)";
		
		return code;
	}
	
	
	protected String doReturn(String var, boolean noReplace, HookContextProperty requiredHooksContext) throws Exception {
		
		CtMethod themethod = getOrigMethod();
			
		if (!outputIsObject)  {
			if(themethod.getReturnType()!=CtClass.voidType) 
				return "return output;";
			else return "";
		}	
		
		String startCode="";
		String returnCode="";
		String code="";
		if (themethod.getReturnType()!=CtClass.voidType)    {
			returnCode+="return ";
			
			if (themethod.getReturnType().isPrimitive())  {
				String type = ClassProcessor.makePrimitiveObject(themethod.getReturnType());
				startCode+=type+" res = "+"("+type+")";
				returnCode+="res."+themethod.getReturnType().getName()+"Value();";
			}
			else  {
				String type = themethod.getReturnType().getName();
				startCode+=type+" res = "+" ("+type+") ";
				returnCode+=" res;";
			}
		
			code+=startCode;
			code+=var+";"+JavaGenerator.NEWLINE;
			code+=returnCode;
		}
		else code="return;"+JavaGenerator.NEWLINE;
		
	
		
		return code;
	}
	
	public boolean requiresInvocationThrowingReturning(String adviceType, IHook theHook) {
		
		boolean b = super.requiresInvocationThrowingReturning(adviceType,theHook);
		if(!b)
			return false;
		
		CtMethod theMethod = this.findReflectMethod(theHook,adviceType,false);
		return theMethod!=null;
	}
	
	
	protected CtMethod initCtMethod(JascoMethod method) throws Exception {
		
		CtMethod res = method.getJavassistMethod();
		return res;
	}
	
	protected void makeFieldsStaticPublic(CtClass clazz) throws Exception {
	
		CtField[] fields = clazz.getDeclaredFields();
		
		for(int i=0;i<fields.length;++i)  {
			CtField field = fields[i];
			int modifiers = field.getModifiers();
			modifiers=Modifier.setPublic(modifiers);
			modifiers|= Modifier.STATIC;
			field.setModifiers(modifiers);
		}
	}
	
	protected void addField(String f, CtClass clz) throws Exception {
		
		addFieldImpl("public static "+f,clz); 
	}

	
	protected String moreInvokeCode(IHook hook, String accessHook, String conditionboolean) {
		
		if(!(hook._Jasco_getConstructorRep() instanceof PCutpointConstructorRegular))
			return super.moreInvokeCode(hook,accessHook, conditionboolean);
		
		if(!InlineCompiler.getDefault().isJumpingAspectsEnabled(hook))
			return super.moreInvokeCode(hook,accessHook,conditionboolean); 
		
	
		String code = super.moreInvokeCode(hook,accessHook,conditionboolean);
		code+="if("+conditionboolean+")"+"jasco.runtime.inline.JumpingAspectsWeaver#getInstance().reweave("+accessHook+", method);";

		return code;
	}
	
	protected String generateFactoryInit(DynamicHookElement element, CtClass fieldClass, int index) throws Exception {
		
		if(!element.doNotCache())
			return "";
		
		String code = super.generateFactoryInit(element,fieldClass,index);
		
		String hname = getHookElementName(element);
		
		if(element.getAspectFactory() instanceof IPredefinedAspectFactory) {
			String facname = getFactoryName(element);
			addField(IPredefinedAspectFactory.class.getName()+" "+facname+";",fieldClass);
			code+=accessField(facname)+"="+"("+IPredefinedAspectFactory.class.getName()+")"+accessField(hname)+".getAspectFactory();"+JavaGenerator.NEWLINE;	
		}
		
		return code;
		
	}
	
	protected String generateFactoryAspectsForDynamicHookElements(DynamicHookElement el, String assignStartCode) {
	
		String code = "";
		if(! el.doNotCache())
			return "";
		
		String type = el.getCutpoint().getClass().getName();
		String hookElementField = accessField(getHookElementName(el));
		if((el.getAspectFactory() instanceof IPredefinedAspectFactory)) { //optimzie away!
			IPredefinedAspectFactory fac = (IPredefinedAspectFactory) el.getAspectFactory();
			String factoryField = accessField(this.getFactoryName(el));
			
			String fetchKey = fac.getKeyFetchSourceCode("method",getJoinpoint());
			String facInvokeCode =  "("+type+") "+factoryField+".getAspectInstance("+fetchKey+","+hookElementField+");"+JavaGenerator.NEWLINE;
			if(PerInstanceOptimizer.getInstance().isOptmized(el)) //no hookElements and factories
				code+=PerInstanceOptimizer.getInstance().generateOptimziedFactoryInit(el,facInvokeCode,assignStartCode,this.getJoinpoint());
			else code+=assignStartCode+facInvokeCode;
		}
		else code+=super.generateFactoryAspectsForDynamicHookElements(el,assignStartCode);
		
		return code;
	}
}
