/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.runtime.transform;

import jasco.options.Options;
import jasco.runtime.hotswap.JAsCoClassLoader;
import jasco.util.generators.ClassGenerator;
import jasco.util.generators.JavaGenerator;
import jasco.util.generators.MethodGenerator;

import java.io.File;
import java.io.FileOutputStream;
import java.util.Iterator;

import javassist.CtClass;
import javassist.CtConstructor;
import javassist.CtMethod;
import javassist.CtNewConstructor;
import javassist.CtNewMethod;
import javassist.Modifier;

public class CallbackGenerator extends MethodProcessor  {
	
	private StringBuffer codeBuffer = new StringBuffer();
	private boolean noElse = true;
	private ClassGenerator callbackClass = null;
	private CtClass generatedClass = null;
	
	public CallbackGenerator(ClassProcessor g)  {
	 
		super(g);
	}
	
	protected CtClass toJavaAssistInner(ClassGenerator classGen) throws Exception {
		
		String fullClassName = classGen.getPackageName()+"."+classGen.getName();
		if(classGen.getPackageName().equals(""))
			fullClassName=classGen.getName();
		CtClass cachedAspect = getClassPool().makeClass(fullClassName);
		CtConstructor c = CtNewConstructor.defaultConstructor(cachedAspect);
		cachedAspect.addConstructor(c);
		Iterator i = classGen.getMethods();
		while (i.hasNext())  {
			MethodGenerator g = (MethodGenerator) i.next();
			if (g.getReturnType().equals(""))  { //constructor
				CtConstructor method = CtNewConstructor.make(g.toString(),cachedAspect);
				cachedAspect.addConstructor(method);
			}
			else  {
			//	System.out.println("NOCON"+g.toString());
			CtMethod method = CtNewMethod.make(g.toString(),cachedAspect);
			cachedAspect.addMethod(method);
			}
		}
		
		i = classGen.getInterfaces();
		while (i.hasNext())  {
			String g = (String) i.next();
			CtClass cl = getClassPool().get(g);
			cachedAspect.addInterface(cl);
		}
		return cachedAspect;
	}
	
	public void generateClass(File dir) throws Exception {
		
		String classFileName = dir.getPath()+File.separator+getTargetClass().getName().replace('.','/')+"\\$"+callbackClass.getName();
		String classFileNameWithoutEscape = dir.getPath()+File.separator+getTargetClass().getName().replace('.','/')+"$"+callbackClass.getName();
		
		
		
		if(Options.isWindows())
			classFileName = classFileNameWithoutEscape;
		String className = getTargetClass().getSimpleName()+"$"+callbackClass.getName();
		callbackClass.setName(className);
		callbackClass.setPackage(getTargetClass().getPackageName());
		callbackClass.setModifiers(0);
		
		Iterator j = callbackClass.getConstructors();
		while (j.hasNext())  {
			MethodGenerator gen = (MethodGenerator) j.next();
			gen.setName(className);
		}
		
		CtClass inner = toJavaAssistInner(callbackClass);
		
		CtClass serializable = JAsCoClassLoader.getSingleton().loadClass("java.io.Serializable");
		inner.addInterface(serializable);
		
		addNoAspects(inner); 
		
		generatedClass=inner;
		makeDir(dir,inner);
		
		File classFile = new File(classFileNameWithoutEscape+".class");
		FileOutputStream writer = new FileOutputStream(classFile);
		inner.stopPruning(true);
		writer.write(inner.toBytecode());
		writer.close();
		inner.defrost();
		
		GeneratedFileManager.registerClassFile(classFile);
	}
	
	public ClassGenerator getResultingCallback()  {
	
		return callbackClass;
	}
	
	public String getResultingCallbackName() {
		
		return callbackClass.getPackageName()+"."+callbackClass.getName();
	}
	
	public void generateCallbackClassField(CtClass clz) throws Exception {
		
			//String thefield = "private static final jasco.runtime.JAsCoBeanCallback INSTANCE = new "+TransformerConstants.getCallbackClassFullname(getTargetClass(),getUniqueClassID())+"();"+JavaGenerator.NEWLINE;
			String thefield = "public static jasco.runtime.JAsCoBeanCallback INSTANCE;"; // new "+TransformerConstants.getCallbackClassFullname(getTargetClass(),getUniqueClassID())+"();"+JavaGenerator.NEWLINE;
			addField(thefield,clz);
		}
	
	public boolean startProcessing(CtClass clz, String classname) throws Exception {
		
		noElse=true;
		String name = TransformerConstants.getCallbackClass(getUniqueClassID());
		callbackClass = new ClassGenerator(name,"");
		callbackClass.addModifier(Modifier.PRIVATE);
		callbackClass.addModifier(Modifier.STATIC);
		callbackClass.addInterface("jasco.runtime.JAsCoBeanCallback");
		
		MethodGenerator method = new MethodGenerator("dispatch","Object");
		method.addModifier(Modifier.PUBLIC);	
		method.addException("java.lang.Exception");
		
		method.addVariable("int","i");
		method.addVariable("java.lang.Object","objectA");
		method.addVariable("java.lang.Object[]","v");
		
		String fullName= TransformerConstants.getCallbackClassFullname(clz,getUniqueClassID());
		
		String code	="";
		code+=getSourceName(clz)+" object = ("+getSourceName(clz)+") objectA;"+JavaGenerator.NEWLINE;
		codeBuffer.append(code);
		
		//START generating methods
		super.startProcessing(clz,classname);
		//END generating methods
		
		code = "";
		code+="return null;"+JavaGenerator.NEWLINE;
		codeBuffer.append(code);
		code=codeBuffer.toString();
		

		method.setImplementation(code);
		callbackClass.addMethod(method);
	
		return true; 
	}

	
	public boolean processMethod(CtMethod method, int methodid) throws Exception {
		
		Long id = new Long(methodid);
		String code = "";
		CtMethod oldMethod=getNewMethod(method);
		
		if (noElse)  {
			code+="if(i=="+id.toString()+")"+"{"+JavaGenerator.NEWLINE;
			noElse=false;
		}
		else code+="else if(i=="+id.toString()+")"+"{"+JavaGenerator.NEWLINE;
		String runCode = "";
		String endCode="";
		if(oldMethod.getReturnType()!=CtClass.voidType) {
			runCode+="return ";
			if(oldMethod.getReturnType().isPrimitive())	{
				runCode+="new "+makePrimitiveObject(oldMethod.getReturnType())+"(";
				endCode=")";
			}
		}
		boolean isStatic = Modifier.isStatic(oldMethod.getModifiers());
		if(isStatic)
			runCode+=getTargetClass().getName()+"."+oldMethod.getName()+"(";
		else runCode+="object."+oldMethod.getName()+"(";	
		
		CtClass[] params = oldMethod.getParameterTypes();
		for(int i=0;i<params.length;++i) {
			String src = getSourceName(params[i]);
			if(params[i].isPrimitive()) 
				code+=src+" var"+i+"= (("+makePrimitiveObject(params[i])+") v["+i+"])."+params[i].getName()+"Value();"+JavaGenerator.NEWLINE;
			else code+=src+" var"+i+"= ("+src+") v["+i+"];"+JavaGenerator.NEWLINE;
			runCode+="var"+i;
			if(i<params.length-1)
				runCode+=", ";
		}
			
		runCode+=")"+endCode+";"+JavaGenerator.NEWLINE;
		code+=runCode;
		if (oldMethod.getReturnType()==CtClass.voidType)  {
			code+="return null;"+JavaGenerator.NEWLINE;
		}
		code+="}"+JavaGenerator.NEWLINE+JavaGenerator.NEWLINE;
		
		codeBuffer.append(code);
		
		return true;
	}
	
}
		