/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.runtime.transform;

import jasco.util.logging.Logger;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileWriter;
import java.util.Vector;

import javassist.ClassPool;
import javassist.CtClass;
import javassist.CtField;
import javassist.CtMethod;
import javassist.NotFoundException;

public abstract class ClassProcessor  {

	private ClassPool myClassPool;
	private CtClass theClass;
	private CtClass fieldClass;
	private static CtClass NO_ASPECTS;
	
	private static long idsupplier = 0;
	
	private long theID;
	
	private FileWriter debugWriter;
	
	
	public ClassProcessor(ClassPool pool)  {
		this.myClassPool=pool;
		theID=idsupplier++;
		try {
		NO_ASPECTS=pool.get(jasco.runtime.NoAspects.class.getName());
		}catch (NotFoundException ex) {
			Logger.getInstance().showError(ex);
			NO_ASPECTS=null;
		}
	}
	
	public ClassProcessor(ClassProcessor prev)  {
	
		this(prev.getClassPool());
		theID=prev.getUniqueClassID();
		setFieldClass(prev.getFieldClass());
	}
	
	public long getUniqueClassID()  {
	
		return theID;
	}
	
	protected Vector debug()  {
		
		Vector res = new Vector();
		//System.out.println("hotswapped: "+getTargetClass().getName());
		CtMethod[] methods = getTargetClass().getDeclaredMethods();
		for(int i=0;i<methods.length;++i)
			res.add(methods[i].getName());
		
		return res;	
	}		
	
	protected void makeDir(File dir,CtClass clazz)  {
	
		String classname = clazz.getName();
		String classFileName=classname.replace('.','/');
		File classFile = new File(dir.getPath()+File.separator + classFileName + ".class");
		new File(classFile.getParent()).mkdirs();	
	}
	
	public void addNoAspects(CtClass clz) {
		
		clz.addInterface(NO_ASPECTS);
	}
	
	public ClassProcessor()  {
		
		this(ClassPool.getDefault());
	}
	
	public ClassPool getClassPool()  {
	
		return myClassPool;
	}
	
	public void setClassPool(ClassPool pool)  {
	
		this.myClassPool=pool;
	}
	
	public CtClass getTargetClass()  {
	
		return theClass;
	}
	
	public String getNewMethodName(CtMethod method)  {
		return TransformerConstants.getNewMethodName(method.getName(),method.getDeclaringClass().getSimpleName());
	}
	
	protected void addField(String fields, CtClass cl) throws Exception {
		
		//System.out.println(fields);
			
			
		CtField field = CtField.make(fields,cl);
		cl.addField(field);	
	}
	
	protected String getFieldClassName(CtClass main)  {
		
		return main.getName()+"$_JAsCo_FIELDS"+theID;
	}
	
	public void setUseFieldClass(CtClass clz, boolean b)  {
		
		if (b)  {
			String name = getFieldClassName(clz);
			/*
			try  {
				fieldClass = getClassPool().get(name); //already exists...
			}catch(Exception ex)  {
			*/
			fieldClass = getClassPool().makeClass(name);
			
		}
		else fieldClass=null;
	}
	
	public CtClass getFieldClass()  {
	
		return fieldClass;
	}
	
	public void setFieldClass(CtClass cl)  {
	
		this.fieldClass=cl;
	}
	
	protected String accessField(String f)  {
		
		String name = getTargetClass().getName();
		if(getFieldClass()!=null)
			name=getFieldClass().getName();
			
		return name+"#"+f;
	}
	
	public static String makePrimitiveObject(CtClass aClass) 
	{
		if(aClass==CtClass.booleanType)
			return "Boolean";
		else if(aClass==CtClass.intType)
			return "Integer";
		else if(aClass==CtClass.floatType)
			return "Float";
		else if(aClass==CtClass.doubleType)
			return "Double";
		else if(aClass==CtClass.charType)
			return "Character";
		else if(aClass==CtClass.byteType)
			return "Byte";
		else if(aClass==CtClass.shortType)
			return "Short";
		else if(aClass==CtClass.longType)
			return "Long";
		else return"Not recognized Primitive!!";
		
	}
	
	public CtMethod getNewMethod(CtMethod method)  {
	
		String newname = getNewMethodName(method);
		
		try  {
		return getTargetClass().getDeclaredMethod(newname,method.getParameterTypes());
		}
		catch(Exception ex)  {
			Logger.getInstance().showFullError(ex);
		}
		
		return null;
	}
	
	public String getSignatureSourceName(CtClass aClass)  {
		
		String name = aClass.getName();
		//name=name.replace('$','.');
		return name;
		
	}
	
	protected String getSourceName(CtClass aClass) throws Exception {
		
		String name = getSignatureSourceName(aClass);
		return name;
	}
	
	protected void setTargetClass(CtClass clz)  {
		
		clz.defrost();
		this.theClass=clz;
	}
	
	public byte[] getTargetByteCode() throws Exception {
	
		if(getTargetClass()==null)
			return null;
		getTargetClass().stopPruning(true);
		byte[] codes = getTargetClass().toBytecode();
	
		getTargetClass().defrost();
		return codes;
	}
	
	protected abstract boolean startProcessing(CtClass clz, String classname) throws Exception;
	
	public final boolean process(byte[] bytes, String classname) throws Exception {
		
		return process(getClassPool().makeClass(new ByteArrayInputStream(bytes)), classname);
	}
	
	public final boolean process(CtClass clz, String classname) throws Exception {
		
		if(clz.isInterface()) {
			return false;
		}
		
		setTargetClass(clz);
		
		return startProcessing(clz,classname);
	}
	
	
	
}