/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.runtime.transform;

import jasco.runtime.hotswap.JoinpointContainer;
import javassist.CtConstructor;
import javassist.CtMethod;

public class FieldGeneratorForInlineJutta extends FieldGenerator  {
	
	private boolean callbackExists = false;
	
	public FieldGeneratorForInlineJutta(ClassProcessor g)  {
		
		this(g,true);
	}
	
	public FieldGeneratorForInlineJutta(ClassProcessor g, boolean callbackExists)  {
	
		super(g,null);
		
		this.callbackExists=callbackExists;
	}
	
	protected boolean callbackExists() {
		
		return callbackExists;
	}
	
	public boolean processMethod(CtMethod method, int methodid) throws Exception {
		
		
		Long id = new Long(methodid);
		
		if(callbackExists())
			addInitCode("_Jasco_JascoMethod"+id.toString()+".setCallbackObject("+generateCallbackObject(method)+");");
		
		String jascoMethodField = "public static final jasco.runtime.JascoMethod _Jasco_JascoMethod"+id.toString()+" ";
		jascoMethodField+=" = (jasco.runtime.JascoMethod)"+JoinpointContainer.class.getName()+"#getInstance().getJoinpoint(\""+JoinpointContainer.getInstance().getUniqueId(method)+"\");";
	
		this.addField(jascoMethodField,getFieldClass());
		 
		return true;
	}
	
	//no use here, already inited
	public void initClassLoader(ClassLoader loader) throws Exception{
		
		
	}

	
	protected void generateStaticInit() throws Exception{
		
		String code = getInitCode();
		CtConstructor c = getFieldClass().makeClassInitializer();
		c.insertAfter(code);
	}
	
	
	public void generateCallbackClassField() throws Exception {
	
		if(callbackExists()) super.generateCallbackClassField();
	} 
}
