/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
/*
 * Created on 20-okt-2004
 */
package jasco.testing;

import jasco.options.Options;
import jasco.util.logging.Logger;

import java.io.File;
import java.io.InputStream;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Vector;

import junit.framework.Test;
import junit.framework.TestResult;

/**
 * @author Wim
 */ 

//TODO add following tests: run system with all compiled connectors, adapt runJuggler to add/remove deleter connector
public class JAsCoTestMain implements Test {
	
	private static Vector tests = new Vector();
	private static HashMap depends = new HashMap();
	
	private static boolean redoTests = false;
	private static Vector doneTests = new Vector();
	
	private static final String pacoSuiteDirString = "jasco.testing.pacosuitedir";
	private static final String benchBaseResultString = "jasco.testing.benchbaseresult";
	
	private static String pacoSuiteDir = null;
	private static int benchBaseResult = RunBenchTest.DONT_CHECK;
	
	public static Test suite() {
		emptyFolder(new File(DefaultJAsCoTest.getOutputDir()),false);
		redoTests = false;
		initOptions();
		initTests();
		
		return new JAsCoTestMain();
	}
	
	public static void initOptions() {
		
		String tmp = System.getProperty(pacoSuiteDirString);
		if(tmp!=null&&!tmp.equals("")) {
			File f = new File(tmp);
			if(!f.exists())
				Logger.getInstance().showWarning("Illegal directory for PacoSuite");
			else pacoSuiteDir=tmp;
		}
		
		tmp = System.getProperty(benchBaseResultString);
		
		if(tmp!=null&&!tmp.equals("")) {
			try {
				int tmpint  = new Integer(tmp).intValue();
				benchBaseResult = tmpint;
			}
			catch(NumberFormatException ex) {
				Logger.getInstance().showWarning("Illegal bench base result");
			}
		}
	}
	
	public int countTestCases() {
		int count= 0;
		
		Iterator i = tests.iterator();
		while(i.hasNext()) {
			Test test= (Test)i.next();
			count += test.countTestCases();
		}
		Vector counted = new Vector();
		counted.addAll(tests);
		i = depends.entrySet().iterator();
		while(i.hasNext()) {
			Map.Entry entry = (Map.Entry)i.next();
			Vector v = (Vector) entry.getValue();
			Iterator j = v.iterator();
			while(j.hasNext()) {
				Test test = (Test)j.next();
				if(!counted.contains(test)) { //not already counted
					count += test.countTestCases();
					counted.add(test);
				}
			}
		}
		return count;
	}
	
	public void run(TestResult result) {
		Iterator i = tests.iterator();
		while(i.hasNext()) {
			if (result.shouldStop())
				break;
			executeJunit((IJAsCoTest)i.next(),result,false);
		}
	}

	protected void executeJunit(IJAsCoTest test, TestResult result, boolean alwaysExecute) {
		
		if(!alwaysExecute&&test.succeededPreviously()&&!redoTests)
			return;
		
		if(doneTests.contains(test))
			return;
		
		Iterator depends = getDepends(test);
		while(depends.hasNext()) {
			IJAsCoTest dependsTest = (IJAsCoTest) depends.next();
			executeJunit(dependsTest,result,true);
		}
		test.run(result);
		doneTests.add(test);
	}
	
	public static void main(String[] args) throws Exception{
		
		Logger.getInstance().showOutput("Starting JAsCo testing framework "+Options.getVersion());
		
		String logdir = DefaultJAsCoTest.getLogDir();
		Logger.getInstance().showOutput("Logs are generated in "+logdir);
		
		StopThread stopThread = new StopThread();
		stopThread.start();
		Thread.sleep(10);
		Logger.getInstance().showOutputNoNewLine("Press 'r' within three seconds to redo all tests instead of failed tests "); 
		Thread.sleep(1000);
		Logger.getInstance().showOutputNoNewLine(".");
		Thread.sleep(1000);
		Logger.getInstance().showOutputNoNewLine(".");
		Thread.sleep(1000);
		Logger.getInstance().showOutput(".");
		Thread.sleep(1000);
		
		if(redoTests) {
			Logger.getInstance().showOutput("Redoing all tests");
		}
		
		Logger.getInstance().showOutput("Cleaning temp dir: "+DefaultJAsCoTest.getOutputDir());
		emptyFolder(new File(DefaultJAsCoTest.getOutputDir()),false);
		
		Logger.getInstance().showOutput("Press 'q' at any time to quit"); 
		
		Logger.getInstance().showOutputNoNewLine("Preparing tests ..... ");
		initTests();
		Logger.getInstance().showOutput("done");
		Logger.getInstance().showOutput("");
		
		int failed = 0;
		int succes = 0;
		java.util.Iterator i = tests.iterator();
		while(i.hasNext()) {
			boolean b = executeTest((IJAsCoTest)i.next(),false);
			if(b)
				succes++;
			else failed++;
		}
		
		
		Logger.getInstance().showOutput("");
		if(failed==0) {
			Logger.getInstance().showOutput("DONE; All tests succeeded!");
			Logger.getInstance().showOutput("Cleaning temp dir: "+DefaultJAsCoTest.getOutputDir());
			emptyFolder(new File(DefaultJAsCoTest.getOutputDir()),false);
		}
		
		else Logger.getInstance().showOutput("DONE; "+succes+" tests succeeded, but "+failed+" tests failed.");
	}
	
	protected static boolean executeTest(IJAsCoTest test, boolean alwaysExecute) {
		
		if(!alwaysExecute&&test.succeededPreviously()&&!redoTests)
			return true;
		
		if(doneTests.contains(test))
			return true;
		
		Iterator depends = getDepends(test);
		while(depends.hasNext()) {
			IJAsCoTest dependsTest = (IJAsCoTest) depends.next();
			executeTest(dependsTest,true);
		}
		
		boolean  b = test.startTest();
		doneTests.add(test);
		return b;
	}
	
	protected static void emptyFolder(File dir) {
		emptyFolder(dir,true);
	}
	
	protected static void emptyFolder(File dir, boolean deleteDir)  {
		
			String[] files = dir.list();
			if(files==null)
				return;
			for(int i=0; i<files.length; ++i)  {
				File f = new File(dir,files[i]);
				if (f.isDirectory()) 
					emptyFolder(f);
				f.delete();
			}
			if(deleteDir)
				dir.delete();
	}
	
	public static void initTests() {
		
		tests.clear();
		
		
		
		CompileAspectTest testa1 = new CompileAspectTest("basic.asp");
		tests.add(testa1);
		
		CompileAspectTest testa2 = new CompileAspectTest("tester.asp");
		testa2.setTarget("1.5");
		tests.add(testa2);
		
		CompileLoggingAspectTest testa3 = new CompileLoggingAspectTest();
		tests.add(testa3);
		
		CompileAspectTest testa4 = new CompileAspectTest("extended.asp");
		addDepend(testa4,testa1);
		tests.add(testa4);
		
		CompileAspectTest testa5 = new CompileAspectTest("cflowClass.asp");
		tests.add(testa5);
		
		CompileAspectTest testa6 = new CompileAspectTest("deleter.asp");
		tests.add(testa6);
		
		CompileAspectTest testa16 = new CompileAspectTest("Stateful.asp");
		tests.add(testa16);
		
		CompileAspectTest testa17 = new CompileAspectTest("TargetTest.asp");
		tests.add(testa17);
		
		CompileAspectTest testa18 = new CompileAspectTest("AfterReturningTest.asp");
		tests.add(testa18);
		
		
		CompileConnectorTest testc3 = new CompileConnectorTest("aspectonaspecttracing.con");
		addDepend(testc3,testa3);
		addDepend(testc3,testa1);
		tests.add(testc3);
	
		CompileConnectorTest testc30 = new CompileConnectorTest("extendedConnector.con");
		addDepend(testc30,testa4);
		tests.add(testc30);
		
		
		SimpleTransformTest testtransform = new SimpleTransformTest("java.awt.Component");
		testtransform.addOption("-noevents");
		tests.add(testtransform);
		
		SimpleTransformTest testtransform2 = new SimpleTransformTest("javax.swing.JComponent");
		testtransform2.addOption("-A");
		tests.add(testtransform2);
	
//		prio tests
		{
		CompileConnectorTest testc1 = new CompileConnectorTest("basicConnector.con");
		addDepend(testc1,testa1);
			
		RemoveConnectors removeConnectorsPrio = new RemoveConnectors();
		CompileConnectorTest testc10 = new CompileConnectorTest("basicConnectorCopy.con");
		addDepend(testc10,testa1);
		
		Vector prioDepend = new Vector();
		prioDepend.add(removeConnectorsPrio);
		prioDepend.add(testc10);
		prioDepend.add(testc1);
		
		RunConnectorPriorityTest runprio1 = new RunConnectorPriorityTest(RunPacoSuiteTest.RUNTIME_WEAVER);
		addDepend(runprio1,prioDepend);
		tests.add(runprio1);
		
		RunConnectorPriorityTest runprio2 = new RunConnectorPriorityTest(RunPacoSuiteTest.HOTSWAP2);
		addDepend(runprio2,prioDepend);
		tests.add(runprio2);
		
		RunConnectorPriorityTest runprio3 = new RunConnectorPriorityTest(RunPacoSuiteTest.NO_JUTTA);
		addDepend(runprio3,prioDepend);
		tests.add(runprio3);
		}
		
				
		//meta data tests
		Vector dependsMetaData = new Vector();
		RemoveConnectors removeConnectorsMetaData = new RemoveConnectors();
		
		
		CompileConnectorTest testc12 = new CompileConnectorTest("MetaDataConnector.con");
		addDepend(testc12,testa1);
		dependsMetaData.add(removeConnectorsMetaData);
		dependsMetaData.add(testc12);
		
		RunMetaDataTest metadata1 = new RunMetaDataTest(RunPacoSuiteTest.HOTSWAP1);
		addDepend(metadata1,dependsMetaData);
		tests.add(metadata1);
		
		
		RunMetaDataTest metadata2 = new RunMetaDataTest(RunPacoSuiteTest.HOTSWAP2);
		addDepend(metadata2,dependsMetaData);
		tests.add(metadata2);
		
		RunMetaDataTest metadata3 = new RunMetaDataTest(RunPacoSuiteTest.RUNTIME_WEAVER);
		addDepend(metadata3,dependsMetaData);
		tests.add(metadata3);
		 
		
		tests.add(testc12);

		
		//transforms juggler and JButton, has to be after run juggler tests...
		CompileTransformConnectorTest testc2 = new CompileTransformConnectorTest();
		addDepend(testc2,testa2);
		addDepend(testc2,testa5);
		tests.add(testc2);
		
		
		RunJAsCoInterfaceTest interfaceTest = new RunJAsCoInterfaceTest();
		RemoveConnectors removeConnectorsI = new RemoveConnectors();
		addDepend(interfaceTest,removeConnectorsI);
		tests.add(interfaceTest);
		
		//		stateful cflow tests
		createTests("StatefulControlFlow.asp","StatefulControlFlowConnector.con",RunStatefulControlFlowTest.class);
		
		//covariance tests
		{
			createTests("basic.asp","basicConnector.con",RunJAsCoCovarianceTest.class);
		     
		}
		//		Distributed tests, compile only so far
		{
			CompileAspectTest dist = new CompileAspectTest("distributed.asp");
			tests.add(dist);
		}
		
//		Enable-disable tests
		{	
			createTests("HelloAspect.asp","HelloConnector.con",RunJAsCoEnableTest.class);
			 
		}
		
//		non-static connector tests
		{	
			createTests("HelloAspect.asp","HelloConnector.con",RunJAsCoStaticTest.class);
		}
		
//		before&reflection tests
		{	
			createTests("BeforeTest.asp","BeforeTestConnector.con",RunJAsCoBeforeTest.class);
		}
		
		//		Aspect Bean Instantiation tests
		{
			createTests("AspectBeanInstantiationTest.asp","AspectBeanInstantiation.con",RunAspectInstantiationTest.class);
			
		}
		
		//benchmark tests
		{
		RemoveConnectors removeConnectors1 = new RemoveConnectors();
		
		CompileAspectTest testb1 = new CompileAspectTest("VoidAspect.asp");
		
		CompileConnectorTest testb2 = new CompileConnectorTest("BenchConnector.con");
		
		Vector dependsRun = new Vector();
		dependsRun.add(removeConnectors1);
		dependsRun.add(testb1);
		dependsRun.add(testb2);
		
		RunJAsCoProgramTest runbench2 = new RunBenchTest(RunPacoSuiteTest.HOTSWAP1, 400);
	    addDepend(runbench2,dependsRun);
		tests.add(runbench2);
		
		RunJAsCoProgramTest runbench3 = new RunBenchTest(RunPacoSuiteTest.HOTSWAP2, 350);
		addDepend(runbench3,dependsRun);
		tests.add(runbench3);
		
		RunJAsCoProgramTest runbench4 = new RunBenchTest(RunPacoSuiteTest.RUNTIME_WEAVER, 175);
		addDepend(runbench4,dependsRun);
		tests.add(runbench4);
	}
		//counting tests
		{
			createTests("Counting.asp","CountingConnector.con",RunCountTest.class);
		
	}
		{	
		// after throwing/returning tests
			createTests("Around.asp","aroundRTConnector.con",RunAroundThrowingReturningTest.class);
		}
		
		{	
			// subtype test...
			createTests("basic.asp","SubtypeTestConnector.con",RunSubtypeTest.class);
			
		}
		
		{	
			// limited subtype test...
			createTests("basic.asp","LimitedSubtypeTestConnector.con",RunLimitedSubtypeTest.class);
		}

		
		{	
			// jumping aspects tests...
			Vector dependsJumping = new Vector();
			RemoveConnectors removeConnectorsJumping = new RemoveConnectors();
			
			CompileAspectTest testjumping1 = new CompileAspectTest("Jumping.asp");
			CompileConnectorTest testjumping2 = new CompileConnectorTest("JumpingLogging.con");
			
			dependsJumping.add(removeConnectorsJumping);
			dependsJumping.add(testjumping1);
			dependsJumping.add(testjumping2);
			
			RunJAsCoProgramTest runjumping = new RunJumpingAspectsTest();
			addDepend(runjumping,dependsJumping);
			tests.add(runjumping);
			}
		{	
			//aspects on aspects test
			Vector dependsAonA = new Vector();
			RemoveConnectors removeConnectorsAonA = new RemoveConnectors();
			
			CompileAspectTest testAonA1 = new CompileAspectTest("logging.asp");
			CompileConnectorTest testAonA2 = new CompileConnectorTest("JumpingTestLoggingConnector.con");
			addDepend(testAonA2,testa1);
			CompileConnectorTest testAonA3 = new CompileConnectorTest("aspectonaspecttracing.con");
			addDepend(testAonA3,testAonA1);
			addDepend(testAonA3,testa1);
			
			dependsAonA.add(removeConnectorsAonA);
			dependsAonA.add(testAonA2);
			dependsAonA.add(testAonA3);
			
//			test aspects on aspects trap adding at runtime:
			/* tests fail because aspects are loaded by boot classpath loader and are thus not subject for hotswapping!
			 * 
			 * RunJAsCoProgramTest runAonA1 = new RunAspectOnAspectTest(RunJAsCoProgramTest.HOTSWAP2,true);
			addDepend(runAonA1,dependsAonA);
			tests.add(runAonA1);
			
			RunJAsCoProgramTest runAonA2 = new RunAspectOnAspectTest(RunJAsCoProgramTest.RUNTIME_WEAVER,true);
			addDepend(runAonA2,dependsAonA);
			tests.add(runAonA2);*/
			
//			test aspects on aspects trap adding at compile time:
			CompileAspectTest testAonA0 = new CompileAspectTest("basic.asp");
			testAonA0.addOption("-X");
			
			CompileConnectorTest testAonA5 = new CompileConnectorTest("JumpingTestLoggingConnector.con");
			addDepend(testAonA5,testAonA0);
			CompileConnectorTest testAonA4 = new CompileConnectorTest("aspectonaspecttracing.con");
			addDepend(testAonA4,testAonA1);
			addDepend(testAonA4,testAonA0);
	
			dependsAonA.clear();
			dependsAonA.add(new RemoveConnectors());
			dependsAonA.add(testAonA4);
			dependsAonA.add(testAonA5);
			
			RunJAsCoProgramTest runAonA3 = new RunAspectOnAspectTest(RunJAsCoProgramTest.HOTSWAP2,false);
			addDepend(runAonA3,dependsAonA);
			tests.add(runAonA3);
			
			RunJAsCoProgramTest runAonA4 = new RunAspectOnAspectTest(RunJAsCoProgramTest.RUNTIME_WEAVER,false);
			addDepend(runAonA4,dependsAonA);
			tests.add(runAonA4);
			}
		
		{ //refinement module tests
			
			RemoveConnectors remove= new RemoveConnectors();
			
			CompileAspectTest testaspect = new CompileAspectTest("ToRefineAspect.asp");
			testaspect.setTarget("1.5");
			tests.add(testaspect);
			
			CompileRefinementTest testref = new CompileRefinementTest("TestRefinement.ref");
			addDepend(testref,testaspect);
			tests.add(testref);
			
			CompileRefinementTest testref2 = new CompileRefinementTest("TestRefinementChild.ref");
			addDepend(testref2,testaspect);
			addDepend(testref2,testref);
			tests.add(testref2);
			
			CompileConnectorTest testcon1 = new CompileConnectorTest("InlineRefiningConnector.con");
			addDepend(testcon1,testaspect);
			
			CompileConnectorTest testcon2 = new CompileConnectorTest("RefineConnector.con");
			addDepend(testcon2,testaspect);
			
			RunRefiningClassesTests run = new RunRefiningClassesTests();
			addDepend(run,remove);
			addDepend(run,testaspect);
			addDepend(run,testcon2);
			addDepend(run,testcon1);
			addDepend(run,testref);
			addDepend(run,testref2);
			
			
			tests.add(run);
			
		}
		
		{ //mixin tests 
			createTests("MixinTest.asp","MixinTestConnector.con",RunMixinTest.class);
			
			/*RemoveConnectors remove= new RemoveConnectors();
			
			CompileAspectTest testaspect = new CompileAspectTest("MixinTest.asp");
			tests.add(testaspect);
			
			CompileConnectorTest testcon1 = new CompileConnectorTest("MixinTestConnector.con");
			addDepend(testcon1,testaspect);
			
			RunMixinTest run = new RunMixinTest(RunMixinTest.HOTSWAP2);
			addDepend(run,remove);
			addDepend(run,testcon1);
			
			RunMixinTest run2 = new RunMixinTest(RunMixinTest.RUNTIME_WEAVER);
			addDepend(run2,remove);
			addDepend(run2,testcon1);
			
			tests.add(run);
			tests.add(run2);*/
		}
		
		{ //strict stateful aspects tests (context)
			RemoveConnectors remove= new RemoveConnectors();
			
			CompileAspectTest testaspect = new CompileAspectTest("Stateful1.asp");
			tests.add(testaspect);
			
			CompileConnectorTest testcon1 = new CompileConnectorTest("StatefulCon1.con");
			addDepend(testcon1,testaspect);
			
			RunStrictTest run = new RunStrictTest(true);
			addDepend(run,remove);
			addDepend(run,testcon1);
			 
			tests.add(run);
		}
		
		{ //strict stateful aspects tests (NO context)
			RemoveConnectors remove= new RemoveConnectors();
			
			CompileAspectTest testaspect = new CompileAspectTest("Stateful1.asp");
			tests.add(testaspect);
			
			CompileConnectorTest testcon1 = new CompileConnectorTest("StatefulCon2.con");
			addDepend(testcon1,testaspect);
			
			RunStrictTest run = new RunStrictTest(false);
			addDepend(run,remove);
			addDepend(run,testcon1);
			
			tests.add(run);
		}
		
		{ //polymorphic hook instantiation tests..
			RemoveConnectors remove= new RemoveConnectors();
			CompileAspectTest testploy1 = new CompileAspectTest("AbstractAspect.asp");
			tests.add(testploy1);
			
			CompileAspectTest testploy2 = new CompileAspectTest("ConcreteAspect.asp");
			addDepend(testploy2,testploy1);
			tests.add(testploy2);

			CompileConnectorTest testploy3 = new CompileConnectorTest("PolyConnector.con");
			addDepend(testploy3,testploy2);
			addDepend(testploy3,testploy1);
			
			RunSubhookTest run = new RunSubhookTest(RunSubhookTest.HOTSWAP2);
			RunSubhookTest run2 = new RunSubhookTest(RunSubhookTest.RUNTIME_WEAVER);
			addDepend(run,remove);
			addDepend(run,testploy3);
			addDepend(run2,remove);
			addDepend(run2,testploy3);
			
			tests.add(run);
			tests.add(run2);
		}
		
		{ //dynamic arguments tests...
			RemoveConnectors remove= new RemoveConnectors();
			CompileAspectTest test1 = new CompileAspectTest("DynArgsTestAspect.asp");
			tests.add(test1);
			
			CompileConnectorTest test2 = new CompileConnectorTest("DynArgsConnector.con");
			addDepend(test2,test1);
			
			RunDynArgsTest run = new RunDynArgsTest(RunDynArgsTest.NO_JUTTA);
			RunDynArgsTest run2 = new RunDynArgsTest(RunDynArgsTest.HOTSWAP2);
			RunDynArgsTest run3 = new RunDynArgsTest(RunDynArgsTest.RUNTIME_WEAVER);
			addDepend(run,remove);
			addDepend(run,test2);
			addDepend(run2,remove);
			addDepend(run2,test2);
			addDepend(run3,remove);
			addDepend(run3,test2);
			
			tests.add(run);
			tests.add(run2);tests.add(run3);
			
		}
		
		{ //after tests...
			RemoveConnectors remove= new RemoveConnectors();
			CompileAspectTest test1 = new CompileAspectTest("AfterTest.asp");
			tests.add(test1);
			
			CompileConnectorTest test2 = new CompileConnectorTest("AfterConnector.con");
			addDepend(test2,test1);
			
			RunAfterTest run = new RunAfterTest(RunAfterTest.NO_JUTTA);
			RunAfterTest run2 = new RunAfterTest(RunAfterTest.HOTSWAP2);
			RunAfterTest run3 = new RunAfterTest(RunAfterTest.RUNTIME_WEAVER);
			addDepend(run,remove);
			addDepend(run,test2);
			addDepend(run2,remove);
			addDepend(run2,test2);
			addDepend(run3,remove);
			addDepend(run3,test2);
			
			tests.add(run);
			tests.add(run2);
			tests.add(run3);
			
		}
		
		{	//conditional around tests
			RemoveConnectors remove= new RemoveConnectors();
			CompileAspectTest conAround = new CompileAspectTest("ConditionalAround.asp");
			tests.add(conAround);
			CompileConnectorTest conAroundCon = new CompileConnectorTest("ConditionalAroundConnector.con");
			addDepend(conAroundCon,conAround);
			
			RunConditionalAroundTest run = new RunConditionalAroundTest(RunAfterTest.NO_JUTTA);
			RunConditionalAroundTest run2 = new RunConditionalAroundTest(RunAfterTest.HOTSWAP2);
			RunConditionalAroundTest run3 = new RunConditionalAroundTest(RunAfterTest.RUNTIME_WEAVER);
			addDepend(run,remove);
			addDepend(run,conAroundCon);
			addDepend(run2,remove);
			addDepend(run2,conAroundCon);
			addDepend(run3,remove);
			addDepend(run3,conAroundCon);
			
			tests.add(run);
			tests.add(run2);
			tests.add(run3);
		}

		{	//object tests
			RemoveConnectors remove= new RemoveConnectors();
			CompileAspectTest conAround = new CompileAspectTest("ConditionalAround.asp");
			tests.add(conAround);
			CompileConnectorTest conAroundCon = new CompileConnectorTest("ObjectTestConnector.con");
			addDepend(conAroundCon,conAround);
			
			RunObjectTest run = new RunObjectTest(RunAfterTest.NO_JUTTA);
			RunObjectTest run2 = new RunObjectTest(RunAfterTest.HOTSWAP2);
			RunObjectTest run3 = new RunObjectTest(RunAfterTest.RUNTIME_WEAVER);
			addDepend(run,remove);
			addDepend(run,conAroundCon);
			addDepend(run2,remove);
			addDepend(run2,conAroundCon);
			addDepend(run3,remove);
			addDepend(run3,conAroundCon);
			
			tests.add(run);
			tests.add(run2);
			tests.add(run3);
		}
		
		{	//per* tests
			RemoveConnectors remove= new RemoveConnectors();
			CompileAspectTest conAround = new CompileAspectTest("PerTestAspect.asp");
			tests.add(conAround);
			CompileConnectorTest conAroundCon = new CompileConnectorTest("PerTestConnector.con");
			addDepend(conAroundCon,conAround);
			
			CompileConnectorTest conAroundCon2 = new CompileConnectorTest("PerStaticConnectorTest.con");
			addDepend(conAroundCon2,conAround);
			
			RunPerTest run = new RunPerTest(RunAfterTest.NO_JUTTA);
			RunPerTest run2 = new RunPerTest(RunAfterTest.HOTSWAP2);
			RunPerTest run3 = new RunPerTest(RunAfterTest.RUNTIME_WEAVER);
			addDepend(run,remove);
			addDepend(run,conAroundCon);
			addDepend(run,conAroundCon2);
			addDepend(run2,remove);
			addDepend(run2,conAroundCon);
			addDepend(run2,conAroundCon2);
			addDepend(run3,remove);
			addDepend(run3,conAroundCon);
			addDepend(run3,conAroundCon2);
			
			tests.add(run);
			tests.add(run2);
			tests.add(run3);
			
		}
		
		{
			//traversal test
			
			CompileTraversalTest testt1 = new CompileTraversalTest("TestTraversal.trv");
			addDepend(testt1,testa1);
			
			RemoveConnectors remTrav = new RemoveConnectors();
			RunTraversalTest runtrav = new RunTraversalTest();
			addDepend(runtrav,remTrav);
			addDepend(runtrav,testt1);
			tests.add(runtrav);	
		}
		
		{
			//Multiple Execs Test
			RemoveConnectors rem = new RemoveConnectors();
			CompileAspectTest m = new CompileAspectTest("MultipleExec.asp");
			tests.add(m);
			CompileConnectorTest con = new CompileConnectorTest("MutlipleExecConnector.con");
			addDepend(con,m);
			 
			
			RunMultipleExecTest run1 = new RunMultipleExecTest(RunMultipleExecTest.NO_JUTTA);
			RunMultipleExecTest run2 = new RunMultipleExecTest(RunMultipleExecTest.HOTSWAP2);
			RunMultipleExecTest run3 = new RunMultipleExecTest(RunMultipleExecTest.RUNTIME_WEAVER);
			Vector depends = new Vector();
			depends.add(rem);depends.add(con);
			addDepend(run1,depends);
			addDepend(run2,depends);
			addDepend(run3,depends);
			tests.add(run1);	
			tests.add(run2);
			tests.add(run3);
		}
		
		{
			//Returning Primitive around tests
			CompileAspectTest m = new CompileAspectTest("PrimitiveAroundTest.asp");
			tests.add(m);
			CompileConnectorTest con = new CompileConnectorTest("PrimitiveAroundConnector.con");
			addDepend(con,m);
			 
			addRunJAsCoProgramTests(RunPrimitiveAroundTest.class,"",con,m);
			
			{
//			Returning Primitive around tests returning
			CompileConnectorTest con2 = new CompileConnectorTest("PrimitiveAroundReturningConnector.con");
			addDepend(con2,m);
			addRunJAsCoProgramTests(RunPrimitiveAroundTest.class," returning",con2,m);
			}
			
			{
//				Returning Primitive around tests no inline
				CompileConnectorTest con2 = new CompileConnectorTest("PrimitiveAroundNoInlineConnector.con");
				addDepend(con2,m);
				addRunJAsCoProgramTests(RunPrimitiveAroundTest.class," no inline",con2,m);
			}
		}
		
		{
		//Returning Object around tests
		CompileAspectTest m = new CompileAspectTest("ObjectAroundTest.asp");
		tests.add(m);
		CompileConnectorTest con = new CompileConnectorTest("ObjectAroundConnector.con");
		addDepend(con,m);
		 
		addRunJAsCoProgramTests(RunObjectAroundTest.class,"",con,m);
		
		{
//		Returning Object around tests returning
		CompileConnectorTest con2 = new CompileConnectorTest("ObjectAroundReturningConnector.con");
		addDepend(con2,m);
		addRunJAsCoProgramTests(RunObjectAroundTest.class," returning",con2,m);
		}
		
		{
//			Returning Object around tests no inline
			CompileConnectorTest con2 = new CompileConnectorTest("ObjectAroundNoInlineConnector.con");
			addDepend(con2,m);
			addRunJAsCoProgramTests(RunObjectAroundTest.class," no inline",con2,m);
		}
		}
		
		{//		run juggler tests.... 
			Vector dependsRunJuggler = new Vector();
			RemoveConnectors removeConnectorsJuggler = new RemoveConnectors();
			
			CompileConnectorTest testc1 = new CompileConnectorTest("basicConnector.con");
			addDepend(testc1,testa1);
			
			CompileConnectorTest testc123 = new CompileConnectorTest("subtypeConnector.con");
			addDepend(testc123,testa1);
			
			CompileConnectorTest testc4 = new CompileConnectorTest("buttonDelConnector.con");
			testc4.addOption("-target:1.5");
			addDepend(testc4,testa6);
			
			
			CompileConnectorTest testc9 = new CompileConnectorTest("basic2Connector.con");
			addDepend(testc9,testa2);
			addDepend(testc9,testa5);
			
			
			CompileConnectorTest testc11 = new CompileConnectorTest("basicConnectorArray.con");
			addDepend(testc11,testa1);
			
			
			CompileConnectorTest testc13 = new CompileConnectorTest("TargetConnector.con");
			addDepend(testc13,testa17);
			
			
			CompileConnectorTest testc14 = new CompileConnectorTest("AfterReturningConnector.con");
			addDepend(testc14,testa18);
			
			
			dependsRunJuggler.add(removeConnectorsJuggler);
			dependsRunJuggler.add(testc4);
			dependsRunJuggler.add(testc1);
			dependsRunJuggler.add(testc9);
			dependsRunJuggler.add(testc14);
			dependsRunJuggler.add(testc13);
			dependsRunJuggler.add(testc11);
			dependsRunJuggler.add(testc123);
			

			RunJugglerTest runjugglerbench1 = new RunJugglerTest(RunPacoSuiteTest.HOTSWAP2);
			addDepend(runjugglerbench1,dependsRunJuggler);
			tests.add(runjugglerbench1);
			
			RunJugglerTest runjugglerbench2 = new RunJugglerTest(RunPacoSuiteTest.RUNTIME_WEAVER);
			addDepend(runjugglerbench2,dependsRunJuggler);
			tests.add(runjugglerbench2);}
		
		
//		pacosuite tests
		
		if(getPacoSuiteDir()!=null)  {//no pacosuite tests
			RemoveConnectors removeConnectors2 = new RemoveConnectors();
			
			TransformBeanTest testtransform3 = new TransformBeanTest();
			tests.add(testtransform3);
		
			RunPacoSuiteTest paco1 = new RunPacoSuiteTest(RunPacoSuiteTest.TRANSFORM_BEAN);
			addDepend(paco1,removeConnectors2);
			addDepend(paco1,testtransform3);
			tests.add(paco1);
		
			RunPacoSuiteTest paco2 = new RunPacoSuiteTest(RunPacoSuiteTest.HOTSWAP1);
			addDepend(paco2,removeConnectors2);
			tests.add(paco2);
		
			RunPacoSuiteTest paco3 = new RunPacoSuiteTest(RunPacoSuiteTest.HOTSWAP2);
			addDepend(paco3,removeConnectors2);
			tests.add(paco3);
		
			RunPacoSuiteTest paco4 = new RunPacoSuiteTest(RunPacoSuiteTest.RUNTIME_WEAVER);
			addDepend(paco1,removeConnectors2);
			tests.add(paco4);

			//			app server tests
			RunPacoSuiteTest paco5 = new RunPacoSuiteTest(RunPacoSuiteTest.HOTSWAP2);
			addDepend(paco5,removeConnectors2);
			paco5.addOption("-Djasco.hotswap.applicationserver=true");
			tests.add(paco5); 
		} else Logger.getInstance().showWarning("jasco.testing.pacosuiteDir not set: skipping PacoSuite tests");
		
		 
		 
		//TODO: comb strat and precedence test
		//TODO: test heuristsics of inline compiler
	}
 
	public static void addRunJAsCoProgramTests(Class runner, String descr, IJAsCoTest... depends) {
		
		try {
		Vector dependsV = new Vector();
		RemoveConnectors remove = new RemoveConnectors();
		dependsV.add(remove);
		
		
		dependsV.addAll(java.util.Arrays.asList(depends));
		
		Constructor constr = null;
		Object[] args = null;
		if(descr==null) {
			constr=runner.getConstructor(Integer.TYPE);
			args=new Object[1];
		}
		else { 
			constr=runner.getConstructor(Integer.TYPE,String.class);
			args=new Object[2];
			args[1]=descr;
		}
		
		args[0]=RunJAsCoProgramTest.NO_JUTTA;
		RunJAsCoProgramTest run1 = (RunJAsCoProgramTest)constr.newInstance(args); 
		args[0]=RunJAsCoProgramTest.HOTSWAP2;
		RunJAsCoProgramTest run2 = (RunJAsCoProgramTest)constr.newInstance(args); 
		args[0]=RunJAsCoProgramTest.RUNTIME_WEAVER;
		RunJAsCoProgramTest run3 = (RunJAsCoProgramTest)constr.newInstance(args); 
		
		addDepend(run1,dependsV);
		addDepend(run2,dependsV);
		addDepend(run3,dependsV);
		
		tests.add(run1);	
		tests.add(run2);
		tests.add(run3);
		}catch(Throwable t) {
			Logger.getInstance().showError("Failed to init tests: "+runner.getName());
		}
	}
	
	public static String getPacoSuiteDir() {
		
		return pacoSuiteDir;
	}
	
	public static int getBenchBaseResult() {
		
		return benchBaseResult;
	}
	
	public static void addDepend(IJAsCoTest test, Vector dependsOn) {
		
		 Vector v = (Vector) depends.get(test);
		 if(v==null) {
		 	v=new Vector();
		 	depends.put(test,v);
		 }
		v.addAll(dependsOn);
	}
	
	public static void addDepend(IJAsCoTest test, IJAsCoTest dependsOn) {
		
		 Vector v = (Vector) depends.get(test);
		 if(v==null) {
		 	v=new Vector();
		 	depends.put(test,v);
		 }
		 v.add(dependsOn);
	}
	
	public static Iterator getDepends(IJAsCoTest test) {
		
		Vector v =  (Vector) depends.get(test);
		if(v==null) v=new Vector();
		
		return v.iterator();
	}
	
	static class StopThread extends Thread {
		
		StopThread() {
			setDaemon(true);
		}
		
		public void run() {
			try {
			InputStream is = System.in;
			while(true) {
				int i = is.read();
				if(i==-1)
					break;
				if(i=='q') {
					Logger.getInstance().showOutput("User intervention detected, halting processes .....");
					System.exit(0);
				}
				if(i=='r') {
					redoTests=true;
				}
					
			
			}
			}catch(Exception ex) {}
		}
	}
	
	private static int[] testsToRun = {RunJAsCoProgramTest.NO_JUTTA,RunJAsCoProgramTest.HOTSWAP2,RunJAsCoProgramTest.RUNTIME_WEAVER};
	
	private static void createTests(String aspect, String connector, Class<? extends RunJAsCoProgramTest> test) {
		createTests(new String[] {aspect},new String[] {connector},test);
	}
	
	public static void createTests(String[] aspects, String[] connectors, Class<? extends RunJAsCoProgramTest> testClass) {
		
		RemoveConnectors remove = new RemoveConnectors();
		
		Vector depends = new Vector();
		depends.add(remove);
		
		Vector<CompileAspectTest> aspectsVector = new Vector<CompileAspectTest>();
		for(int i=0;i<aspects.length;++i) {
			CompileAspectTest helloAspect = new CompileAspectTest(aspects[i]);
			aspectsVector.add(helloAspect);
		}
		
		Vector<CompileConnectorTest> connectorsVector = new Vector<CompileConnectorTest>();
		for(int i=0;i<aspects.length;++i) {
			CompileConnectorTest helloConnector = new CompileConnectorTest(connectors[i]);
			connectorsVector.add(helloConnector);
			addDepend(helloConnector,aspectsVector);
		}
		
		depends.addAll(connectorsVector);
		
		try {
			Constructor<? extends RunJAsCoProgramTest> testClassConstructor = testClass.getConstructor(new Class[] {Integer.TYPE});
			for(int i=0;i<testsToRun.length;++i) {
				
				RunJAsCoProgramTest run1 = testClassConstructor.newInstance(testsToRun[i]);
				addDepend(run1,depends);
				tests.add(run1);
			}
			
		} catch (SecurityException e) {
			jasco.util.logging.Logger.getInstance().showError(e);
		} catch (NoSuchMethodException e) {
			jasco.util.logging.Logger.getInstance().showError(e);
		} catch (IllegalArgumentException e) {
			jasco.util.logging.Logger.getInstance().showError(e);
		} catch (InstantiationException e) {
			jasco.util.logging.Logger.getInstance().showError(e);
		} catch (IllegalAccessException e) {
			jasco.util.logging.Logger.getInstance().showError(e);
		} catch (InvocationTargetException e) {
			jasco.util.logging.Logger.getInstance().showError(e);
		}
	}
}
