/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.tools;

import jasco.options.Options;
import jasco.runtime.hotswap.JAsCoClassLoader;
import jasco.runtime.hotswap.TransformClassChecker;
import jasco.runtime.transform.ClassPreTransformer;
import jasco.runtime.transform.GeneratedFileManager;
import jasco.runtime.transform.JAsCoClassTransformer;
import jasco.runtime.transform.MethodAccumulator;
import jasco.util.logging.Logger;

import java.io.File;
import java.io.IOException;
import java.util.Vector;

import javassist.CtClass;
import javassist.CtField;
import javassist.NotFoundException;

public class BeanTransformerJavassist 
{
	public static final String JASCO_BEAN_INT = "jasco.runtime.JAsCoBean";
	public static final String JASCO_BEAN_FIELD = "_JAsCo_isAJAsCoBean";
	 
	public static boolean isAJascoBean(CtClass theClass)  {
	
		try { 
			
		CtClass[] ints = theClass.getInterfaces();
		for(int i=0;i<ints.length;++i) {
			
			if(ints[i].getName().equals(JASCO_BEAN_INT))
				return true;
		}
		}
		catch(Exception x)  {
			
		}
		
		try {
		theClass.getField(JASCO_BEAN_FIELD);
		return true;
		}
		catch(NotFoundException x)  {
			
		}
		return false;
	}
	
	protected TransformClassChecker getTransformChecker() { 
		
		return TransformClassChecker.getSingleton();
	}
	
	public boolean transformBean(String classname,Vector theexcludedmethods)  {
		
		File tmpDir = Options.getTempDir();
		Options.setTempDir(new File(Options.getOutputDir()));
		boolean b = Options.transformHooks();
		Options.setTransformHooks(true);
		
		Logger.getInstance().showDebug("Transforming class: "+classname);
		try {
			JAsCoClassLoader.getSingleton().appendListToClassPath(Options.getClassPath());
			CtClass clz = JAsCoClassLoader.getSingleton().loadClass(classname);
			
			if(isAJascoBean(clz)) {
				Logger.getInstance().showOutput("Transforming class failed, is already JAsCo bean: "+classname);
				return false;
			}
			
			if(getTransformChecker().checkValidType(clz.getName())==null) {
				Logger.getInstance().showOutput("Transforming class failed, illegal type: "+classname);
				return false;
			}
			
			
			MethodAccumulator methodAcc = new MethodAccumulator(theexcludedmethods);
			methodAcc.setDebugString("Preprocessing");
			ClassPreTransformer preprocessor = new ClassPreTransformer();
			
			methodAcc.process(clz,clz.getName());
			preprocessor.setProcessMethods(methodAcc.getResultingMethods());

			boolean result = preprocessor.process(clz,clz.getName());
			if(!result) {//no changes made...
				return true;
			}
			
			JAsCoClassTransformer transformer = new JAsCoClassTransformer(clz);
			transformer.insertTrapsAtRuntime(theexcludedmethods,null,ClassLoader.getSystemClassLoader());
			 
			String dir = Options.getOutputDir();
			while(dir.endsWith(File.separator))
				dir=dir.substring(0,dir.length()-1);
			
			makeJAsCoBean(clz);
			clz.writeFile(dir);
			
			return true;
			
			//FileWriter writer = new FileWriter();
			
		} catch(NotFoundException ex) {
			{Logger.getInstance().showWarning("Transforming class failed, no class def or jar found error: "+ex.getMessage());}
		}
		catch(IOException ex) {
			{Logger.getInstance().showError("Transforming class failed, reading/writing class failed: "+classname);
			Logger.getInstance().showError(ex);}
		}
		catch(Exception ex) {
			{Logger.getInstance().showError("Transforming class failed: "+classname);Logger.getInstance().showError(ex);}
			
		}
		finally {
			Options.setTempDir(tmpDir);
			Options.setTransformHooks(b);
			GeneratedFileManager.unregisterAllTempFiles();
		}
		
		return false;
	}
	
	protected void makeJAsCoBean(CtClass clz2)  {
		
		try {
			CtClass clz = JAsCoClassLoader.getSingleton().loadClass(JASCO_BEAN_INT);
			clz2.addInterface(clz);
			CtField field = new CtField(clz,JASCO_BEAN_FIELD,clz2);
			clz2.addField(field);
		}catch(Exception ex )  {
			Logger.getInstance().showError(ex);
		}
	}
			
}
