/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.tools.aspectparser;

import jasco.runtime.aspect.PCutpointConstructorAnd;
import jasco.runtime.aspect.PCutpointConstructorApplicationDesignator;
import jasco.runtime.aspect.PCutpointConstructorArgs;
import jasco.runtime.aspect.PCutpointConstructorCall;
import jasco.runtime.aspect.PCutpointConstructorCflow;
import jasco.runtime.aspect.PCutpointConstructorComplement;
import jasco.runtime.aspect.PCutpointConstructorExecute;
import jasco.runtime.aspect.PCutpointConstructorExecutionEqualsJoinpointHost;
import jasco.runtime.aspect.PCutpointConstructorExecutionHost;
import jasco.runtime.aspect.PCutpointConstructorMatchAll;
import jasco.runtime.aspect.PCutpointConstructorNot;
import jasco.runtime.aspect.PCutpointConstructorOr;
import jasco.runtime.aspect.PCutpointConstructorPHost;
import jasco.runtime.aspect.PCutpointConstructorRegular;
import jasco.runtime.aspect.PCutpointConstructorStrict;
import jasco.runtime.aspect.PCutpointConstructorStrictComplement;
import jasco.runtime.aspect.PCutpointConstructorTarget;
import jasco.runtime.aspect.PCutpointConstructorWithincode;
import jasco.runtime.aspect.State;
import jasco.runtime.distribution.hostselection.IHostSelectionStrategy;
import jasco.tools.jascoparser.JascoParseException;
import jasco.tools.jascoparser.JascoParserTokenManager;
import jasco.tools.jascoparser.PImport;
import jasco.tools.jascoparser.PJavaCode;
import jasco.util.javacompiler.CompileError;
import jasco.util.logging.Logger;

import java.util.Iterator;
import java.util.TreeMap;
import java.util.Vector;

public class AspectParser extends jasco.tools.jascoparser.JascoParser
{

    private PAspect aspect;
    private PCutpoint tempcutpoint;
    
	private static final String NOCONTEXT=  "nocontext";
	private static final String JOINPOINTHOST_STR = "joinpointhost";

	public PAspect getParsedAspect()
    {
        return aspect;
    }

    
    public AspectParser(JascoParserTokenManager aAspectParserTokenManager, String aFileName)
    {	
		super(aAspectParserTokenManager,aFileName);
        aspect = new PAspect();
        tempcutpoint = new PCutpoint();
    }
	
    public void start()
    {
        try
        {
            parsePackageDeclaration(aspect);
            aspect.setImports(parseImportDeclaration());
            aspect.addImport(new PImport(aspect.getPackage().getPackageName()+".*",aspect.getPackage().getLineNumber()));
            parseAspectDeclaration();
        }
        catch(JascoParseException exception)
        {
            //System.err.println(exception.toString());
            //exception.printStackTrace();
            //System.err.println("1 error");
            Logger.getInstance().showError(exception.getError());
            Logger.getInstance().showOutput("1 error");
            
			jasco.options.Options.exit(1);
        }
    }
    
    @Override
	protected boolean isExistingClassName(String name) {
		if(aspect!=null&&aspect.getName().equals(name))
			return true;
		else if(aspect!=null&&aspect.getFullName().equals(name))
			return true;

		else return super.isExistingClassName(name);
	}
    
    

    private void parseAspectDeclaration()throws JascoParseException
    {	
		String aspectdeclarationstring = getCurrentImage();
		StringBuffer annotations = new StringBuffer();
		
		int modifiers = parseModifiers(annotations, aspect.getImports());
		
		
        String errorstring = " 'class' expected\n";
        CompileError error = new CompileError(getFileName(), getPreviousLine(), errorstring);
        assertToken(getCurrentId(), CLASS, error);
		
        getNextToken();
		
        checkIDENTIFIER(aspectdeclarationstring);
        aspect.setAspectName(getCurrentImage());
		aspect.setModifiers(modifiers);
		aspect.setLine(getPreviousLine());
		aspect.setAnnotations(annotations.toString());
		
		aspect.getImports().addImport(new PImport(aspect.getFullName(),PImport.SYNTHETIC));
		aspect.getImports().addImport(new PImport(aspect.getName(),PImport.SYNTHETIC));
		
		checkFileName(aspect.getAspectName());
		
		addSyntheticImports(aspect);
		
		aspectdeclarationstring = aspectdeclarationstring + " " + getCurrentImage();
        getNextToken();
        
        this.parseSuperDeclarations(aspect,aspect.getImports());
		
        checkLBRACE(aspectdeclarationstring);
			
        getNextToken();
        parseAspectBody();
       
    }

    private void parseAspectBody()throws JascoParseException
    {	
    	StringBuffer annotations = new StringBuffer();
		int modifiers = parseModifiers(annotations, aspect.getImports());
		
        if(getCurrentId() == HOOKDEFINITION)
        {
            parseCutpointDeclaration(annotations,modifiers);
            parseAspectBody();
        } 
		else 
		{
        	if(getCurrentId() != RBRACE)
        	{
            	parseJavaCode(aspect,aspect.getImports(),annotations,modifiers);
            	parseAspectBody();
        	}
		}
    } 
	

	
	private void parseCutpointDeclaration(StringBuffer annotations, int modifiers) throws JascoParseException
    {
        tempcutpoint = new PCutpoint();
        
        modifiers=this.setPublic(modifiers);
        
        tempcutpoint.setAnnotations(annotations.toString());
		
		String cutpointdeclarationstring = getCurrentImage();
        getNextToken();
		
		
        checkIDENTIFIER(cutpointdeclarationstring);
        tempcutpoint.setCutpointName(getCurrentImage());
        tempcutpoint.setLine(getPreviousLine());
		
		cutpointdeclarationstring = cutpointdeclarationstring + " " + getCurrentImage();
        getNextToken();
        
        if(getCurrentId()==EXTENDS) {
			getNextToken();
			checkIDENTIFIER("super class name expected for"+tempcutpoint.getName());
			String next = getCurrentImage();
			getNextToken();
			String type = accumulateType(aspect.getName()+"."+next,aspect.getImports(),TYPE_ASPECT);
			tempcutpoint.setSuperClass(type);
		}
		
        this.parseSuperDeclarations(tempcutpoint,aspect.getImports()); //only for interfaces
        
		checkLBRACE(cutpointdeclarationstring);
        getNextToken();
			
        parseCutpointBody();
        if(!aspect.addCutpoint(tempcutpoint))
        	throw new JascoParseException(this.newError("Hook name already defined: "+tempcutpoint.getCutpointName()));
        tempcutpoint = null;
      
    }

    private void parseCutpointBody() throws JascoParseException
    {
        if(getCurrentId() == BEFORE || getCurrentId() == AFTER || getCurrentId() == AROUND)
        {
            parseCutpointAdaptation();
            parseCutpointBody();
        } 
		else 
		{
        	if(getCurrentImage().equals(tempcutpoint.getCutpointName()))
        	{
            	parseCutpointConstructor();
            	parseCutpointBody();
        	}
			else if (getCurrentImage().equals("isApplicable"))  {
				parseApplicableStatement();
				parseCutpointBody();
			}	
        	else 
        	{
        		if(getCurrentId() == RBRACE)
       	 		{ 
            		getNextToken();// TODO Auto-generated method stub
        		} 
				else
        		{
            		parseJavaCode(tempcutpoint,aspect.getImports());
            		parseCutpointBody();
        		}
    		}
		}
    }
	
	private void parseApplicableStatement() throws JascoParseException //no adaptation parameter are parsed for the moment
    {
        PJavaCode code = new PJavaCode();    
        int adaptationtype = getCurrentId();
        String cutpointadaptationstring = getCurrentImage();
		 getNextToken();
		 
		 
		String name = null;
		if (getCurrentId()==IDENTIFIER)  {
			cutpointadaptationstring = cutpointadaptationstring + " " + getCurrentImage();
			name=getCurrentImage();
			getNextToken();
		}
		
       
		
		checkLPAREN(cutpointadaptationstring);
		
		cutpointadaptationstring = cutpointadaptationstring + " " + getCurrentImage();
		getNextToken();
		
		checkRPAREN(cutpointadaptationstring);
		
		cutpointadaptationstring = cutpointadaptationstring + " " + getCurrentImage();
		getNextToken();
		
		checkLBRACE(cutpointadaptationstring);
		
		getNextToken();
		
		String temp = "";
		int numberofbraces = 1;
        int cachedlinenumber = getLine();
        
		while(numberofbraces != 0) 
        {
			String errorstring = " '}' or ';' expected";
            CompileError error = new CompileError(getFileName(), getPreviousLine(), errorstring);
            assertNotToken(getCurrentId(), HOOKDEFINITION , error);
			assertNotToken(getCurrentId(), BEFORE , error);
			assertNotToken(getCurrentId(), AFTER , error);
			assertNotToken(getCurrentId(), AROUND , error);	
            assertNotToken(getCurrentId(), EOF, error);	
			
            if (getCurrentId() == LBRACE) 
			{
                numberofbraces++;
			}
			
			if (getCurrentId() == RBRACE) 
			{
                numberofbraces--;
			}
			
			if (numberofbraces != 0)  
			{
				temp = temp + getCurrentImage() + " ";
			}
			
            getNextToken();
			
            if(cachedlinenumber != getLine())
            {
               code.addJavaCode(temp, cachedlinenumber);
               cachedlinenumber = getLine();
               temp = "";
            }
        }
		
        PCutpointTest adap = new PCutpointTest(getPreviousLine());
        adap.setLine(this.getPreviousLine());
		adap.setName(name);
        adap.setImplementation(code);
		tempcutpoint.addTest(adap);
        
    }

    private void parseCutpointAdaptation() throws JascoParseException //no adaptation parameter are parsed for the moment
    {
        PJavaCode code = new PJavaCode();    
        int adaptationtype = getCurrentId();
		
		String cutpointadaptationstring = getCurrentImage();
        getNextToken();
		
        int line = this.getPreviousLine();
        
		boolean after_throwing = false;
		boolean after_returning = false;
		boolean around_throwing = false;
		boolean around_returning = false;
		String name = null;
		if (getCurrentId()==IDENTIFIER||getCurrentId()==THROWING||getCurrentId()==RETURNING||getCurrentId()==COMPLEMENT)  {
			cutpointadaptationstring = cutpointadaptationstring + " " + getCurrentImage();
			name=getCurrentImage();
			
			if ((getCurrentId()==THROWING||getCurrentId()==RETURNING)&&(adaptationtype==AFTER||adaptationtype==AROUND))  {
				
				
				if(adaptationtype==AFTER) {
					if(getCurrentId()==THROWING)
						after_throwing=true;
					else if(getCurrentId()==RETURNING) 
						after_returning=true;
				}
				else if(adaptationtype==AROUND) {
					if(getCurrentId()==THROWING)
						around_throwing=true;
					else if(getCurrentId()==RETURNING) 
						around_returning=true;
				}
			
				getNextToken();
				
				name=null;
				
				if (getCurrentId()==IDENTIFIER||getCurrentId()==COMPLEMENT)  {
					cutpointadaptationstring = cutpointadaptationstring + " " + getCurrentImage();
					name=getCurrentImage();
					getNextToken();
				}
			}
			else getNextToken();
		}
		
		checkLPAREN(cutpointadaptationstring);
		cutpointadaptationstring = cutpointadaptationstring + " " + getCurrentImage();
		getNextToken();
		
		String exceptionType = "";
		String exceptionArg = ""; 
		if (after_throwing||after_returning||around_throwing||around_returning)  {
			exceptionType=accumulateType(aspect.getImports());
			cutpointadaptationstring+=" "+exceptionType;
			checkIDENTIFIER(cutpointadaptationstring);
			exceptionArg+=getCurrentImage();
			getNextToken();
		
			if (exceptionType==null)  {
				CompileError error = new CompileError(getFileName(), getPreviousLine(), "class not found: "+exceptionType);
				throw new JascoParseException(error);
			}
		}
		
		checkRPAREN(cutpointadaptationstring);
		
		cutpointadaptationstring = cutpointadaptationstring + " " + getCurrentImage();
		getNextToken();
		
		checkLBRACE(cutpointadaptationstring);
		
		getNextToken();
		
		PJavaCode jcode = parseMethodBody();
		
        PCutpointAdaptation adap = new PCutpointAdaptation(getPreviousLine());
        adap.setCutpointAdaptationImplementation(jcode);
		adap.setName(name);
		adap.setLine(line);
		
		if (after_throwing)  {
			adap.setArgName(exceptionArg);
	        tempcutpoint.addAfterThrowingCutpointAdaptation(exceptionType,adap);
		}
		else if (after_returning)  {
			adap.setArgName(exceptionArg);
	        tempcutpoint.addAfterReturningCutpointAdaptation(exceptionType,adap);
		}
		else if (around_throwing)  {
			adap.setArgName(exceptionArg);
	        tempcutpoint.addAroundThrowingCutpointAdaptation(exceptionType,adap);
		}
		else if (around_returning)  {
			adap.setArgName(exceptionArg);
	        tempcutpoint.addAroundReturningCutpointAdaptation(exceptionType,adap);
		}
		else if (adaptationtype == BEFORE) 
		{
            tempcutpoint.addBeforeCutpointAdaptation(adap);
		}
        else if (adaptationtype == AFTER) 
		{
            tempcutpoint.addAfterCutpointAdaptation(adap);
		}
        else if (adaptationtype == AROUND) 
		{
            tempcutpoint.addReplaceCutpointAdaptation(adap);
		}
		
    }
	
	protected boolean isMultipleElements(String s)  {
	
		return s.equals("multipleelements");
	}
	
	protected int getNumberTag(PCutpointConstructor constructor, String image) throws JascoParseException {
		
		int x =  constructor.getNumberTag(image);
		if (x==-1)  {
			CompileError error = new CompileError(getFileName(), getPreviousLine(), image +" is not a valid identifier");
			throw new JascoParseException(error);
		}
		return x;
	}	
	
	


    private void parseCutpointConstructor() throws JascoParseException
    {	
        PCutpointConstructor constructor = new PCutpointConstructor();
		
		String constructorstring = getCurrentImage();
		getNextToken();
		
		checkLPAREN(constructorstring);
		
		constructorstring = constructorstring + " " + getCurrentImage();
		getNextToken();
		
		constructor.setLine(getPreviousLine());
		
		while(getCurrentId() != RPAREN) 
        {
	
        
        PCutpointConstructorMethod constructorsignature = new PCutpointConstructorMethod();
		
		String returnType = getCurrentImage();
		getNextToken();
		
		String methodName = returnType;
		
		if (getCurrentId() != LPAREN)  { //return type specified
		    
			returnType=accumulateType(returnType,aspect.getImports());
			 
			constructorstring = constructorstring + " " + returnType;
			
		    methodName = getCurrentImage();
			checkIDENTIFIER(constructorstring);
	        constructorstring = constructorstring + " " + getCurrentImage();
        	getNextToken();
			constructorsignature.setReturnType(returnType);
		}
		else constructorstring = constructorstring + " " + methodName;
		constructorsignature.setCutpointConstructorMethodName(methodName);
		
      	checkLPAREN(constructorstring);
		
   	    constructorstring = constructorstring + " " + getCurrentImage();
        getNextToken();
		
		
		
        while(getCurrentId() != RPAREN) 
        {	
            String errorstring = " '..' or type expected" + "\n\t" + constructorstring + "\n\t" + generatePointer(constructorstring.length());
            CompileError error = new CompileError(getFileName(), getPreviousLine(), errorstring);
            //assertOrToken(getCurrentId(), MULTIPLEELEMENTS , IDENTIFIER , error);
			
            StringBuffer annotations = new StringBuffer();
            this.parseModifiers(annotations,aspect.getImports());
			  
            if( getCurrentId() == MULTIPLEELEMENTS)
            {	
				constructorstring = constructorstring + getCurrentImage();
                getNextToken();
                checkIDENTIFIER(constructorstring);
                constructorsignature.addArgument(getCurrentImage(), "multipleelements",annotations.toString());
            }
			else 
            {	
                String type =accumulateType(aspect.getImports());
				
				constructorstring = constructorstring + type;
				
                checkIDENTIFIER(constructorstring);
                constructorsignature.addArgument(getCurrentImage(), type,annotations.toString());
				
				checkClass(type,aspect.getImports());
            }
            constructorstring = constructorstring + " " + getCurrentImage();
            getNextToken();
			
            errorstring = " ',' or ')' expected" + "\n\t" + constructorstring + "\n\t" + generatePointer(constructorstring.length());
            error = new CompileError(getFileName(), getPreviousLine(), errorstring);
			
            assertOrToken(getCurrentId(), COMMA , RPAREN , error);
            if (getCurrentId() == COMMA) 
			{
                getNextToken();
			}
        }
		
        constructorstring = constructorstring + " " + getCurrentImage();
        getNextToken();
		
		if (getCurrentId() == COMMA)  {
			constructorstring = constructorstring + " " + getCurrentImage();
			getNextToken();
		}
		
		constructor.addCutpointConstructorMethod(constructorsignature);
		
        }
			
		constructorstring = constructorstring + " " + getCurrentImage();
        getNextToken();
		
        checkLBRACE(constructorstring);
		
        constructorstring = "";
        getNextToken();
		
		StringBuffer constructorStringBuffer = new StringBuffer();
		
        PCutpointConstructorApplicationDesignator firstParam = parseConstructorBody(constructor,constructorStringBuffer);
					
       	constructorstring=constructorStringBuffer.toString();
		constructor.setConstructorBody(constructorstring);


		tempcutpoint.setCutpointConstructor(constructor);
		constructor.setConstructorApplicationDesignator(firstParam);		
        constructorstring = constructorstring + " " + getCurrentImage();
		
        checkRBRACE(constructorstring);
		
        getNextToken();
		
    }
    
    protected PCutpointConstructorApplicationDesignator parseStrictRegularPointcut(PCutpointConstructor constructor, StringBuffer constructorstring) throws JascoParseException {
    	
    	constructorstring.append(" "+getCurrentImage());
    	getNextToken();
    	
    	boolean strictcomplement =false;
    	
    	if(getCurrentId()==COMPLEMENT) {
    		constructorstring.append(" "+getCurrentImage());
        	getNextToken();
        	strictcomplement=true;
    	}
    	
    	PCutpointConstructorApplicationDesignator context = null;
    	
    	if(getCurrentId()==COLON) { //matchall context
    		
    		context =  new PCutpointConstructorMatchAll();
    	}
    	else {
    		checkLBRACKET(constructorstring.toString());
    	
    		constructorstring.append(" "+getCurrentImage());
    		getNextToken();
    	
    		context =  parseAbstractPointcut(constructor,constructorstring);
    	
    		checkRBRACKET(constructorstring.toString());
    		constructorstring.append(" "+getCurrentImage());
    		getNextToken();
    	}
    	
    	PCutpointConstructorRegular strict = null ;
    	if(strictcomplement)
    		strict=new PCutpointConstructorStrictComplement(context);
    	else strict = new PCutpointConstructorStrict(context);
    	
    	constructorstring.append(" "+getCurrentImage());
    	checkCOLON(constructorstring.toString());
    	getNextToken();
    	
    	parseRegularPointcut(constructor,constructorstring,strict);
    	
    	return strict;
    }
    
    protected PCutpointConstructorApplicationDesignator parseComplementRegularPointcut(PCutpointConstructor constructor, StringBuffer constructorstring) throws JascoParseException {
    	
    	constructorstring.append(" "+getCurrentImage());
    	getNextToken();
    	
    	boolean strictcomplement =false;
    	
    	if(getCurrentId()==STRICT) {
    		constructorstring.append(" "+getCurrentImage());
        	getNextToken();
        	strictcomplement=true;
    	}
    	
    	PCutpointConstructorApplicationDesignator context = null;
    	if(getCurrentId()==COLON) { //matchall context
    		
    		context =  new PCutpointConstructorMatchAll();
    	}
    	else {
    		checkLBRACKET(constructorstring.toString());
    	
    		constructorstring.append(" "+getCurrentImage());
    		getNextToken();
    	
    		context = parseAbstractPointcut(constructor,constructorstring);
    	
    		checkRBRACKET(constructorstring.toString());
    		constructorstring.append(" "+getCurrentImage());
    		getNextToken();
    	}
    	
    	PCutpointConstructorRegular compl=null; 
    	if(strictcomplement)
    		compl=new PCutpointConstructorStrictComplement(context);
    	else compl= new PCutpointConstructorComplement(context);
    	
    	constructorstring.append(" "+getCurrentImage());
    	checkCOLON(constructorstring.toString());
    	getNextToken();
    	
    	parseRegularPointcut(constructor,constructorstring,compl);
    	
    	return compl;
    }
	
	protected PCutpointConstructorApplicationDesignator parseConstructorBody(PCutpointConstructor constructor, StringBuffer constructorstring) throws JascoParseException  {
		
		if (getCurrentId()==EXECUTE||getCurrentId()==CALL)   { //this means without regular expressions
			PCutpointConstructorApplicationDesignator des = parseAbstractPointcut(constructor,constructorstring);
			checkSEMICOLON(constructorstring.toString());
			getNextToken();
			return des;
		} 
		else if (getCurrentId()==COMPLEMENT)  { //this means with complement regular expressions 
			return parseComplementRegularPointcut(constructor,constructorstring);
		} 
		else if (getCurrentId()==STRICT)  { //this means with strict regular expressions 
			return parseStrictRegularPointcut(constructor,constructorstring);
		}
		else if (getCurrentId()==IDENTIFIER)  { //this means with regular expressions 
			if (getCurrentImage().equals("execute"))   //backward syntax check, issue warning
					Logger.getInstance().showWarning("'execute' keyword changed to 'execution'");
			return parseRegularPointcut(constructor,constructorstring,new PCutpointConstructorRegular());
	
		}
		else  {
				String errorstring = " 'execution' or 'call' or identifier for regular expression expected\n\t";
    		    CompileError error = new CompileError(getFileName(), getPreviousLine(), errorstring);
				throw new JascoParseException(error);
		}
	}
	
	protected PCutpointConstructorApplicationDesignator parseRegularPointcut(PCutpointConstructor constructor,StringBuffer constructorstring, PCutpointConstructorRegular regularConstructor) throws JascoParseException {
		
		
		
		TreeMap stmap = new TreeMap();
		PRegularConstructorStatement first = null;
		
		PRegularConstructorStatement startTrans = null;
		
		while (getCurrentId()!=RBRACE)  {
			PRegularConstructorStatement res = parseRegularPointcutStmnt(constructor,constructorstring);
			if(first==null)
				first=res;
			if (res.getName().equals(PRegularConstructorStatement.START)&&startTrans!=null)  {
				String errorstring = "Only one start definition is possible in constructor body: \n\t";
    		    CompileError error = new CompileError(getFileName(), getPreviousLine(), errorstring);
				throw new JascoParseException(error);
			}
			else if (res.getName().equals(PRegularConstructorStatement.START))
				startTrans=res;	
			stmap.put(res.getName(),res);
		}
		State start = new State("START");
		regularConstructor.setStartState(start);
		
		if (startTrans==null)  {
			Vector next = new Vector();
			next.add(first.getName());
			PRegularConstructorStatement newStart = new PRegularConstructorStatement(PRegularConstructorStatement.START,null,next);
			startTrans=newStart;
		}
		
		Iterator startTransI = startTrans.getNext();
		TreeMap stateMap = new TreeMap();
		while (startTransI.hasNext()) 	 {
			String next = (String) 	startTransI.next();
			PRegularConstructorStatement stmt = (PRegularConstructorStatement)stmap.get(next);
			if (stmt == null)  {
				String errorstring = "Undefined destination pointcut in constructor body: "+next+"\n\t";
    		    CompileError error = new CompileError(getFileName(), getPreviousLine(), errorstring);
				throw new JascoParseException(error);
			}
			transformToStateMachine(stmt,start,stmap,stateMap,regularConstructor);
		}
		
		regularConstructor.setStatementMap(stmap);
		return regularConstructor;
	}
	
	protected void transformToStateMachine(PRegularConstructorStatement current, State prev, TreeMap statementMap, TreeMap stateMap, PCutpointConstructorRegular regularConstructor) throws JascoParseException {
		
		State nextState = (State)stateMap.get(current.getName());
		boolean nextStateDone=false;
		if (nextState==null)  {
			nextState = new State(current.getName());
			stateMap.put(nextState.getLabel(),nextState);
			regularConstructor.addState(nextState);
		}
		else nextStateDone=true;
		regularConstructor.addTransition(prev,current.getID(),current.getPointcut(),nextState);
		
		if(nextStateDone)
			return;
		
		Iterator nextI = current.getNext();
		while (nextI.hasNext())  {
			String next = (String) nextI.next();
			PRegularConstructorStatement nextStatement = (PRegularConstructorStatement)statementMap.get(next);
			if (nextStatement == null)  {
				String errorstring = "Undefined destination pointcut in constructor body: "+next+"\n\t";
    		    CompileError error = new CompileError(getFileName(), getPreviousLine(), errorstring);
				throw new JascoParseException(error);
			}
			transformToStateMachine(nextStatement,nextState,statementMap,stateMap,regularConstructor);
		}
		
	}
	
	protected PRegularConstructorStatement parseRegularPointcutStmnt(PCutpointConstructor constructor,StringBuffer constructorstring) throws JascoParseException {
		
		String name=getCurrentImage();
		
		if (name.equals(PRegularConstructorStatement.START))  {
			getNextToken();
			checkGT(constructorstring.toString());
			constructorstring.append(" "+getCurrentImage());
			
			Vector next = parseRegularConstructorStatementNext(constructorstring);
			return new PRegularConstructorStatement(name,null,next);
		}
		checkIDENTIFIER(constructorstring.toString());
		constructorstring.append(" "+name);
		
		getNextToken();
		checkCOLON(constructorstring.toString());
		constructorstring.append(" "+getCurrentImage());
		
		getNextToken();
		PCutpointConstructorApplicationDesignator designator = parseAbstractPointcut(constructor,constructorstring);
		
		String errorstring = " > or ; expected\n\t";
        CompileError error = new CompileError(getFileName(), getPreviousLine(), errorstring);
        assertOrToken(getCurrentId(), GT, SEMICOLON,  error);
		
		Vector next = new Vector();
		if (getCurrentId()==SEMICOLON)  {
			getNextToken();
		}
		else if (getCurrentId()==GT)  {
			next = parseRegularConstructorStatementNext(constructorstring);
		}
		 
		return new PRegularConstructorStatement(name,designator,next);
	}
	
	protected Vector parseRegularConstructorStatementNext(StringBuffer constructorstring) throws JascoParseException {
	
		Vector next=new Vector();
		constructorstring.append(" "+getCurrentImage());
		getNextToken();
		constructorstring.append(" "+getCurrentImage());
		checkIDENTIFIER(constructorstring.toString());
			
		next.add(getCurrentImage());
		getNextToken();
			
		while (getCurrentId()!=SEMICOLON&&getCurrentId()!=EOF)  {
			constructorstring.append(" "+getCurrentImage());
			String errorstring="; or || expected\n\t";
			CompileError error = new CompileError(getFileName(), getPreviousLine(), errorstring);
			assertToken(getCurrentId(), SC_OR, error);
				
			getNextToken();
			constructorstring.append(" "+getCurrentImage());
			checkIDENTIFIER(constructorstring.toString());
			next.add(getCurrentImage());
			getNextToken();
		}
		checkSEMICOLON(constructorstring.toString());
		
		constructorstring.append(" "+getCurrentImage());
		getNextToken();
		return next;
	}
	
	
	protected PCutpointConstructorApplicationDesignator parseAbstractPointcut(PCutpointConstructor constructor,StringBuffer constructorstring) throws JascoParseException {
	
	 	String errorstring = " 'execution' expected\n\t";
        CompileError error = new CompileError(getFileName(), getPreviousLine(), errorstring);
        assertOrToken(getCurrentId(), EXECUTE, CALL, error);
		
        boolean execute = true;
        if(getCurrentId()==CALL)
        	execute=false;
        
        constructorstring.append(" " + getCurrentImage());
		getNextToken();
		
        checkLPAREN(constructorstring.toString());
		
        constructorstring.append(" " + getCurrentImage());
        getNextToken();
		
		boolean fix=false;
		if (getCurrentId()==PLUS)  {
			fix=true;
			getNextToken();
			constructorstring.append(" " + getCurrentImage());
		}
		
        checkIDENTIFIER(constructorstring.toString());
		
		int numberTag = getNumberTag(constructor,getCurrentImage());
		
		PCutpointConstructorApplicationDesignator firstParam = null ;
		if(execute)
			firstParam=new PCutpointConstructorExecute(numberTag+"",fix);
		else firstParam=new PCutpointConstructorCall(numberTag+"");
		
		PCutpointConstructorArgs args = new PCutpointConstructorArgs(null,numberTag);
		PCutpointConstructorAnd result = new PCutpointConstructorAnd(args, firstParam);
		 
        constructorstring.append(" " + getCurrentImage());
        getNextToken();
		
        checkRPAREN(constructorstring.toString());
		 
        constructorstring.append(" " + getCurrentImage());
		
		getNextToken();
		
		firstParam=createDesignatorBool(result,constructorstring,constructor);
		
		return firstParam;
	}
	
	protected PCutpointConstructorApplicationDesignator createDesignatorBool(PCutpointConstructorApplicationDesignator prev, StringBuffer log, PCutpointConstructor constructor)  throws JascoParseException{
	
		int currentId = getCurrentId();
		log.append(" "+getCurrentImage());
		switch (currentId)  {
		
			case SC_AND: {
				getNextToken();
				PCutpointConstructorApplicationDesignator secondParam = createDesignatorAll(log,constructor);
				PCutpointConstructorApplicationDesignator and = new PCutpointConstructorAnd(prev,secondParam);
				return createDesignatorBool(and,log,constructor);
			}
			case SC_OR: {
				getNextToken();
				PCutpointConstructorApplicationDesignator secondParam = createDesignatorAll(log,constructor);
				PCutpointConstructorApplicationDesignator and = new PCutpointConstructorOr(prev,secondParam);
				return createDesignatorBool(and,log,constructor);
			}
			case RPAREN:
				getNextToken();
				return prev;
			
			case SEMICOLON:
			case GT:
		    case RBRACKET:
				return prev;
			
			default:
				String errorstring = " illegal boolean expression \n\t";
        		CompileError error = new CompileError(getFileName(), getPreviousLine(), errorstring);
				throw new JascoParseException(error);
				
		}
	}
	
	protected PCutpointConstructorApplicationDesignator createDesignatorAll(StringBuffer log, PCutpointConstructor constructor) throws JascoParseException {
		
		int currentId = getCurrentId();
		
		switch (currentId)  {
		
			case BANG:{
				log.append(" "+getCurrentImage());
				getNextToken();
				PCutpointConstructorApplicationDesignator secondParam = createDesignatorAll(log,constructor);
				return new PCutpointConstructorNot(secondParam);
			}	
				
			case LPAREN: {
				log.append(" "+getCurrentImage());
				getNextToken();
				PCutpointConstructorApplicationDesignator secondParam = createDesignatorAll(log,constructor);
				return createDesignatorBool(secondParam,log,constructor);
			}
			
			case TARGET: {
				
				log.append(" "+getCurrentImage());
				getNextToken();
				log.append(" "+getCurrentImage());
				checkLPAREN(log.toString());
				getNextToken();
				String type = accumulateType(aspect.getImports());
				log.append(" "+type);
				PCutpointConstructorApplicationDesignator secondParam = new PCutpointConstructorTarget(type);
				log.append(" "+getCurrentImage());
				checkRPAREN(log.toString());
				getNextToken();
				return secondParam;
			}
				
			
			case CFLOW: {
				
				log.append(" "+getCurrentImage());
				getNextToken();
				log.append(" "+getCurrentImage());
				checkLPAREN(log.toString());
				getNextToken();
				log.append(" "+getCurrentImage());
				PCutpointConstructorApplicationDesignator secondParam = new PCutpointConstructorCflow(getNumberTag(constructor,getCurrentImage())+"");
				getNextToken();
				log.append(" "+getCurrentImage());
				checkRPAREN(log.toString());
				getNextToken();
				return secondParam;
			}
			
			case WITHINCODE: {
				log.append(" "+getCurrentImage());
				getNextToken();
				checkLPAREN(log.toString());
				getNextToken();
				log.append(" "+getCurrentImage());
				PCutpointConstructorApplicationDesignator secondParam = new PCutpointConstructorWithincode(getNumberTag(constructor,getCurrentImage())+"");
				getNextToken();
				checkRPAREN(log.toString());
				getNextToken();
				return secondParam;
			}
			
			case EXECUTE: {
				log.append(" "+getCurrentImage());
				getNextToken();
				boolean fix=false;
				if (getCurrentId()==PLUS)  { 
					fix=true;
					log.append(" "+getCurrentImage());
					getNextToken();
				}
				checkLPAREN(log.toString());
				getNextToken();
				log.append(" "+getCurrentImage());
				PCutpointConstructorApplicationDesignator secondParam = new PCutpointConstructorExecute(getNumberTag(constructor,getCurrentImage())+"",fix);
				getNextToken();
				checkRPAREN(log.toString());
				getNextToken();
				return secondParam;
			}
			
			case EXECUTIONHOST: {
				log.append(" "+getCurrentImage());
				getNextToken();
				log.append(" "+getCurrentImage());
				checkLPAREN(log.toString());
				getNextToken();
				log.append(" "+getCurrentImage());
				String host = parseHost(log);
				getNextToken();
				log.append(" "+getCurrentImage());
				
				String strategy = null;
				if(this.getCurrentId()==COMMA) {
					
					getNextToken();
					strategy = this.accumulateType(aspect.getImports());
					checkHostSelectionStrategy(strategy);
				}
				
				PCutpointConstructorApplicationDesignator secondParam;
				if(host.equals(JOINPOINTHOST_STR))
					secondParam = new PCutpointConstructorExecutionEqualsJoinpointHost();
				else secondParam = new PCutpointConstructorExecutionHost(host,strategy);
				checkRPAREN(log.toString());
				getNextToken();
				return secondParam;
			} 
			
			case JOINPOINTHOST: {
				log.append(" "+getCurrentImage());
				getNextToken();
				checkLPAREN(log.toString());
				getNextToken();
				String host = parseHost(log);
				PCutpointConstructorApplicationDesignator secondParam = new PCutpointConstructorPHost(host);
				getNextToken();
				checkRPAREN(log.toString());
				getNextToken();
				return secondParam;
			}
			
			default:
				String errorstring = " illegal boolean expression \n\t";
        		CompileError error = new CompileError(getFileName(), getPreviousLine(), errorstring);
				throw new JascoParseException(error);
			
				
		}
	}
	
	private void checkHostSelectionStrategy(String strategy) {
		try {
			Class clz = this.loadClassOrPrim(strategy);
			if(!IHostSelectionStrategy.class.isAssignableFrom(clz))
				throw new JascoParseException(newError("Supplied strategy does not implement "+IHostSelectionStrategy.class.getName()+": "+strategy ));
			
		} catch (ClassNotFoundException e) {
			throw new JascoParseException(newError("Class bot found: "+strategy));
		}
	}


	private String parseHost(StringBuffer log) {
		
		String host;
		
		switch(getCurrentId()) {
		
			case STRING_LITERAL:
				host = removeQuotes(getCurrentImage());
				log.append(" "+getCurrentImage());
				break;
			
			case IDENTIFIER:
				host = getCurrentImage();
				log.append(" "+getCurrentImage());
				break;
				
			case JOINPOINTHOST:
				host= JOINPOINTHOST_STR;
				break;
				
				default:
					throw new JascoParseException(newError("string or identifier expected for host"));
		}
		
		return host;
	}


	private String removeQuotes(String currentImage) {

		if(currentImage.charAt(0)=='"')
			currentImage = currentImage.substring(1);
		if(currentImage.charAt(currentImage.length()-1)=='"')
			currentImage = currentImage.substring(0,currentImage.length()-1);
		
		return currentImage;
	}
		
				
			
		
		
			
			
			
			

}
