/*
 * Created on 9-mrt-2005
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package jasco.tools.aspectparser;

import jasco.util.logging.Logger;

import java.util.Iterator;

import javassist.CannotCompileException;
import javassist.ClassPool;
import javassist.CtClass;
import javassist.CtMethod;
import javassist.NotFoundException;
import javassist.bytecode.AnnotationsAttribute;
import javassist.bytecode.annotation.Annotation;
import javassist.expr.MethodCall;

/* 
 Copyright (C) 2002-2005 JAsCo Development Team
 http://ssel.vub.ac.be/jasco
 
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as
 published by the Free Software Foundation; either version 2 of the
 License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 USA 
 */

/**
 * @author Wim
 * 
 * only works when aspect is compiled using Java 1.5! 
 */

public class RefinableMethodsInvocationChecker {

	private final ClassPool pool = new ClassPool(null); //each time new classPool for new receiving CtClass
	private final String classname;
	
	public RefinableMethodsInvocationChecker(String classname, String saveDir) {
		super();
		
		try {
			initPool(saveDir);
		} catch (NotFoundException e) {
			Logger.getInstance().showError(e);
		}
		this.classname=classname;
	}
	
	private void initPool(String saveDir) throws NotFoundException {
		
		pool.appendSystemPath();
		Iterator i = jasco.options.Options.getClassPathVector().iterator();
		while (i.hasNext())  {
			String s = (String) i.next();
			
			pool.appendClassPath(s);
		}	
		pool.insertClassPath(saveDir);
	}
	/**
	 * 
	 * @return name of method with illegal invocation or null if ok
	 * @throws NotFoundException
	 */
	public String check() throws NotFoundException {
		
		CtClass clz = pool.get(classname);
		
		CtMethod[] methods = clz.getMethods();
		for(int i=0;i<methods.length;++i) {
			if(isAdvice(methods[i]))
				continue;
			if(hasInvocation(methods[i]))
				return methods[i].getName();
		}
			
		return null;
	}
	
	/**
	 * @param method
	 * @return
	 */
	private boolean isAdvice(CtMethod method) {
		return hasAnnotation(method,AspectClassesGenerator.ADVICE_ANNOTATION);
	}

	/**
	 * @param method
	 * @return
	 */
	private boolean hasInvocation(CtMethod method) {
		
		CallEditor editor = new CallEditor();
		try {
			method.instrument(editor);
		} catch (CannotCompileException e) {
			
			Logger.getInstance().showError(e);
		}
		if(editor.getResult())
			return true;
		else return false;
	}
	
	/*
	 * 
	 * bug in javassist: no method has annotations
	 */
	private boolean hasAnnotation(CtMethod method, String name) {
		
		System.out.println(method.getName());
		AnnotationsAttribute annot;
		
		annot = (AnnotationsAttribute) method.getMethodInfo().getAttribute(AnnotationsAttribute.invisibleTag);
		if(annot==null)
			return false;
		
		Annotation[] annotations = annot.getAnnotations();
		if(annotations==null)
			return false;
    	for(int i=0;i<annotations.length;++i)
    		if(annotations[i].getTypeName().equals(name)) {
    			System.out.println(name);
    			return true;
    		}
		
		return false;
	}

	class CallEditor extends javassist.expr.ExprEditor {
		
		private boolean result = false;
		
		public void edit(MethodCall m) throws CannotCompileException {
	        
			try {
				if(hasAnnotation(m.getMethod(),AspectClassesGenerator.REFINABLE_ANNOTATION))
						result=true;
			} catch (NotFoundException e) {
			
				Logger.getInstance().showError(e);
			}	
	        	
	     }
		
		public boolean getResult() {
			
			return result;
		}
	}
}
