/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.tools.connectorparser;

import jasco.options.Options;
import jasco.runtime.aspect.ConstructorMethod;
import jasco.runtime.aspect.IHook;
import jasco.runtime.aspect.factory.IAspectFactory;
import jasco.runtime.aspect.factory.PerAllAspectFactory;
import jasco.runtime.aspect.factory.PerClassAspectFactory;
import jasco.runtime.aspect.factory.PerMethodAspectFactory;
import jasco.runtime.aspect.factory.PerObjectAspectFactory;
import jasco.runtime.aspect.factory.PerThreadAspectFactory;
import jasco.runtime.connector.MethodSignature;
import jasco.tools.jascoparser.JascoParseException;
import jasco.tools.jascoparser.JascoParser;
import jasco.tools.jascoparser.PJavaClass;
import jasco.tools.jascoparser.PJavaCode;
import jasco.tools.jascoparser.PMethod;
import jasco.util.javacompiler.CompileError;
import jasco.util.logging.Logger;

import java.util.Iterator;
import java.util.Vector;

public class ConnectorParser extends JascoParser {
	private PConnector connector;
	
	private PCutpointDeclaration tempcutpointdeclaration;
	

	private PCutpointDeclarationParameter tempcutpointdeclarationparameter;

	private int cutpointexecutioncounter = 0;

	public ConnectorParser(
			jasco.tools.jascoparser.JascoParserTokenManager aConnectorParserTokenManager,
			String aFileName) {
		super(aConnectorParserTokenManager, aFileName);
		connector = new PConnector();
		tempcutpointdeclaration = new PCutpointDeclaration();
		tempcutpointdeclarationparameter = new PCutpointDeclarationParameter();
	}

	private boolean isHookInitialisation(String aHook) {
		boolean returnvalue = false;
		PCutpointDeclarations declarations = connector
				.getCutpointDeclarations();
		for (int i = 0; i < declarations.getCutpointDeclarationsSize(); i++) {
			PCutpointDeclaration temp = declarations.getCutpointDeclarations(i);
			String name = temp.getCutpointDeclarationName();
			
			if (name.equals(aHook)) {
				returnvalue = true;
			}
		}
		
		return returnvalue;
	}

	public void start() {
		try {
			if(getCurrentId()==PACKAGE)
				throw new JascoParseException(newError("Package definition for connector not allowed (package is always Connector)"));
			connector.setImports(parseImportDeclaration());

			parseConnector();

		} catch (JascoParseException exception) {
			Logger.getInstance().showError(exception.getError());
			Logger.getInstance().showOutput("1 error");
			jasco.options.Options.exit(1);
		}
	}

	private void parseConnector() throws JascoParseException {
		String connectordeclarationstring = "";
		int definedpriority = 0;
		
		while(getCurrentId()==ANNOTATION) {
			StringBuffer annotation = new StringBuffer();
			
			this.parseAnnotation(annotation,connector.getImports());
			String annotationString = annotation.toString();
			connector.addAnnotation(annotationString);
		}
		
		String errorstring = " 'static' or 'connector' expected";
		CompileError error = new CompileError(getFileName(), getPreviousLine(),
				errorstring);
		assertOrToken(getCurrentId(), STATIC, CONNECTOR, error);
		
		if (getCurrentId() == STATIC) {
			connector.setStaticConnector(true);

			connectordeclarationstring = connectordeclarationstring + " "
					+ getCurrentImage();
			getNextToken();
			
		} else {
			connector.setStaticConnector(false);
		}
		
		errorstring = " 'connector expected'" + "\n\t"
			+ connectordeclarationstring + "\n\t"
			+ generatePointer(connectordeclarationstring.length());
		error = new CompileError(getFileName(), getPreviousLine(),errorstring);
		assertToken(getCurrentId(), CONNECTOR, error);	
		
		connectordeclarationstring = connectordeclarationstring + " "
				+ getCurrentImage();
		getNextToken();

		checkIDENTIFIER(connectordeclarationstring);
		connector.setConnectorName(getCurrentImage());
		
		checkFileName(connector.getConnectorName());
 
		connectordeclarationstring = connectordeclarationstring + " "
				+ getCurrentImage();
		getNextToken();
		
		connector.setLine(getPreviousLine());

		if (getCurrentId() == EXTENDS) {
			connectordeclarationstring = connectordeclarationstring + " "
					+ getCurrentImage();
			getNextToken();

			checkIDENTIFIER(connectordeclarationstring);
			connector.setExtendingConnectorName(getCurrentImage());

			connectordeclarationstring = connectordeclarationstring + " "
					+ getCurrentImage();
			getNextToken();

			checkLBRACE(connectordeclarationstring);

			getNextToken();
			parseConnectorBody();
		} else {
			checkLBRACE(connectordeclarationstring);

			getNextToken();
			parseConnectorBody();
		}
	}

	private void parseConnectorBody() throws JascoParseException {
		String aspectFactory = null;

		if (getCurrentId() != RBRACE) {

			if (getCurrentId() == LBRACE) {
				parseBlockedJavaCode();
				parseConnectorBody();
				return;
			} else if (getCurrentId() == PEROBJECT) {
				aspectFactory = PerObjectAspectFactory.class.getName();
				getNextToken();
			}else if (getCurrentId() == PER) {
				getNextToken();
				checkLPAREN("per");
				getNextToken();
				String type = accumulateType(this.connector.getImports());
				checkRPAREN("per("+type);
				aspectFactory = type;
				getNextToken();
			} 
			else if (getCurrentId() == PERALL) {
				aspectFactory = PerAllAspectFactory.class.getName();
				getNextToken();
			} else if (getCurrentId() == PERTHREAD) {
				aspectFactory = PerThreadAspectFactory.class.getName();
				getNextToken();
			} else if (getCurrentId() == PERCLASS) {
				aspectFactory = PerClassAspectFactory.class.getName();
				getNextToken();
			} else if (getCurrentId() == PERMETHOD) {
				aspectFactory = PerMethodAspectFactory.class.getName();
				getNextToken();
			} else if (getCurrentImage().equals("addCombinationStrategy")) {
				parseCombinationStrategy("");
				return;
			} else if (getCurrentId() == SEMICOLON) {
				getNextToken();
				parseConnectorBody();
				return;
			}

			Vector identifiers = new Vector();

			String connectorbodystring = getCurrentImage();
			String token1string = getCurrentImage();
			if (getCurrentId() == FINAL) {

				getNextToken();
				connectorbodystring = connectorbodystring + " "
						+ getCurrentImage();
			}

			checkTYPE(token1string);

			identifiers.add(getCurrentImage());

			getNextToken();
			connectorbodystring = connectorbodystring + " " + getCurrentImage();

			while (getCurrentId() == DOT) {
				getNextToken();
				connectorbodystring = connectorbodystring + " "
						+ getCurrentImage();

				identifiers.add(getCurrentImage());

				getNextToken();
				connectorbodystring = connectorbodystring + " "
						+ getCurrentImage();
			}

			int type = checkType(identifiers);

			switch (type) {

			case REGULAR_JAVA:

				getNextToken();
				parseJavaCode(connectorbodystring);
				break;

			case CRE_COMBINATION_STRATEGY:
				getNextToken();
				parseCombinationStrategy(connectorbodystring);
				break;

			case HOOK_DECLARATION:

				parseHookInitialisation(identifiers, connectorbodystring,
						aspectFactory);
				break;
				
			case ASPECT_DECLARATION:

				parseAspectInitialisation(identifiers, connectorbodystring,
						aspectFactory);
				break;

			case METHOD_EXECUTION:

				String declaration = (String) identifiers.elementAt(0);
				String kind = (String) identifiers.elementAt(1);

				if ((kind.equals("before")) || (kind.equals("after"))
						|| (kind.equals(IHook.REPLACE)) || (kind.equals("default"))) {

					parseCutpointExecution(declaration, kind,
							connectorbodystring);
				}

				else {
					parseCutpointMethodExecution(declaration, kind,
							connectorbodystring);
				}

				break;

			default:
				throw new IllegalArgumentException("unknown type " + type
						+ " in parseConnectorBody!");
			}

		}
		getNextToken();
		String errorstring = "code past end of the connector: "
				+ getCurrentImage() + "!";
		CompileError error = new CompileError(getFileName(), getPreviousLine(),
				errorstring);

		if (getCurrentId() != EOF)
			throw new JascoParseException(error);
	}

	private void parseAspectInitialisation(Vector identifiers, String connectorbodystring, String aspectFactory) {
		
		this.checkIDENTIFIER(connectorbodystring);
		
		getNextToken();
		connectorbodystring+=" "+getCurrentImage();
		this.checkASSIGN(connectorbodystring);
		getNextToken();
		connectorbodystring+=" "+getCurrentImage();
		this.checkNEW(connectorbodystring);
		getNextToken();
		
		String instantiationString = "new ";
		
		String type = accumulateType(connector.getImports());
		String firsttype = getType(identifiers);
		if(!type.equals(firsttype))
			throw new JascoParseException(this.newError("Non assignable types: "+firsttype+ " and "+type));
		connectorbodystring+=" "+type;
		instantiationString+=type+" ";
		
		this.checkLPAREN(connectorbodystring);
		connectorbodystring+=" "+getCurrentImage();
		instantiationString+=" "+getCurrentImage();
		getNextToken();
		
		while (getCurrentId() != SEMICOLON) {
			instantiationString += " " + getCurrentImage();
			getNextToken();
			if(this.getCurrentId()==EOF)
				throw new JascoParseException(this.newError("Uexpected end of file")); 
		}
		getNextToken();
		
		if(connector.hasAspectBeanDeclaration(type))
			throw new JascoParseException(this.newError("Aspect Bean "+type+" already defined!"));
		
		PAspectBeanDeclaration decl = new PAspectBeanDeclaration(type,instantiationString);
		connector.addAspectDeclaration(decl);
		
		parseConnectorBody();
	}
	
	private String getType(Vector identifiers) {
		
		String tmp = "";
		Iterator i = identifiers.iterator();
		while(i.hasNext()) {
			tmp+=i.next().toString();
			if(i.hasNext())
				tmp+=".";
		}
		return tmp;
	}

	private void parseBlockedJavaCode() throws JascoParseException {
		PJavaCode code = new PJavaCode();
		int cachedlinenumber = getLine();
		String temp = "";
		code.addJavaCode("{", cachedlinenumber);
		int numberofbraces = 1;
		getNextToken();
		while (numberofbraces != 0) {
			String errorstring = " '}' or ';' expected";
			CompileError error = new CompileError(getFileName(),
					getPreviousLine(), errorstring);
			assertNotToken(getCurrentId(), EOF, error);

			if ((getCurrentId() == IDENTIFIER)
					&& (isHookInitialisation(getCurrentImage()))) {
				temp += "myinstance." + getCurrentImage() + " ";
			} else {
				temp += getCurrentImage() + " ";
			}

			if (getCurrentId() == LBRACE)
				numberofbraces++;

			else if (getCurrentId() == RBRACE)
				numberofbraces--;

			getNextToken();

			if (cachedlinenumber != getLine()) {
				code.addJavaCode(temp, cachedlinenumber);
				cachedlinenumber = getLine();
				temp = "";
			}

		}
		connector.addJavaCode(code);
		parseConnectorBody();
	}

	private void parseJavaCode(String currentJavaCode)
			throws JascoParseException {
		//regular java code not allowed anymore:
		/*String errorStr = "Normal java code can not be used in a connector anymore. A connector should be used solely for instantiating and initilaising hooks and for declaring how they work together.";
		errorStr += " Other logic should be confined to other components.";
		errorStr += " Notice that it is still allowed to initialse a hook by calling a method on it, e.g.: hook.setOutputString(\"BOE!\");";
		errorStr += jasco.util.generators.JavaGenerator.NEWLINE
				+ " If you really have to use normal Java code (DISCOURAGED!), use a local block (using braces {})";
				*/
		String errorStr = "aspect not found error: "+currentJavaCode;
		CompileError error1 = new CompileError(getFileName(), getLine(),
				errorStr);
		throw new JascoParseException(error1);
		/*
		 * PJavaCode code = new PJavaCode(); int cachedlinenumber = getLine();
		 * String temp = currentJavaCode; int numberofbraces = -1; Object while
		 * (numberofbraces != 0) { String errorstring = " '}' or ';' expected";
		 * CompileError error = new CompileError(getFileName(),
		 * getPreviousLine(), errorstring); assertNotToken(getCurrentId(), EOF,
		 * error);
		 * 
		 * if ((getCurrentId() == IDENTIFIER) &&
		 * (isHookInitialisation(getCurrentImage()))) { temp = temp +
		 * "myinstance." + getCurrentImage() + " "; } else { temp = temp +
		 * getCurrentImage() + " "; }
		 * 
		 * if (getCurrentId() == LBRACE) { if (numberofbraces == -1) {
		 * numberofbraces = 1; } else { numberofbraces++; } }
		 * 
		 * if (getCurrentId() == RBRACE) { numberofbraces--; }
		 * 
		 * if ((getCurrentId() == SEMICOLON) && (numberofbraces == -1)) {
		 * numberofbraces = 0; }
		 * 
		 * getNextToken();
		 * 
		 * if(cachedlinenumber != getLine()) { code.addJavaCode(temp,
		 * cachedlinenumber); cachedlinenumber = getLine(); temp = ""; }
		 *  } connector.addJavaCode(code); parseConnectorBody();
		 */
	}

	private void parseCombinationStrategy(String currentJavaCode)
			throws JascoParseException {

		PJavaCode code = new PJavaCode();
		int cachedlinenumber = getLine();
		String temp = currentJavaCode;
		int numberofbraces = -1;

		while (numberofbraces != 0) {
			String errorstring = " '}' or ';' expected";
			CompileError error = new CompileError(getFileName(),
					getPreviousLine(), errorstring);
			assertNotToken(getCurrentId(), EOF, error);

			if ((getCurrentId() == IDENTIFIER)
					&& (isHookInitialisation(getCurrentImage()))) {
				temp = temp + "myinstance." + getCurrentImage() + " ";
			} else {
				temp = temp + getCurrentImage() + " ";
			}

			if (getCurrentId() == LBRACE) {
				if (numberofbraces == -1) {
					numberofbraces = 1;
				} else {
					numberofbraces++;
				}
			}

			if (getCurrentId() == RBRACE) {
				numberofbraces--;
			}

			if ((getCurrentId() == SEMICOLON) && (numberofbraces == -1)) {
				numberofbraces = 0;
			}

			getNextToken();

			if (cachedlinenumber != getLine()) {
				code.addJavaCode(temp, cachedlinenumber);
				cachedlinenumber = getLine();
				temp = "";
			}

		}
		connector.addJavaCode(code);
		parseConnectorBody();

	}

	private void parseCutpointMethodExecution(String cutpointdeclaration,
			String method, String error) throws JascoParseException {
		String cutpointmethodexecutionstring = error;
		String methodexecution = cutpointdeclaration + "." + method + "(";

		getNextToken();

		while (getCurrentId() != SEMICOLON) {
			//methodexecution = methodexecution + getCurrentImage();

			methodexecution = methodexecution + " " + getCurrentImage();
			getNextToken();

			//String errorstring = " ',' or ')' expected" + "\n\t" +
			// cutpointmethodexecutionstring + "\n\t" +
			// generatePointer(cutpointmethodexecutionstring.length());
			//CompileError theerror = new CompileError(getFileName(),
			// getPreviousLine(), errorstring);
			//assertOrToken(getCurrentId(), COMMA , RPAREN , theerror);

			//if (getCurrentId() == COMMA)
			//{
			//     methodexecution = methodexecution + ",";
			//     cutpointmethodexecutionstring = cutpointmethodexecutionstring + "
			// " + getCurrentImage();
			// getNextToken();
			//}
		}

		methodexecution = methodexecution + ";";

		//cutpointmethodexecutionstring = cutpointmethodexecutionstring + " " +
		// getCurrentImage();
		//getNextToken();

		//checkSEMICOLON(cutpointmethodexecutionstring);

		//methodexecution = methodexecution + ";";

		getNextToken();

		connector.addCutpointMethodExecution(methodexecution);

		parseConnectorBody();
	}

	

	private String getAspectPackage(String decl) {

		int i = decl.lastIndexOf(".");
		if (i == -1)
			return "";

		else
			return decl.substring(0, i);
	}
	
	private void checkFactory(String type) {
		
		if(type==null)
			return;
		
		Class clz;
		try {
			clz = Options.loadClass(type);
		
		Class factory = IAspectFactory.class;
		
		if(factory.isAssignableFrom(clz))
			return;
		
		throw new JascoParseException(newError("custom aspect factory has to implement "+factory.getName()));
		} catch (ClassNotFoundException e) {
			throw new JascoParseException(newError("custom aspect factory class not found "+type));
		}
		
	}

	private void parseHookInitialisation(Vector declaration, String theerror,
			String aspectFactory) throws JascoParseException {

		PCutpointDeclaration cutpointdeclaration = new PCutpointDeclaration();
		cutpointdeclaration.setLine(this.getLine());
		checkFactory(aspectFactory);
		cutpointdeclaration.setAspectFactory(aspectFactory);

		String hookinitialisationstring = theerror;

		int numberofidentifiers = declaration.size();

		String hookidentifier = (String) declaration
				.elementAt(numberofidentifiers - 1);
		String aspectidentifier = (String) declaration
				.elementAt(numberofidentifiers - 2);

		String packageidentifier = "";
		for (int i = 0; i < (numberofidentifiers - 2); i++) {
			if (i == 0)
				packageidentifier = (String) declaration.elementAt(i);
			else
				packageidentifier = packageidentifier + "."
						+ (String) declaration.elementAt(i);
		}

		String fullPackage = "";
		if (!packageidentifier.equals(""))
			fullPackage = packageidentifier + ".";
		String declarationtype = fullPackage + aspectidentifier + "_"
				+ hookidentifier;

		String isAspect = isHook(declarationtype,connector.getImports());
		if (isAspect != null) {
			declarationtype = isAspect;
			
			cutpointdeclaration.setLine(getPreviousLine());
			
			checkIDENTIFIER(hookinitialisationstring);

			cutpointdeclaration.setCutpointDeclarationName(getCurrentImage());
			
			if(this.isHookInitialisation(cutpointdeclaration.getCutpointDeclarationName()))
				throw new JascoParseException(this.newError("variable name "+cutpointdeclaration.getCutpointDeclarationName()+" already defined!"));

			hookinitialisationstring = hookinitialisationstring + " "
					+ getCurrentImage();
			getNextToken();

			checkASSIGN(hookinitialisationstring);

			hookinitialisationstring = hookinitialisationstring + " "
					+ getCurrentImage();
			getNextToken();

			checkNEW(hookinitialisationstring);

			hookinitialisationstring = hookinitialisationstring + " "
					+ getCurrentImage();

			String type = "";
			while (getCurrentId() != LPAREN) {
				getNextToken();
				type += getCurrentImage();
				hookinitialisationstring = hookinitialisationstring + " "
						+ getCurrentImage();
				getNextToken();

				assertOrToken(getCurrentId(), DOT, LPAREN, new CompileError(
						getFileName(), getPreviousLine(), ". or ) expected "
								+ Options.NEWLINE
								+ hookinitialisationstring
								+ Options.NEWLINE
								+ generatePointer(hookinitialisationstring
										.length())));
				hookinitialisationstring = hookinitialisationstring + " "
						+ getCurrentImage();
				if (getCurrentId() == DOT)
					type += getCurrentImage();
				//getNextToken();
			}
			int index = type.lastIndexOf(".");
			String errorstring2 = "Not an aspect type: "
					+ type.replace('_', '.');
			CompileError error2 = new CompileError(getFileName(),
					getPreviousLine(), errorstring2);

			if (index == -1)
				throw new JascoParseException(error2);

			type = type.substring(0, index) + "_" + type.substring(index + 1);
			String fullType = isHook(type,connector.getImports());
			if (fullType == null)
				throw new JascoParseException(error2);
			
			cutpointdeclaration.setFullName(fullType);
			
			checkAssignable(declarationtype,fullType);
			
			checkAspectName(cutpointdeclaration);
			
			/*
			if (!fullType.equals(declarationtype))
				throw new JascoParseException(error2);
				*/
			/*
			 * for (int i = 0; i < numberofidentifiers ; i++) { if
			 * (!getCurrentImage().equals((String)declaration.elementAt(i))) {
			 * String errorstring = " '" + (String)declaration.elementAt(i) + "'
			 * expected'" + "\n\t" + hookinitialisationstring + "\n\t" +
			 * generatePointer(hookinitialisationstring.length()); CompileError
			 * error = new CompileError(getFileName(), getPreviousLine(),
			 * errorstring); throw new JascoParseException(error); } else {
			 * hookinitialisationstring = hookinitialisationstring + " " +
			 * getCurrentImage(); getNextToken();
			 * 
			 * if (i < (numberofidentifiers - 1)) {
			 * checkDOT(hookinitialisationstring);
			 * 
			 * hookinitialisationstring = hookinitialisationstring + " " +
			 * getCurrentImage(); getNextToken(); } } }
			 */

			checkLPAREN(hookinitialisationstring);

			hookinitialisationstring = hookinitialisationstring + " "
					+ getCurrentImage();
			getNextToken();

			while (getCurrentId() != RPAREN) {
				checkTYPEorLBRACE(hookinitialisationstring);

				PCutpointDeclarationParameter cutpointdeclarationparameter = new PCutpointDeclarationParameter();

				if (getCurrentId() == LBRACE) {
					hookinitialisationstring = hookinitialisationstring + " "
							+ getCurrentImage();
					getNextToken();

					Vector returnvalue = parseMultipleHookArguments(hookinitialisationstring);
					Vector parameters = (Vector) returnvalue.elementAt(0);
					String errorstr = (String) returnvalue.elementAt(1);

					for (int i = 0; i < parameters.size(); i++) {
						PCutpointDeclarationParameterInput parameter = (PCutpointDeclarationParameterInput) parameters
								.elementAt(i);
						cutpointdeclarationparameter
								.addCutpointDeclarationParameterInput(parameter);
					}
					hookinitialisationstring = errorstr;
				} else {
					Vector returnvalue = parseHookArgument(hookinitialisationstring);
					PCutpointDeclarationParameterInput parameter = (PCutpointDeclarationParameterInput) returnvalue
							.elementAt(0);
					String errorstr = (String) returnvalue.elementAt(1);

					cutpointdeclarationparameter
							.addCutpointDeclarationParameterInput(parameter);
					hookinitialisationstring = errorstr;
				}

				cutpointdeclaration
						.addCutpointDeclarationParameter(cutpointdeclarationparameter);

				String errorstring = " ',' or ')' expected" + "\n\t"
						+ hookinitialisationstring + "\n\t"
						+ generatePointer(hookinitialisationstring.length());
				CompileError error = new CompileError(getFileName(),
						getPreviousLine(), errorstring);
				assertOrToken(getCurrentId(), COMMA, RPAREN, error);

				if (getCurrentId() == COMMA) {
					hookinitialisationstring = hookinitialisationstring + " "
							+ getCurrentImage();
					getNextToken();
				}
			}

			checkParameters(cutpointdeclaration);

			hookinitialisationstring = hookinitialisationstring + " "
					+ getCurrentImage();

			getNextToken();

			//!!!!!!!!!!!!!!!!!!!!new CODE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
			checkSEMICOLONorLBRACE(hookinitialisationstring);

			if (getCurrentId() != SEMICOLON) {
				getNextToken();

				jasco.tools.jascoparser.PJavaClass tmpClass = new PJavaClass(
						"_JAsCo_RefinableMethodImpl");

				while (getCurrentId() != RBRACE) {
					parseJavaCode(tmpClass,connector.getImports());
				}

				Iterator i = tmpClass.getMethods();
				while (i.hasNext()) {
					PMethod method = (PMethod) i.next();
					PAbstractImplementation imp = new PAbstractImplementation(
							cutpointdeclaration, method); 
					cutpointdeclaration.addAbstractImplementation(imp);
				}

			}
			//!!!!!!!!!!!!!!!!!!!!new CODE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
			getNextToken();
			connector.addCutpointDeclaration(cutpointdeclaration);

			parseConnectorBody();
		} else {

			getNextToken();
			parseJavaCode(hookinitialisationstring);
		}
	}
	
	/**
	 * @param decl
	 */
	private void checkAspectName(PCutpointDeclaration decl) {
		
		String name = decl.getAspectType();
		
		if(name.equals(connector.getConnectorName()))
			throw new JascoParseException(newError("a connector can't have the same name as one of the deployed aspect beans"));
		
	}

	private void checkAssignable(String expected, String given) throws JascoParseException {
		
		if(!this.isCheckingClasses())
			return;
		
		String errorstring2 = "Incompatible type for "
				+ expected.replace('_', '.') + ": "
				+ given.replace('_', '.');
		CompileError error2 = new CompileError(getFileName(), getPreviousLine(),
				errorstring2);
		
		Class clzExpected = null;
		Class clzGiven = null;
		
		try{ 
			clzExpected = this.loadClassOrPrim(expected);
		}catch(ClassNotFoundException ex) {
			throw new JascoParseException(this.newError("Hook not found: "+expected.replace('_', '.')));
		}
		try{ 
			clzGiven = this.loadClassOrPrim(given);
		}catch(ClassNotFoundException ex) {
			throw new JascoParseException(this.newError("Hook not found: "+given.replace('_', '.')));
		}
			
		if(clzExpected.isAssignableFrom(clzGiven))
			return;
		else throw new JascoParseException(error2);
	}

	private void checkParameters(PCutpointDeclaration hook)
			throws JascoParseException {
		
		if(!isCheckingClasses())
			return;
		
		Class cl = null;
		java.lang.reflect.Field field = null;
		ConstructorMethod[] value = null;
		try {

			cl = Options.loadClass(hook.getFullClassName());
			field = cl
					.getField(jasco.tools.aspectparser.AspectClassesGenerator.CONSTRUCTOR_ARGS);

			value = (ConstructorMethod[]) field.get(null);
			/*
			 * should not be done with hotswap! if(value==null) throw new
			 * JascoParseException(new CompileError(getFileName(),
			 * getPreviousLine(),"Corrupt transformed class:
			 * "+hook.getFullClassName()+" (detail:
			 * "+jasco.aspectparser.AspectClassesGenerator.CONSTRUCTOR_ARGS+"has
			 * no value )"));
			 */
		} catch (Exception ex) {
			Logger.getInstance().showError(ex);
			ex.printStackTrace();
			throw new JascoParseException(new CompileError(getFileName(),
					getPreviousLine(),
					"Class not found or no correct version hook: "
							+ hook.getFullClassName()));
		}

		PCutpointDeclarationParameters parameters = hook
				.getCutpointDeclarationParameters();
		int paramsize = parameters.getCutpointDeclarationParametersSize();
		if (value.length != paramsize) {
			CompileError error = new CompileError(getFileName(),
					getPreviousLine(),
					"Wrong number of arguments for constructor of hook "
							+ hook.getFullClassName() + " ; " + value.length
							+ " expected while " + paramsize + " are supplied!");
			throw new JascoParseException(error);
		}

		for (int i = 0; i < parameters.getCutpointDeclarationParametersSize(); ++i) {
			PCutpointDeclarationParameter param = parameters
					.getCutpointDeclarationParameter(i);
			for (int j = 0; j < param
					.getCutpointDeclarationParameterInputSize(); ++j) {
				PCutpointDeclarationParameterInput input = param
						.getCutpointDeclarationParameterInput(j);
				checkCompatibility(input, value[i]);

			}
		}

	}

	public void checkCompatibility(PCutpointDeclarationParameterInput method,
			ConstructorMethod constmethod) throws JascoParseException {
		
		String returnType = method.getType();

		try {
			if (!hasWildcards(returnType) && !isEvent(returnType) && !isArray(returnType) && !isGenerics(returnType)) {
				
				Class returnCl = loadClassOrPrim(returnType);

				if (!constmethod.isCompatibleReturn(returnCl)) {
					
					String errorstr = "Return type of "
							+ method.getMethodName()
							+ " not compatible with return type defined in hook: "
							+ returnType;
					CompileError error = new CompileError(getFileName(),
							getPreviousLine(), errorstr);
					throw new JascoParseException(error);
				}
			}

			Vector v = method.getMethodParameters();
			for (int i = 0; i < v.size(); ++i) {
				String argname = (String) v.elementAt(i);
				if (hasWildcards(argname))
					continue;
				
				if (isArray(argname))
					continue;
				
				if (isGenerics(argname))
					continue;

				Class cl = loadClassOrPrim(argname);
				if (!constmethod.isCompatibleArg(cl, i)) {
					String errorstr = "Argument type of "
							+ method.getMethodName()
							+ " not compatible with argument type defined in hook: "
							+ argname;
					CompileError error = new CompileError(getFileName(),
							getPreviousLine(), errorstr);
					throw new JascoParseException(error);
				}
			}

		} catch (ClassNotFoundException ex) {
			CompileError mNotFound = new CompileError(getFileName(),
					getPreviousLine(),
					"System err @connectorparser.checkCompatibility: "
							+ ex.getMessage());
			throw new JascoParseException(mNotFound);
		}
	}

	private Vector parseMultipleHookArguments(String errorstr)
			throws JascoParseException {
		String theerror = errorstr;
		Vector parameters = new Vector();

		while (getCurrentId() != RBRACE) {
			Vector tempparam = parseHookArgument(theerror);
			parameters
					.addElement((PCutpointDeclarationParameterInput) tempparam
							.elementAt(0));
			theerror = (String) tempparam.elementAt(1);

			theerror = theerror + " " + getCurrentImage();

			String errorstring = " ',' or '}' expected" + "\n\t" + theerror
					+ "\n\t" + generatePointer(theerror.length());
			CompileError error = new CompileError(getFileName(),
					getPreviousLine(), errorstring);
			assertOrToken(getCurrentId(), COMMA, RBRACE, error);

			if (getCurrentId() == COMMA) {
				theerror = theerror + " " + getCurrentImage();
				getNextToken();
			}
		}

		theerror = theerror + " " + getCurrentImage();
		getNextToken();

		Vector returnvalue = new Vector();
		returnvalue.addElement(parameters);
		returnvalue.addElement(theerror);

		return returnvalue;
	}

	private Vector parseHookArgument(String errorString)
			throws JascoParseException {
		
		String hookargumentstring = errorString;
		
		String annotationString = null;
		if(getCurrentId()==ANNOTATION) { //hooks triggering on annotations
			getNextToken();
			annotationString = accumulateType(connector.getImports());
		}
		
		String typestring = accumulateType(connector.getImports());
		hookargumentstring+=typestring;

		String classstring = null;
		String methodstring = null;
		
		classstring=accumulateMethod(connector.getImports());
		hookargumentstring+=classstring;
		
		int index = classstring.lastIndexOf(".");
		if(index==-1)
			throw new JascoParseException(newError("no method name found "+classstring));
		
		methodstring=classstring.substring(index+1);
		classstring=classstring.substring(0,index);
		
		int matchSubtypes = MethodSignature.NO_SUBTYPES;
		if(getCurrentId()==PLUS) {
			matchSubtypes=MethodSignature.ALL_SUBTYPES;
			getNextToken();
		}
		else if(getCurrentId()==INCR) { 
			matchSubtypes=MethodSignature.IMPL_SUBTYPES;;
			getNextToken();
		} 
		
		checkLPAREN(hookargumentstring);

		hookargumentstring = hookargumentstring + " " + getCurrentImage();
		getNextToken();

		Vector arguments = new Vector();

		while (getCurrentId() != RPAREN) {

			String argument = accumulateType(connector.getImports());
			hookargumentstring = hookargumentstring + " " +argument;
			
			arguments.addElement(argument);

			String errorstring = " ',' or ')' expected" + "\n\t"
					+ hookargumentstring + "\n\t"
					+ generatePointer(hookargumentstring.length());
			CompileError error = new CompileError(getFileName(),
					getPreviousLine(), errorstring);
			assertOrToken(getCurrentId(), COMMA, RPAREN, error);

			if (getCurrentId() == COMMA) {
				hookargumentstring = hookargumentstring + " "
						+ getCurrentImage();
				getNextToken();
			}
		}

		hookargumentstring = hookargumentstring + " " + getCurrentImage();
		getNextToken();

		PCutpointDeclarationParameterInput cutpointdeclarationparameterinput = new PCutpointDeclarationParameterInput();
		cutpointdeclarationparameterinput.setType(typestring);
		cutpointdeclarationparameterinput.setClassName(classstring);
		cutpointdeclarationparameterinput.setMethodName(methodstring);
		cutpointdeclarationparameterinput.matchSubtypes(matchSubtypes);
		checkClassName(classstring,matchSubtypes);
		
		
		if(annotationString!=null)
			cutpointdeclarationparameterinput.setAnnotation(annotationString); 
		
		for (int i = 0; i < arguments.size(); i++) {
			String argname = (String) arguments.elementAt(i);
			cutpointdeclarationparameterinput.addMethodParameter(argname);
		}

		checkMethod(cutpointdeclarationparameterinput);

		Vector returnvalue = new Vector();
		returnvalue.addElement(cutpointdeclarationparameterinput);
		returnvalue.addElement(hookargumentstring);

		return returnvalue;
	}

	/**
	 * @param classstring
	 * @param matchSubtypes
	 */
	private void checkClassName(String classstring, int matchSubtypes) {
		if(matchSubtypes!=MethodSignature.NO_SUBTYPES&&this.hasWildcards(classstring)) {
			throw new JascoParseException(newError("matching subtypes not supported with wildcard class name:"+classstring));
		}
	}

	private void checkMethod(PCutpointDeclarationParameterInput method)
			throws JascoParseException {
		
		if(!this.isCheckingClasses())
			return;
		
		String clName = method.getClassName();
		String methodname = method.getMethodName();
		String returnType = method.getType();

		if (isEvent(returnType) || hasWildcards(clName)
				|| hasWildcards(methodname) || isGenerics(clName))
			return;

		try {
			Class cl = loadClassOrPrim(clName);
			//Class returnCl = loadClassOrPrim(returnType);

			Vector v = method.getMethodParameters();
			Class[] args = new Class[v.size()];
			boolean doParamCheck = true;
			for (int i = 0; i < v.size(); ++i) {
				String argname = (String) v.elementAt(i);
				if (hasWildcards(argname)||isArray(argname)||isGenerics(argname)) {
					doParamCheck = false;
					break;
				}
				args[i] = loadClassOrPrim(argname);
			}

			java.lang.reflect.Method theMethod = null;
			if (doParamCheck) {

				theMethod = cl.getDeclaredMethod(methodname, args);
			} else {

				java.lang.reflect.Method[] methods = cl.getDeclaredMethods();
				for (int i = 0; i < methods.length; ++i)
					if (methods[i].getName().equals(methodname)) {
						theMethod = methods[i];
						if (java.lang.reflect.Modifier.isPublic(methods[i]
								.getModifiers()))
							break;
					}
			}

			if (theMethod == null) {
				CompileError mNotFound = new CompileError(getFileName(),
						getPreviousLine(), "Method not found error: "
								+ method.getMethodName() + " in class "
								+ method.getClassName());
				throw new JascoParseException(mNotFound);
			} else if (!java.lang.reflect.Modifier.isPublic(theMethod
					.getModifiers()))
				Logger
						.getInstance()
						.showOutput(
								"Warning: method "
										+ methodname
										+ " in class "
										+ clName
										+ " is not a public method. Non-public methods are normally not trapped!");
		} catch (ClassNotFoundException ex) {
			CompileError mNotFound = new CompileError(getFileName(),
					getPreviousLine(), "Class not found error: "
							+ ex.getMessage() + " (system" + ex.getMessage());
			throw new JascoParseException(mNotFound);
		} catch (NoSuchMethodException ex) {
			CompileError mNotFound = new CompileError(getFileName(),
					getPreviousLine(), "Method not found error: "
							+ method.getMethodName() + " in class "
							+ method.getClassName());
			throw new JascoParseException(mNotFound);
		}
	}

	/*
	 * protected String checkClass(String name, ) throws JascoParseException {
	 * 
	 * String orig=name; if(hasWildcards(name)) return name;
	 * if(isPrimitive(name)) return name; if(isEvent(name)) return name;
	 * CompileError classNotFound = new CompileError(getFileName(),
	 * getPreviousLine(), "Class not found error: "+name); name=isClass(name);
	 * if (name == null) { int index = orig.lastIndexOf("."); if (index!=-1) {
	 * name=isClass(orig.substring(0,index)+"_"+orig.substring(index+1)); } }
	 * if(name==null) throw new JascoParseException(classNotFound); else return
	 * name; }
	 */
	/*
	 * private String checkMethod(String name,String classStr,Vector args)
	 * throws JascoParseException {
	 * 
	 * 
	 * 
	 * 
	 * if(hasWildcards(name)) return name; CompileError classNotFound = new
	 * CompileError(getFileName(), getPreviousLine(), "Method not found error in
	 * class "+classStr+": "+name); try { Class cl =
	 * Options.loadClass(classStr);
	 * 
	 * if(name == null) throw new JascoParseException(classNotFound); else
	 * return name;
	 *  } catch(ClassNotFoundException ex) { CompileError classNotFound = new
	 * CompileError(getFileName(), getPreviousLine(), "Class not found error:
	 * "+classStr); throw new JascoParseException(classNotFound); } }
	 */

	private void parseCutpointExecution(String cutpointdeclaration,
			String adaptation, String error) throws JascoParseException {

		String cutpointexecutionstring = error;
		String parameters = "";

		PCutpointExecution cutpointexecution = new PCutpointExecution();

		cutpointexecution.setExecutionNumber(cutpointexecutioncounter++);

		cutpointexecution.setCutpointExecutionName(cutpointdeclaration);
		if (adaptation.equals("before")) {
			cutpointexecution.setCutpointAdaptationType(0);
		}
		if (adaptation.equals("after")) {
			cutpointexecution.setCutpointAdaptationType(1);
		}
		if (adaptation.equals(IHook.REPLACE)) {
			cutpointexecution.setCutpointAdaptationType(2);
		}
		if (adaptation.equals("default")) {
			cutpointexecution.setCutpointAdaptationType(3);
		}

		checkLPAREN(cutpointexecutionstring);
		cutpointexecution.setLine(getPreviousLine());

		cutpointexecutionstring = cutpointexecutionstring + " "
				+ getCurrentImage();
		getNextToken();

		String errorstring = " ')' or <identifier> expected" + "\n\t"
				+ cutpointexecutionstring + "\n\t"
				+ generatePointer(cutpointexecutionstring.length());
		CompileError theerror = new CompileError(getFileName(),
				getPreviousLine(), errorstring);
		assertOrToken(getCurrentId(), IDENTIFIER, RPAREN, theerror);

		while (getCurrentId() != RPAREN) {
			parameters = parameters + getCurrentImage();

			checkIDENTIFIER(cutpointexecutionstring);

			cutpointexecutionstring = cutpointexecutionstring + " "
					+ getCurrentImage();
			getNextToken();

			errorstring = " ',' or ')' expected" + "\n\t"
					+ cutpointexecutionstring + "\n\t"
					+ generatePointer(cutpointexecutionstring.length());
			theerror = new CompileError(getFileName(), getPreviousLine(),
					errorstring);
			assertOrToken(getCurrentId(), COMMA, RPAREN, theerror);

			if (getCurrentId() == COMMA) {
				parameters = parameters + " " + getCurrentImage();
				getNextToken();
			}
		}

		cutpointexecution.setParameter(parameters);

		cutpointexecutionstring = cutpointexecutionstring + " "
				+ getCurrentImage();
		getNextToken();

		checkSEMICOLON(cutpointexecutionstring);

		getNextToken();

		connector.addCutpointExecution(cutpointexecution);

		parseConnectorBody();
	}

	public PConnector getParsedConnector() {
		return connector;
	}

	public boolean isCutpointDeclaration(String anIdentifier) {
		boolean result = false;
		PCutpointDeclarations declarations = connector
				.getCutpointDeclarations();
		for (int i = 0; i < declarations.getCutpointDeclarationsSize(); i++) {
			PCutpointDeclaration declaration = declarations
					.getCutpointDeclarations(i);
			if (anIdentifier.equals(declaration.getCutpointDeclarationName())) {
				result = true;
			}
		}
		return result;
	}

	private static final int REGULAR_JAVA = 0;

	private static final int HOOK_DECLARATION = 1;

	private static final int METHOD_EXECUTION = 2;

	private static final int ADD_COMBINATION_STRATEGY = 3;

	private static final int CRE_COMBINATION_STRATEGY = 4;
	
	private static final int ASPECT_DECLARATION = 5;

	private int checkType(Vector identifiers) throws JascoParseException {

		String type = (String) identifiers.elementAt(0);

		for (int i = 1; i < identifiers.size(); i++) {
			type = type + "." + (String) identifiers.elementAt(i);
		}

		int index = type.lastIndexOf(".");
		String aspectType = "";
		if (index > 0)
			aspectType = type.substring(0, index) + "_"
					+ type.substring(index + 1);

		if (identifiers.size() > 1) {
			String aspectType2 = isHook(aspectType,connector.getImports());
			
			if (isCutpointDeclaration((String) identifiers.elementAt(0))) {
				return METHOD_EXECUTION;
			} else if (aspectType2 != null) {
				return HOOK_DECLARATION;
			}
			else if (jasco.options.Options.isCombinationStrategy(type)) {
				return CRE_COMBINATION_STRATEGY;
			} else if(Options.isAspectBean(type)) {
				
				return ASPECT_DECLARATION;
		    }
			else if (jasco.options.Options.isClass(type)) {
				return REGULAR_JAVA;
			} else {
				String tempType = type;
				while (index > 0) {
					tempType = tempType.substring(0, index);
					index = tempType.lastIndexOf(".");
					aspectType = "";
					if (index > 0)
						aspectType = tempType.substring(0, index) + "_"
								+ tempType.substring(index + 1);
					if (jasco.options.Options.isClass(tempType)) {
						return REGULAR_JAVA;
					} else if (jasco.options.Options.isHook(aspectType)) {
						return REGULAR_JAVA;
					}

				}
				throw new JascoParseException(new CompileError(getFileName(),
						getLine(), "No class or hook def found error: " + type));

			}

		} else {

			return REGULAR_JAVA;
		}
	}
}