/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.tools.gui;

import jasco.runtime.connector.Connector;

import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.BorderFactory;
import javax.swing.ButtonGroup;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.border.BevelBorder;
import javax.swing.border.TitledBorder;

public class ConnectorTreeNode extends AbstractTreeNode {
	
	private Connector connector;
	
	public ConnectorTreeNode(Connector c)  {
	
		super(c.getName());
		connector=c;
	}
	
	public Connector getConnector()  {
	
		return connector;
	}
	
	public void initInfoPanel(JPanel infoPanel)  {
		
		JLabel top = createTop(getName());
		top.setBorder(BorderFactory.createBevelBorder(BevelBorder.RAISED));
		
		
		infoPanel.add("North",top);
		
		JPanel main = new JPanel();
		
		
		final JRadioButton cadapt = new JRadioButton("Static connector (adapt on classes)");
		final JRadioButton iadapt = new JRadioButton("Non-static connector (Adapt on specific instances)");
		
		if(getConnector().adaptOnClasses())
			cadapt.setSelected(true);
		else iadapt.setSelected(true);	
		
		ButtonGroup group  = new ButtonGroup();
		group.add(cadapt);
		group.add(iadapt);
		
		//cadapt.setEditable(false);
		//iadapt.setEditable(false);
		
		JPanel adaptonclasses = new JPanel();
		adaptonclasses.setBorder(new TitledBorder(""));
		adaptonclasses.setLayout(new GridLayout(2,1,20,20));
		adaptonclasses.add(cadapt);
		adaptonclasses.add(iadapt);
		
		final JRadioButton enabled = new JRadioButton("Connector is enabled");
		final JRadioButton disabled = new JRadioButton("Connector is disabled");
		
		if(getConnector().isEnabled())
			enabled.setSelected(true);
		else disabled.setSelected(true);	
		
		group  = new ButtonGroup();
		group.add(disabled);
		group.add(enabled);
		
		JPanel enabledPanel = new JPanel();
		enabledPanel.setBorder(new TitledBorder(""));
		enabledPanel.setLayout(new GridLayout(2,1,20,20));
		enabledPanel.add(enabled);
		enabledPanel.add(disabled);
		
		
		JPanel apply = new JPanel();
		apply.setLayout(new GridLayout(1,2,20,20));
		JButton doApply = new JButton("APPLY");
		apply.add(doApply);
		JButton doReload = new JButton("RELOAD");
		apply.add(doReload);
		
		doApply.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ev)  {
			
				getConnector().setEnabled(enabled.isSelected());
				getConnector().setAdaptOnClasses(cadapt.isSelected());
			}
		});
		doReload.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ev)  {
			
				if(getConnector().isEnabled())
					enabled.setSelected(true);
				else disabled.setSelected(true);
				
				if(getConnector().adaptOnClasses())
					cadapt.setSelected(true);
				else iadapt.setSelected(true);
			}
		});
		
		main.setLayout(new GridLayout(3,1,20,20));
		main.add(enabledPanel);
		main.add(adaptonclasses);
		main.add(apply);
		infoPanel.add(main);
		
	}
	
}
