package jasco.tools.jascoparser;

/*
 Copyright (C) 2002-2005 JAsCo Development Team
 http://ssel.vub.ac.be/jasco
 
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as
 published by the Free Software Foundation; either version 2 of the
 License, or (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.
 
 You should have received a copy of the GNU Lesser General Public
 License along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 USA
 */

import jasco.options.Options;
import jasco.util.javacompiler.CompileError;

import java.util.Iterator;
import java.util.Vector;

public class JascoParser extends JavaParser implements JascoParserConstants
{
	public static final int REFINABLEMOD =  0x08000000;
	
	public JascoParser(JTokenManager tokenmanager, String aFileName)  {
		super(tokenmanager,aFileName);
	}
	protected void checkSEMICOLONorLBRACE(String aErrorString) throws JascoParseException 
	{
		if ((getCurrentId() != SEMICOLON) && (getCurrentId() != LBRACE)) 
		{
			String errorstring = " ';' or '{' expected" + "\n\t" + aErrorString + "\n\t" + generatePointer(1 + aErrorString.length());
			CompileError error = new CompileError(getFileName(), getPreviousLine(), errorstring);
			throw new JascoParseException(error);
		} 
	}
	
	protected void checkASSIGN(String aErrorString) throws JascoParseException
	{
		String errorstring = " '=' expected" + "\n\t" + aErrorString + "\n\t" + generatePointer(1 + aErrorString.length());
		CompileError error = new CompileError(getFileName(), getPreviousLine(), errorstring);
		assertToken(getCurrentId(), ASSIGN , error);
	}
	
	protected void checkNEW(String aErrorString) throws JascoParseException
	{
		String errorstring = " 'new' expected" + "\n\t" + aErrorString + "\n\t" + generatePointer(1 + aErrorString.length());
		CompileError error = new CompileError(getFileName(), getPreviousLine(), errorstring);
		assertToken(getCurrentId(), NEW , error);
	}
	
	
	protected void checkTYPEorLBRACE(String aErrorString) throws JascoParseException 
	{
		
		
		if (getCurrentId() != ANNOTATION && getCurrentId() != WILDCARDIDENTIFIER && getCurrentId() != STAR && getCurrentId() != BOOLEAN && getCurrentId() != BYTE && getCurrentId() != CHAR && getCurrentId() != DOUBLE && getCurrentId() != FLOAT && getCurrentId() != INT && getCurrentId() != LONG  && getCurrentId() != SHORT && getCurrentId() != VOID && getCurrentId() != IDENTIFIER && getCurrentId() != LBRACE) 
		{
			String errorstring = " 'type' or '{' expected" + "\n\t" + aErrorString + "\n\t" + generatePointer(1 + aErrorString.length());
			CompileError error = new CompileError(getFileName(), getPreviousLine(), errorstring);
			throw new JascoParseException(error);
		}
	}
	
	
	protected void checkTYPE(String aErrorString) throws JascoParseException 
	{	
		if (getCurrentId() != WILDCARDIDENTIFIER && getCurrentId() != STAR && getCurrentId() != BOOLEAN && getCurrentId() != BYTE && getCurrentId() != CHAR && getCurrentId() != DOUBLE && getCurrentId() != FLOAT && getCurrentId() != INT && getCurrentId() != LONG  && getCurrentId() != SHORT && getCurrentId() != VOID && getCurrentId() != IDENTIFIER && getCurrentId() != PEROBJECT) 
		{
			String errorstring = " 'type' expected" + "\n\t" + aErrorString + "\n\t" + generatePointer(1 + aErrorString.length());
			CompileError error = new CompileError(getFileName(), getPreviousLine(), errorstring);
			throw new JascoParseException(error);
		}
	}
	
	protected void checkDOT(String aErrorString) throws JascoParseException
	{
		String errorstring = " '.' expected" + "\n\t" + aErrorString + "\n\t" + generatePointer(1 + aErrorString.length());
		CompileError error = new CompileError(getFileName(), getPreviousLine(), errorstring);
		assertToken(getCurrentId(), DOT , error);
	}
	
	protected void checkLPAREN(String aErrorString) throws JascoParseException
	{
		String errorstring = " '(' expected" + "\n\t" + aErrorString + "\n\t" + generatePointer(1 + aErrorString.length());
		CompileError error = new CompileError(getFileName(), getPreviousLine(), errorstring);
		assertToken(getCurrentId(), JascoParserConstants.LPAREN , error);
	}
	
	protected void checkRPAREN(String aErrorString) throws JascoParseException
	{
		String errorstring = " ')' expected" + "\n\t" + aErrorString + "\n\t" + generatePointer(1 + aErrorString.length());
		CompileError error = new CompileError(getFileName(), getPreviousLine(), errorstring);
		assertToken(getCurrentId(), JascoParserConstants.RPAREN , error);
	}
	
	protected void checkLBRACKET(String aErrorString) throws JascoParseException
	{
		String errorstring = " '[' expected" + "\n\t" + aErrorString + "\n\t" + generatePointer(1 + aErrorString.length());
		CompileError error = new CompileError(getFileName(), getPreviousLine(), errorstring);
		assertToken(getCurrentId(), JascoParserConstants.LBRACKET , error);
	}
	
	protected void checkRBRACE(String aErrorString) throws JascoParseException
	{
		String errorstring = " '}' expected" + "\n\t" + aErrorString + "\n\t" + generatePointer(1 + aErrorString.length());
		CompileError error = new CompileError(getFileName(), getPreviousLine(), errorstring);
		assertToken(getCurrentId(), JascoParserConstants.RBRACE , error);
	}
	
	

	protected void checkCOLON(String aErrorString) throws JascoParseException
	{
		String errorstring = " ':' expected" + "\n\t" + aErrorString + "\n\t" + generatePointer(1 + aErrorString.length());
		CompileError error = new CompileError(getFileName(), getPreviousLine(), errorstring);
		assertToken(getCurrentId(), JascoParserConstants.COLON , error);
	}
	
	protected int parseModifiers(StringBuffer annotations, PImports imports) {

		int mod = 0;
		
		while (true)  {
			
			int id = getCurrentId(); 
			switch (id)  {
			
			case JascoParserConstants.PUBLIC:
				mod|= java.lang.reflect.Modifier.PUBLIC;
				break;
				
			case JascoParserConstants.PROTECTED:
				mod|= java.lang.reflect.Modifier.PROTECTED;
				break;
				
			case JascoParserConstants.STATIC:
				mod|= java.lang.reflect.Modifier.STATIC;
				break;
				
			case JascoParserConstants.STRICTFP:
				mod|= java.lang.reflect.Modifier.STRICT;
				break;
				
			case JascoParserConstants.VOLATILE:
				mod|= java.lang.reflect.Modifier.VOLATILE;
				break;
				
			case JascoParserConstants.TRANSIENT:
				mod|= java.lang.reflect.Modifier.TRANSIENT;
				break;
				
			case JascoParserConstants.NATIVE:
				mod|= java.lang.reflect.Modifier.NATIVE;
				break;
				
			case JascoParserConstants.SYNCHRONIZED:
				mod|= java.lang.reflect.Modifier.SYNCHRONIZED;
				break;
				
			case JascoParserConstants.PRIVATE:
				mod|= java.lang.reflect.Modifier.PRIVATE;
				break;
				
			case JascoParserConstants.ABSTRACT:
				mod|= java.lang.reflect.Modifier.ABSTRACT;
				break;
				
			case JascoParserConstants.FINAL:
				mod|= java.lang.reflect.Modifier.FINAL;
				break;
				
			case JascoParserConstants.REFINABLE:
				mod|= REFINABLEMOD;
				break;
				
			case JascoParserConstants.ANNOTATION:
				parseAnnotation(annotations,imports);
				break;
				
			default:
				return mod;
			}
			if(id!=JascoParserConstants.ANNOTATION) 
				getNextToken();
		}
	}

	public static boolean isRefinable(int i) {
		
		return (i & REFINABLEMOD)!=0;
	}
	
	protected boolean isAbstractModifier(int mod) {
		
		return super.isAbstractModifier(mod) || isRefinable(mod);
	}
	
	protected String checkAnyType(String parsedType, PImports imports, int typeToParse) throws JascoParseException {
		
		if(typeToParse==TYPE_ASPECT) {
			String parsedTypeTemp=isHook(parsedType,imports);
			if(parsedTypeTemp==null) {
				CompileError classNotFound = new CompileError(getFileName(), getPreviousLine(), "Class not found error: "+parsedType);
				throw new JascoParseException(classNotFound);
			}
			else parsedType=parsedTypeTemp;
			return parsedType;
		}
		else return super.checkAnyType(parsedType,imports,typeToParse);
	}
	
	protected String checkClass(String name,PImports imports) {
		
		if(hasWildcards(name))
			return name;
		if(isEvent(name))
			return name;	
		
		else return super.checkClass(name,imports);
	}
	
	protected String isHook(String className, PImports imports) {
		
		if(!this.isCheckingClasses())
			return className;
		
		Vector packages = new Vector();
		Iterator i =imports.getImports();
		while (i.hasNext()) {
			PImport next = (PImport) i.next();
			String pack = next.getImportName();
			if (pack.endsWith("*"))
				packages.add(pack.substring(0, pack.length() - 1));
			else if (pack.endsWith("."+className)||pack.equals(className))
				if (Options.isHook(pack))
					return pack;
				else
					return null;
		}
		//packages.add(cutpointClass.getPackageName()+".");
		packages.add("java.lang.");
		packages.add("");
		
		i = packages.iterator();
		while (i.hasNext()) {
			String next = i.next().toString();
			String newTest = next + className;
			//System.out.println("testing:"+next+currentName);
			//Options.loadClass(next+className);
			boolean b = Options.isHook(newTest);
			if (b)
				return newTest;
			
			int index = newTest.lastIndexOf(".");
			if(index!=-1) {
				newTest=newTest.substring(0,index)+"_"+newTest.substring(index+1);
				
				b = Options.isHook(newTest);
				if (b) 
					return newTest;
				
			}
			
		}
		
		return null;
	}
	
	protected boolean hasWildcards(String name) {
		
		if((name.indexOf("*")!=-1)||(name.indexOf("?")!=-1))
			return true;
		else return false;
	}
	
	public static String getClassCode(String aClass) throws ClassNotFoundException {
		
		if(aClass.equals("boolean"))
			return "Z";
		else if(aClass.equals("int"))
			return "I";
		else if(aClass.equals("float"))
			return"F";
		else if(aClass.equals("double"))
			return "D";
		else if(aClass.equals("char"))
			return "C";
		else if(aClass.equals("byte"))
			return "B";
		else if(aClass.equals("short"))
			return "S";
		else if(aClass.equals("long"))
			return "J";
		else return "L"+aClass+";";
	}
	
	protected Class loadClassOrPrim(String aClass) throws ClassNotFoundException {
		
		if(aClass.equals("boolean"))
			return Boolean.TYPE;
		else if(aClass.equals("int"))
			return Integer.TYPE;
		else if(aClass.equals("float"))
			return Float.TYPE;
		else if(aClass.equals("double"))
			return Double.TYPE;
		else if(aClass.equals("char"))
			return Character.TYPE;
		else if(aClass.equals("byte"))
			return Byte.TYPE;
		else if(aClass.equals("short"))
			return Short.TYPE;
		else if(aClass.equals("long"))
			return Long.TYPE;
		else if(aClass.equals("void"))
			return Void.TYPE;	
		else return Options.loadClass(aClass);
	}
	
	protected Class loadClassOrPrimOrArray(String aClass) throws ClassNotFoundException {
		
		if(isArray(aClass)) {
			String classname = removeArrayStuff(aClass);
			classname = getClassCode(classname);
			int i = aClass.indexOf("[");
			if(i!=-1) {
				int n = (aClass.length()-i)/2;
				for(int j=0;j<n;++j)
					classname="["+classname;
			}
			return Options.loadClass(classname);
		}
		
		return loadClassOrPrim(aClass);
		
		
	}
	
	protected boolean isEvent(String name) {
		
		return name.equals("onevent");
	}
	@Override
	protected int tokenRPAREN() {
		return RPAREN;
	}
	@Override
	protected int tokenANNOTATION() {
		return ANNOTATION;
	}
	@Override
	protected int tokenLPAREN() {
		return LPAREN;
	}
	@Override
	protected int tokenRBRACE() {
		return RBRACE;
	}
	@Override
	protected int tokenLBRACE() {
		return LBRACE;
	}
	@Override
	protected int tokenRBRACKET() {
		return RBRACKET;
	}
	@Override
	protected int tokenLBRACKET() {
		return LBRACKET;
	}
	@Override
	protected int tokenIMPLEMENTS() {
		return IMPLEMENTS;
	}
	@Override
	protected int tokenEXTENDS() {
		return EXTENDS;
	}
	@Override
	protected int tokenCOMMA() {
		return COMMA;
	}
	@Override
	protected int tokenIDENTIFIER() {
		return IDENTIFIER;
	}
	@Override
	protected int tokenTHROWS() {
		return THROWS;
	}
	@Override
	protected int tokenSEMICOLON() {
		return SEMICOLON;
	}
	@Override
	protected int tokenASSIGN() {
		return ASSIGN;
	}
	@Override
	protected int tokenCLASS() {
		return CLASS;
	}
	@Override
	protected int tokenGT() {
		return GT;
	}
	@Override
	protected int tokenDOT() {
		return DOT;
	}
	@Override
	protected int tokenIMPORT() {
		return IMPORT;
	}
	@Override
	protected int tokenEOF() {
		return EOF;
	}
	@Override
	protected int tokenPACKAGE() {
		return PACKAGE;
	}
	@Override
	protected int tokenVARARGS() {
		return VARARGS;
	}
	@Override
	protected boolean isExistingClassName(String name) {
		return Options.isClass(name);
	}
	@Override
	protected int tokenSTATIC() {
		return STATIC;
	}
	@Override
	protected int tokenLT() {
		return LT;
	}
	@Override
	protected int tokenINTERFACE() {
		return INTERFACE;
	}
	@Override
	protected int tokenENUM() {
		return ENUM;
	}
	@Override
	protected int tokenATINTERFACE() {
		return ATINTERFACE;
	}
	@Override
	protected int tokenFINAL() {
		return FINAL;
	}
}
 