/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.util;

import java.io.IOException;


public class CommandRunner implements Runnable {
	private String command = null;
	private boolean typed = true;

	public CommandRunner(boolean typed) {
		this.typed = typed;
	}

	public CommandRunner() {
		this(true);
	}

	public void runCommandInNewThread(String cmd) {
		Thread runner = new Thread(this);
		this.command = cmd;
		runner.start();
	}

	public void run() {
		try {
			runCommand(this.command, typed);
		} catch (Exception e) {
		}
	}
	
	public static int runCommand(String cmd)  {
	
		return runCommand(cmd,false);
	}
	
	

	public static int runCommand(String cmd, boolean typed) {
		try {
			
			//typed=true;
			Runtime rt = Runtime.getRuntime();
			
			if (typed)
				System.out.println();
			if (typed)
				System.out.println("Executing: " + cmd);
			
			Process proc = getProcess(cmd);
			
			// any error message?
			StreamGobbler errorGobbler = new LogStreamGobbler(proc.getErrorStream(), "ERROR", typed);

			// any output?
			StreamGobbler outputGobbler =
				new LogStreamGobbler(proc.getInputStream(), "OUTPUT", typed);

			// kick them off
			errorGobbler.start();
			outputGobbler.start();

			// any error???
			int exitVal = proc.waitFor();
			if (typed)
				System.out.println("ExitValue: " + exitVal);
			
			errorGobbler.flush();
			outputGobbler.flush();
				
			return exitVal;
		} catch (Throwable t) {
			t.printStackTrace();
			return -1;
		}
	}
	
	public static Process getProcess(String cmd)  throws IOException
	{
	    Runtime rt = Runtime.getRuntime();
		String[] commandArray = commandArray(cmd);
		String newCmd ="";
		Process proc = null;
		String osName = System.getProperty("os.name");
		if (!(osName.equalsIgnoreCase("Linux") || osName.equalsIgnoreCase("Mac OS X")))  {
			for(int i=0;i<commandArray.length;++i)
			     	newCmd+=commandArray[i]+" ";
			//System.out.println(newCmd);
			//command as a String
			//System.out.println(newCmd);
		    proc = rt.exec(newCmd);	
		}	
		else  {
			//System.out.println(commandArray[2]);
			//a CommandArray
			proc = rt.exec(commandArray);
		}
		return proc;
	}

	public static String[] commandArray(String cmd) {
		String osName = System.getProperty("os.name");
		String[] commandArray;
		if (osName.equalsIgnoreCase("Linux") || osName.equalsIgnoreCase("Mac OS X")) {
			commandArray = new String[3];
			commandArray[0] = "/bin/sh";
			commandArray[1] = "-c";
			commandArray[2] = cmd;
		} else if (
			!osName.equalsIgnoreCase("Windows 95")
				&& !osName.equalsIgnoreCase("Windows 98")) {
			commandArray = new String[2];
			commandArray[0] = "cmd /C";
			commandArray[1] = cmd;
		} else {
			commandArray = new String[2];
			commandArray[0] = "command.com /C";
			commandArray[1] = cmd;
		}
		return commandArray;
	}

}