package jasco.util.distribution;

import java.io.BufferedInputStream;
import java.io.ByteArrayInputStream;
import java.io.FilterInputStream;
import java.io.InputStream;
import java.io.IOException;

public class JascoInputStream extends FilterInputStream {

	private byte[] header;

	private byte[] body;

	private int restOfBody;

	private boolean enableSettingCflow;

	final int headerSize = 4;

	public JascoInputStream(InputStream in) {
		super(new BufferedInputStream(in));
		this.header = new byte[4];
		this.body = new byte[1];
		this.restOfBody = 0;
		this.enableSettingCflow = true;
	}

	private void setCflowThreadList(String[] b) {
		/**
		 * if (! enableSettingCflow) return ; else enableSettingCflow = false;
		 */
		// String cflowlistArray = new String(b);
		// CflowList.setAll(cflowlistArray);
		// getCflowStringArray(b);
	}

	public synchronized int read() throws IOException {
		int n = read(body, 0, 1);
		if (n < 1)
			return n;
		else
			return body[0] & 0xff;
	}

	public synchronized int read(byte[] b) throws IOException {
		return read(b, 0, b.length);
	}

	public synchronized int read(byte[] b, int off, int len) throws IOException {
		String[] cflowlist = null;
		if (restOfBody == 0) {
			// skip a cflow header here.
			int t = readBuffer(header, 0, headerSize);
			if (t != headerSize)
				if (t == -1)
					return -1;
				else
					throw new IOException("bad header size: " + t);

			// set cflow values to this thread
			int cflowlistSize = read32bitInteger(header);
			System.out.println("IStream size of cflow list " + cflowlistSize);
			if (cflowlistSize != 0) {
				cflowlist = new String[cflowlistSize];
				for (int i = 0; i < cflowlistSize; i++) {
					int t2 = readBuffer(header, 0, headerSize);
					if (t2 != headerSize)
						if (t == -1)
							return -1;
						else
							throw new IOException("bad header size: " + t2);

					// set cflow values to this thread
					int cflowElementSize = read32bitInteger(header);
					System.out.println("IStream size of cflow Element " + cflowElementSize);
					byte[] cflowElement = new byte[cflowElementSize];
					int s = readBuffer(cflowElement, 0, cflowElementSize);
					if (s != cflowElementSize)
						if (s == -1)
							return -1;
						else
							throw new IOException("bad cflowlist size: " + s);
					cflowlist[i] = cflowElement.toString();
				}
				setCflowThreadList(cflowlist);
			}
			// skip a len header here
			int u = readBuffer(header, 0, headerSize);
			if (u != headerSize)
				if (u == -1)
					return -1;
				else
					throw new IOException("bad header size: " + u);
			restOfBody = read32bitInteger(header);
		}

		if (restOfBody < len)
			len = restOfBody;

		int v = in.read(b, off, len);
		restOfBody -= v;
		return v;
	}

	private int readBuffer(byte[] b, int off, int len) throws IOException {
		int t = 0;
		int n = 0;
		do {
			t = in.read(b, off + n, len - n);
			if (t >= 0)
				n = n + t;
			else if (n > 0)
				return n;
			else
				return -1;
		} while (n < len);
		return n;
	}

	public long skip(long n) throws IOException {
		throw new IOException("not supported");
	}

	public boolean markSupported() {
		return false;
	}

	public int available() throws IOException {
		int k = in.available();
		if (k < restOfBody)
			return k;
		else
			return restOfBody;
	}

	/**
	 * Get the cflow list in to a byte array
	 */
	public static String[] getCflowStringArray(byte[] b) {
		byte[] head = new byte[4];
		ByteArrayInputStream temp = new ByteArrayInputStream(b);
		temp.read(head, 0, 4);
		int len = read32bitInteger(head);
		String[] ans = new String[len];
		for (int i = 0; i < len; i++) {
			temp.read(head, 0, 4);
			int lenSubStr = read32bitInteger(head);
			byte[] tmp = new byte[lenSubStr];
			temp.read(tmp, 0, lenSubStr);
			ans[i] = tmp.toString();
			System.out.println("Stack " + i + ": " + ans[i]);
		}
		return ans;
	}

	/**
	 * Reads a 32bit integer at the index.
	 */
	public static int read32bitInteger(byte[] b) {
		return (b[0] << 24) | ((b[1] & 0xff) << 16) | ((b[2] & 0xff) << 8)
				| (b[3] & 0xff);
	}

}
