package jasco.util.distribution;

import java.io.BufferedOutputStream;
import java.io.ByteArrayOutputStream;
import java.io.FilterOutputStream;
import java.io.IOException;
import java.io.OutputStream;

public class JascoOutputStream extends FilterOutputStream {

	public JascoOutputStream(OutputStream o) {
		super(new BufferedOutputStream(o));
	}

	public synchronized void write(byte[] b, int off, int len)
			throws IOException {
		String[] cflowStrings = getCflowStringArray();  
		byte[] cflowlist = getCflowByteArray(cflowStrings);
		if (cflowlist == null) {
			out.write(write32bit(0), 0, 4);
		} else {
			out.write(write32bit(cflowStrings.length), 0, 4);
			out.write(cflowlist, 0, cflowlist.length);
		}
		out.write(write32bit(len), 0, 4);
		out.write(b, off, len);
	}
	
	/**
	 * Get the cflow string array
	 * */
	public static String[] getCflowStringArray(){
		return jasco.runtime.RuntimeContext.getStackContents();
	}

	/**
	 * Get the cflow list in to a byte array
	 */
	public static byte[] getCflowByteArray(String[] cflow) {
		ByteArrayOutputStream temp = new ByteArrayOutputStream();
		try {
			temp.write(write32bit(cflow.length));
			for (int i = 0; i < cflow.length; i++) {
				byte[] tmp = cflow[i].getBytes();
				temp.write(write32bit(tmp.length));
				temp.write(tmp);
			}
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		System.out.println("Tamanio del cflow "+ temp.size());
		return temp.toByteArray();
	}

	/**
	 * Return the byte representation of a 32bit integer.
	 */
	public static byte[] write32bit(int value) {
		byte[] ans = new byte[4];
		ans[0] = (byte) (value >>> 24);
		ans[1] = (byte) (value >>> 16);
		ans[2] = (byte) (value >>> 8);
		ans[3] = (byte) value;
		return ans;
	}

}
